#include "Global.H"
#include "Options.H"
#include "Game1_ST_ForceFieldMove.H"
#include "GameObj.H"    // fixme: remove dependancy on this for better decoupling!
#include "LinAlg.H"

using namespace std;

//#define TEST(x)
#define TEST(x) x
#define TEST2(x)
//#define TEST2(x) x

namespace Game1_SimpleTerrain
{
  //-----------------------------------------------------------------------------
  static Loc get_center_loc(const Object *obj)
  {
    Loc l;
    obj->get_center(l.x, l.y);
    return l;
  }

  //-----------------------------------------------------------------------------
  void ST_ForceFieldMove::add_options()
  {
    ::Options o("ST_ForceFieldMove");
    o.put("-nopush", "turns off pushing");
    o.put("-wpattract", "units are attracted to wps, and not forced to stay on the path");
  }

  //-----------------------------------------------------------------------------
  ST_ForceFieldMove::ST_ForceFieldMove(sint4 /*tiles_x_*/, sint4 /*tiles_y_*/, sint4 /*tile_points_*/, sint4 /*client_pID_*/, sint4 /*neutral_pID_*/) 
    : client_pID(client_pID)
  {
    sint4 seed = 0;
    ::Options::get("-seed", seed);
    rand.srand(seed);

    bool nopush = false;
    ::Options::get("-nopush", nopush);
    pushing_on = !nopush;

    ::Options::get("-wpattract", wp_attract);
  }

  ST_ForceFieldMove::~ST_ForceFieldMove() {};

  //-----------------------------------------------------------------------------
  void ST_ForceFieldMove::execute_tasks(ST_Terrain::Tasks &tasks,
                                        ST_Terrain::Objects &objsToStop,
                                        Vector<MoveCmd> &cmds, 
                                        Vector<StatusMsg> &msgs)
  {
  
    //clear the reserved locations in the object map
    FORALL(objects, it){
      it->second = Loc(-1, -1);
    }
    // Execute all active tasks
    /*for (Tasks::iterator j = tasks.begin(); j != tasks.end();) {
      Tasks::iterator i = j++;*/
    FORALL(tasks, i) {

      // Report failure of planning (no path at all)
      if (i->noPathFound ) {
	TerrainBase::StatusMsg s;
	// fixme: this is not robust to future changes in the meaning of TaskId.
	s.task_id = (TerrainBase::TaskId)&(*i);
	s.obj = *(i->task.objs.begin());
	s.type = TerrainBase::StatusMsg::NO_PATH_FAILURE;
	msgs.push_back(s);
	continue;
      }

      // Skip invalid tasks.
      if (!i->isValid() ) { 
	TEST(cout << *i->task.objs.begin() << " task not valid" << endl;);
	continue;
      }

      UnitPath &cur_path = *(i->path);
      const Object *obj = cur_path.obj;

      // Skip tasks that are waiting for a retry
      if (cur_path.sleeping()) {
	TEST(cout << obj << " path sleeping" << endl;);
	continue;
      }

      // Skip objects for which we gave a (move) action that is still pending
      /*
	if (obj->is_pending_action()) {
	TEST(cout << obj << " pending action" << endl;)
	continue;
	}
      */

      bool atGoal = is_at_goal(obj, i->task, i);

      // Look ahead: if the object will get to a waypoint in the next tick,
      // give the next action now the prevent it from wasting a tick on stopping.
      // This will cut a single tick at each waypoint, though it will really prevent 
      // the object from stopping only if there's no (network or processing) lag
      // between this client and the server.
      // atWaypoint can be true only if the current waypoint is an intermediate and
      // not the *final* waypoint/goal.
      //   bool atWaypoint = cur_path.current_target > 0 &&  // reverse order, goal==0
      //	is_at_location(obj, cur_path.path.at(cur_path.current_target));

      // Skip a moving object which is currently not at a waypoint/goal
      if (/*(obj->get_speed() > 0 || obj->get_moving() != 0) && !atWaypoint && */ !atGoal ) {
	//TEST(cout << obj << " atWaypoint=" << atWaypoint << " speed=" << obj->get_speed() << " is_moving=" << obj->get_moving() << endl;);
	Loc target;
	//	bool deflected = get_forcefield_target(cur_path, target);
	//	bool nearWaypoint = is_near_waypoint(obj, cur_path.path.at(cur_path.current_target));
	// 	if (nearWaypoint && deflected && cur_path.current_target > 0){
	// 	  bool valid = slide_wp(cur_path);
	// 	  if (!valid){
	// 	    i->isFailed = true;
	// 	    continue;
	// 	  }
	get_forcefield_target(cur_path, target);
	//	}
	add_move_command(obj, target, cmds);

	///////////////////////////////
	
	Loc obj_center;
	obj->get_center(obj_center.x, obj_center.y);

// 	FORALL(i->followers, fol_it){
// 	  get_forcefield_target(*fol_it, obj_center, target);
// 	  add_move_command(*fol_it, target, cmds);
// 	}
	
	real8 distance = obj_center.distance(cur_path.path.at(cur_path.current_target));
	if (cur_path.spot != Loc(-1,-1)){
	  real8 spot_distance = cur_path.spot.distance(cur_path.path.at(cur_path.current_target));
	  if (distance >= spot_distance/* - ((real8)obj->get_max_speed())/2*/){
	    if (cur_path.collision_step > 5){
	      if (i->replanned < 1){ //fixed cap on number of replannings
	       i->isFailed = true;
	      }
	    }
	    if (cur_path.collision_step > 20){
	      TerrainBase::StatusMsg s;
	      //    fixme: this is not robust to future changes in the meaning of TaskId.
	      s.task_id = (TerrainBase::TaskId)&(*i);
	      s.obj = obj;
	      s.type = TerrainBase::StatusMsg::MOVEMENT_FAILURE;
	      msgs.push_back(s);
	      cout << "Movement failure!" << endl;
	    }
	   
	    cur_path.collision_step++;
	  }
	  else{
	    cur_path.collision_step = 0;
	    cur_path.spot = obj_center;
	  }
	}
	else{
	  cur_path.spot = obj_center;
	  cur_path.collision_step = 0;
	}
	
	
	///////////////////////////////

	continue;
      }



      // ok, so it needs attention:
      // 1. maybe it had reached the goal
      // 2. maybe it's at a waypoint (with the lookahead)
      // 3. maybe it needs the first orders
      // 4. maybe it unexpectedly collided with something

      if (atGoal ) {

	TEST(cout << obj << " arrived" << endl;)

	  TerrainBase::StatusMsg s;
	// fixme: this is not robust to future changes in the meaning of TaskId.
	s.task_id = (TerrainBase::TaskId)&(*i);
	s.obj = obj; 
	s.type = TerrainBase::StatusMsg::ARRIVED;
	msgs.push_back(s);	

      }
  //     else if( atWaypoint || 
// 		 cur_path.no_orders_yet || 
// 		 cur_path.stopped_to_avoid_collision ) {

// 	// in all these cases we should just issue a move command

// 	TEST(cout << obj << " atWaypoint=" << atWaypoint
// 	     << " no_orders_yet=" << cur_path.no_orders_yet
// 	     << " stopped_to_avoid_collision=" << cur_path.stopped_to_avoid_collision
// 	     << endl; );

// 	// if we are at a waypoint, move to next waypoint
//  	if( atWaypoint )
//  	  cur_path.current_target--;
// 	cur_path.no_orders_yet = false;
// 	cur_path.stopped_to_avoid_collision = false;
// 	Loc target;
// 	get_forcefield_target(cur_path, target);
// 	add_move_command(obj, target, cmds);

//       } 
      else {

	// Unexpected stop. The current policy is:
	// 1-2. First time => wait a little and send the move command again.
	// 3-4. Second stop at the same place => mark as failed and replan.
	// 5. Third stop at the same place => report failure.

// 	TEST(cout << obj << " unexpectedly stopped, collision_step=" << cur_path.collision_step << endl);

// 	Loc cur_position; obj->get_center(cur_position.x, cur_position.y);

// 	// First collision in this position:
// 	if (cur_path.spot != cur_position ) {
// 	  cur_path.collision_step = 1;
// 	  cur_path.mark_spot(cur_position);
// 	  // sleep for a short random time, 500-999 millisecs
// 	  cur_path.sleep(500 + (rand.rand_uint4() % 500));
// 	} else {
// 	  cur_path.collision_step++;

// 	  // Just finished the sleeping:
// 	  if (cur_path.collision_step == 2 ) {
// 	    Loc target;
// 	    get_forcefield_target(cur_path, target);
// 	    add_move_command(obj, target, cmds);
// 	  }
// 	  // Second collision
// 	  else if (cur_path.collision_step == 3 ) {
// 	    i->isFailed = true;
// 	  }
// 	  // Back from replanning
// 	  else if (cur_path.collision_step == 4 ) {
// 	    Loc target;
// 	    get_forcefield_target(cur_path, target);
// 	    add_move_command(obj, target, cmds);
// 	  }
// 	  // Third collision
// 	  else if (cur_path.collision_step == 5 ) {
// 	    TerrainBase::StatusMsg s;
// 	    // fixme: this is not robust to future changes in the meaning of TaskId.
// 	    s.task_id = (TerrainBase::TaskId)&(*i);
// 	    s.obj = obj;
// 	    s.type = TerrainBase::StatusMsg::MOVEMENT_FAILURE;
// 	    msgs.push_back(s);
// 	  }
//	}
      }
    }

    //Pushing!
    if (pushing_on){
      FORALL(objects, objIt){
	const GameObj *gob = dynamic_cast<const GameObj*>(objIt->first);
	if (gob){
	  if (objIt->second == Loc(-1,-1) && *(gob->sod.owner) == client_pID && objIt->first->get_max_speed() > 0){
	    Loc target;
	    get_push_target(objIt->first, target);
            if( target != get_center_loc(objIt->first) )
	      add_move_command(objIt->first, target, cmds);
	  }
	}
      }
    }
   
 // Priority of commands:
    // 1. stops to avoid a collision.
    // 2. moves.
    // 3. other stops (resulting from a cancelled task, etc.).
    
    // 2 above 3:
    FORALL(cmds, c) {
      objsToStop.erase(c->obj);
    }

    // With all movement commands were given while assuming an ideal-no-collisions
    // world, now check if any of our controlled objects are going to collide
    // (because of an existing movement vector, and also consider commands
    // we just intended to give).
    //resolve_friendly_collisions();

    // 1 above 2:
    for(typeof(cmds.begin()) c = cmds.begin() ; c != cmds.end() ; ) {
      FIND(objsToStop, s, c->obj);
      if( s != objsToStop.end() )
	c = cmds.erase(c);
      else
	c++;
    }

    // Return also the stop commands
    FORALL(objsToStop, oi) {
      const Object *obj = *oi;

      TerrainBase::MoveCmd m;
      m.obj = obj;
      obj->get_center(m.next_loc.x, m.next_loc.y); // doesn't really matter
      m.speed = 0; // 0 -> stop
      cmds.push_back(m);
    }
    objsToStop.clear();
  }

  void ST_ForceFieldMove::add_move_command(const Object *obj, Loc target, Vector<TerrainBase::MoveCmd> &cmds){
    //hack: 
    if (abs(target.x) >= 500000){
       target.x = 499999;
     }
     if( abs(target.y) >= 500000){
         target.y = 499999;
     }

    TerrainBase::MoveCmd m;
    m.obj = obj;
    m.next_loc = target;
    m.speed = obj->get_max_speed();
    cmds.push_back(m);

    TEST2(cout << obj << " set move command to " << target << endl; );
  
    FIND (objects, it, obj);
    // assert(it != objects.end() && "hrmm... all objects should be in objects map");
    if (it != objects.end()){
      Loc curr_loc;
      obj->get_center(curr_loc.x, curr_loc.y);
      Vec_2D vec(target.x - curr_loc.x, target.y - curr_loc.y);
      
      real8 k = ((real8)obj->get_max_speed()) / vec.length();
      
      vec = vec.multiply(k);
 
      it->second.x = curr_loc.x + vec.x;
      it->second.y = curr_loc.y + vec.y;
    }
  
  }

  //-----------------------------------------------------------------------------
  sint4 ST_ForceFieldMove::what_team(const Object *obj)
  {
    const GameObj *p = dynamic_cast<const GameObj*>(obj);
    if (p){
      return *(p->sod.owner);
    }
    else
      return -1;
  }

  void ST_ForceFieldMove::get_push_target(const Object *obj, Loc &output){
        //const Object *obj = path.obj;
    if (obj->get_shape() != Object::CIRCLE){
      return; //only circle movement implemented -- this also functions as a crude check for buildings
    }
  

    Vec_2D force_vec(0,0);

    FORALL(objects, i){
      const GameObj *gob = dynamic_cast<const GameObj*>(i->first);
      if (!gob){continue;}
      
      //don't consider any object that's not your own as a force that can push you
      if (*(gob->sod.owner) != client_pID){continue;}
      
      if (obj == i->first){
	continue;
      }
      Vec_2D vec;
      if (i->first->get_shape() == Object::CIRCLE){
	Loc obstacle_center;
	if (i->second == Loc(-1, -1)){
	  i->first->get_center(obstacle_center.x, obstacle_center.y);
	}
	else{
	  obstacle_center = i->second;
	}
	vec = get_point_repulsion(obj, obstacle_center, i->first->get_radius());
	force_vec = force_vec + vec;
      }
      else if (i->first->get_shape() == Object::RECTANGLE){
	Loc u_l, u_r, l_r, l_l; //upper-left, upper-right, lower-right, lower-left
	i->first->get_p1(u_l.x, u_l.y);
	i->first->get_p2(l_r.x, l_r.y);
	u_r.x = l_r.x;
	u_r.y = u_l.y;
      
	l_l.x = u_l.x;
	l_l.y = l_r.y;
      
	Segment seg;
	seg.l1 = u_l;
	seg.l2 = u_r;      
	vec = get_line_repulsion(obj, seg);
	//force_vec = force_vec + vec;

	seg.l1 = u_r;
	seg.l2 = l_r;
	vec = vec + get_line_repulsion(obj, seg);
	//force_vec = force_vec + vec;

	seg.l1 = l_r;
	seg.l2 = l_l;
	vec = vec + get_line_repulsion(obj, seg);
	//force_vec = force_vec + vec;

	seg.l1 = l_l;
	seg.l2 = u_l;
	vec = vec + get_line_repulsion(obj, seg);
	//force_vec = force_vec + vec;
	if (vec == Vec_2D(0,0)){
	  vec = vec + get_point_repulsion(obj, u_l);
	  vec = vec + get_point_repulsion(obj, u_r);
	  vec = vec + get_point_repulsion(obj, l_l);
	  vec = vec + get_point_repulsion(obj, l_r);
	}
	force_vec = force_vec + vec;
	
      }
    }
    Vec_2D wall_repulsion;
    FORALL(segment_ends, ends_it){
      ends_it->second = true;
    }
    FORALL(segments, seg_it){
      Vec_2D vec;
      vec = get_line_repulsion(obj, *seg_it);
      if (vec != Vec_2D(0,0)){
	FIND(segment_ends, ends_it, seg_it->l1);
	if (ends_it != segment_ends.end())
	  ends_it->second = false;
	FIND(segment_ends, ends_it2, seg_it->l2);
	if (ends_it != segment_ends.end())
	  ends_it->second = false;
      }
      wall_repulsion = wall_repulsion + get_line_repulsion(obj, *seg_it);
      // force_vec = force_vec + vec;
    }
    FORALL(segment_ends, endit){
      if (endit->second){
	wall_repulsion = wall_repulsion + get_point_repulsion(obj, endit->first);
      }
    }
    force_vec = force_vec + wall_repulsion;

    if (force_vec != Vec_2D(0,0)){
      Vec_2D rand_vec;
      Vec_2D half_force = Vec_2D(force_vec.x/2, force_vec.y/2);
      
      uint4 rand_angle_int = rand.rand_uint4() % 62;
      real8 rand_angle = ((real8)rand_angle_int)/ 10;
      rand_vec.x = Vec_2D::round(half_force.length() * cos(rand_angle));
      rand_vec.y = Vec_2D::round(half_force.length() * sin(rand_angle));
      
      force_vec = force_vec + rand_vec;
    }

   //  if (force_vec == Vec_2D(0,0)){
//       output = path.path.at(path.current_target);
//       return false;
//     }
    
    Loc object_center;
    obj->get_center(object_center.x, object_center.y);
    
    output.x = object_center.x + force_vec.x;
    output.y = object_center.y + force_vec.y;
  }

  void ST_ForceFieldMove::get_forcefield_target(const Object *obj, const Loc &attractor, Loc &output){
    //const Object *obj = path.obj;
    assert(obj->get_shape() == Object::CIRCLE && "Only Circle Movement implemented");
  

    Vec_2D force_vec(0,0);

    FORALL(objects, i){
      if (obj == i->first){
	continue;
      }
      Vec_2D vec;
      if (i->first->get_shape() == Object::CIRCLE){
	Loc obstacle_center;
	if (i->second == Loc(-1, -1)){
	  i->first->get_center(obstacle_center.x, obstacle_center.y);
	}
	else{
	  obstacle_center = i->second;
	}
	vec = get_point_repulsion(obj, obstacle_center, i->first->get_radius());
	force_vec = force_vec + vec;
      }
      else if (i->first->get_shape() == Object::RECTANGLE){
	Loc u_l, u_r, l_r, l_l; //upper-left, upper-right, lower-right, lower-left
	i->first->get_p1(u_l.x, u_l.y);
	i->first->get_p2(l_r.x, l_r.y);
	u_r.x = l_r.x;
	u_r.y = u_l.y;
      
	l_l.x = u_l.x;
	l_l.y = l_r.y;
      
	Segment seg;
	seg.l1 = u_l;
	seg.l2 = u_r;      
	vec = get_line_repulsion(obj, seg);
	//force_vec = force_vec + vec;

	seg.l1 = u_r;
	seg.l2 = l_r;
	vec = vec + get_line_repulsion(obj, seg);
	//force_vec = force_vec + vec;

	seg.l1 = l_r;
	seg.l2 = l_l;
	vec = vec + get_line_repulsion(obj, seg);
	//force_vec = force_vec + vec;

	seg.l1 = l_l;
	seg.l2 = u_l;
	vec = vec + get_line_repulsion(obj, seg);
	//force_vec = force_vec + vec;
	if (vec == Vec_2D(0,0)){
	  vec = vec + get_point_repulsion(obj, u_l);
	  vec = vec + get_point_repulsion(obj, u_r);
	  vec = vec + get_point_repulsion(obj, l_l);
	  vec = vec + get_point_repulsion(obj, l_r);
	}
	force_vec = force_vec + vec;
	
      }
    }
    Vec_2D wall_repulsion;
    FORALL(segment_ends, ends_it){
      ends_it->second = true;
    }
    FORALL(segments, seg_it){
      Vec_2D vec;
      vec = get_line_repulsion(obj, *seg_it);
      if (vec != Vec_2D(0,0)){
	FIND(segment_ends, ends_it, seg_it->l1);
	if (ends_it != segment_ends.end())
	  ends_it->second = false;
	FIND(segment_ends, ends_it2, seg_it->l2);
	if (ends_it != segment_ends.end())
	  ends_it->second = false;
      }
      wall_repulsion = wall_repulsion + get_line_repulsion(obj, *seg_it);
      // force_vec = force_vec + vec;
    }
    FORALL(segment_ends, endit){
      if (endit->second){
	wall_repulsion = wall_repulsion + get_point_repulsion(obj, endit->first);
      }
    }
    force_vec = force_vec + wall_repulsion;
  
   //  if (force_vec == Vec_2D(0,0)){
//       output = path.path.at(path.current_target);
//       return false;
//     }
    
    Vec_2D attract_vec, scaled_attract_vec;
    Loc obj_center;
    obj->get_center(obj_center.x, obj_center.y);
    attract_vec = Vec_2D(obj_center, attractor);
         
    real8 k;
    if (attract_vec.length() < 1){
      k = ATTRACTOR_FORCE;
    }
    else{
      k = ATTRACTOR_FORCE / attract_vec.length();
    }
    scaled_attract_vec = attract_vec.multiply(k); 
  
    force_vec = force_vec + scaled_attract_vec;
  
    restrict_vector(attract_vec, force_vec);

    if (force_vec == Vec_2D(0,0)){
      force_vec = Vec_2D(attract_vec.y, -attract_vec.x);
    }
    
    Loc object_center;
    obj->get_center(object_center.x, object_center.y);
    
    output.x = object_center.x + force_vec.x;
    output.y = object_center.y + force_vec.y;
 
  }

  void ST_ForceFieldMove::get_forcefield_target(UnitPath &path, Loc &output){
    Loc attractor = get_attractor(path);
    get_forcefield_target(path.obj, attractor, output);
  }
  

  Loc ST_ForceFieldMove::get_attractor(UnitPath &path){
    Loc object_center, attractor;
    Vec_2D to_wp, scaled_to_wp;
    path.obj->get_center(object_center.x, object_center.y);
    if (path.current_target >= ((sint4)path.path.size()) - 1){
      attractor = path.path.at(path.current_target);
    }
    else{
      attractor = find_point_on_line(object_center, path.path.at(path.current_target), path.path.at(path.current_target+1));
    }
    to_wp = Vec_2D(path.path.at(path.current_target).x - attractor.x, path.path.at(path.current_target).y - attractor.y);
    real8 offset = path.obj->get_max_speed() * PATH_LOOK_AHEAD;
    //std::cout << "k = " << offset/to_wp.length() << "\n";
    if (to_wp.length() == 0)
      scaled_to_wp = Vec_2D(1,1);
    else
      scaled_to_wp = to_wp.multiply(offset / to_wp.length());

    //If wp_attract is set, set the attractor to the wp unless you're near it.
    if (wp_attract && scaled_to_wp.length() <= to_wp.length()){
      return path.path.at(path.current_target);
    }
    while (scaled_to_wp.length() > to_wp.length()){
      offset = offset - to_wp.length();
      attractor = path.path.at(path.current_target);
      if (path.current_target == 0){
	scaled_to_wp = Vec_2D(0,0);
	break;
      }
      path.current_target--;
      //std::cout << "new wp: " << path.path.at(path.current_target).x 
      //		<< ", " << path.path.at(path.current_target).y << "\n"; 
      to_wp = Vec_2D(path.path.at(path.current_target).x - attractor.x, path.path.at(path.current_target).y - attractor.y);
      scaled_to_wp = to_wp.multiply(offset / to_wp.length());
    }
    
    attractor.x = attractor.x + scaled_to_wp.x;
    attractor.y = attractor.y + scaled_to_wp.y;
    
    return attractor;
  }
  
#if 0
  Loc ST_ForceFieldMove::find_point_on_line(const Loc &loc, const Loc &line_start, const &Loc line_end){
    
    Vec_2D line_to_obj, n;
   
    line_to_obj = Vec_2D(loc.x - line_start.x, loc.y - line_start.y);
    n = Vec_2D(line_end.y - line_start.y, -(line_end.x - line_start.x)); //perpendicular to segment
  
    line_to_obj = line_to_obj.project_onto(n);
  
    Loc point_on_line(loc.x - line_to_obj.x, loc.y - line_to_obj.y);
    
    return point_on_line;
 
  }
#else
  Loc ST_ForceFieldMove::find_point_on_line(const Loc &loc, const Loc &line_start, const Loc &line_end)
  {
    sint4 line_x = line_end.x - line_start.x;
    sint4 line_y = line_end.y - line_start.y;

    sint4 v_x = loc.x - line_start.x;
    sint4 v_y = loc.y - line_start.y;

    sint4 a = line_x * v_x + line_y * v_y;
    sint4 b = line_x * line_x + line_y * line_y;

    return Loc(line_start.x + line_x * a / b, line_start.y + line_y * a / b);
  }
#endif

  //-------------------------------------------------------------------

//   bool ST_ForceField::slide_wp(UnitPath &path){
  
//     Loc next_wp, curr_wp, temp_wp;
//     next_wp = path.path.at(path.current_target -1);
//     curr_wp = path.path.at(path.current_target);
//     real8 d = next_wp.distance(curr_wp);
  
//     Vec_2D wp_vector(next_wp.x - curr_wp.x, next_wp.y - curr_wp.y);
  
//     for(uint4 i = 1; ; i++){
//       real8 k = ((real8)(i * path.obj->get_max_speed())) / d;
//       if (k > 1){
// 	return false;
//       }
//       Vec_2D scaled_wp_vec = wp_vector.multiply(k);
   
//       temp_wp.x = curr_wp.x + scaled_wp_vec.x;
//       temp_wp.y = curr_wp.y + scaled_wp_vec.y;
//       if (!is_near_waypoint(path.obj, temp_wp))
// 	break;
//     }
//     path.path.at(path.current_target) = temp_wp;
//     return true;
//   }

  void ST_ForceFieldMove::restrict_vector(const Vec_2D &target_vec, Vec_2D &vec){

    real8 dot_prod = vec.dot_prod(target_vec);
    if (dot_prod < 0){
      Vec_2D perpendicular(target_vec.y, -target_vec.x);
      vec = vec.project_onto(perpendicular);
    }
  
  
  }

  Vec_2D ST_ForceFieldMove::get_point_repulsion(const Object *moving_obj, const Loc &point, sint4 radius){
    Loc center;
    moving_obj->get_center(center.x, center.y);
    real8 center_to_point_dist = center.distance(point);
    real8 d = center_to_point_dist - moving_obj->get_radius() - radius +1;
  
    if (d < moving_obj->get_max_speed() * FIELD_RANGE_MULTIPLIER){
      real8 k = REPULSOR_BASELINE / (d * d * center_to_point_dist);
      return Vec_2D( (sint4)((center.x - point.x)*k) , (sint4)((center.y - point.y)*k) );
      /*vec.x = center.x - point.x;
      vec.y = center.y - point.y;
      real8 k = REPULSOR_BASELINE / (d * d * vec.length());
      vec = vec.multiply(k);*/
    }
    return Vec_2D();
  }

#if 0
  Vec_2D ST_ForceFieldMove::get_line_repulsion(const Object *moving_obj, const Segment &seg){
    Loc q, p;
    Vec_2D line_to_obj, n;
    moving_obj->get_center(q.x, q.y);
    p = seg.l1;
    line_to_obj = Vec_2D(q.x - p.x, q.y - p.y);
    n = Vec_2D(seg.l1.y - seg.l2.y, -(seg.l1.x - seg.l2.x)); //perpendicular to segment
  
    line_to_obj = line_to_obj.project_onto(n);
  
    Loc point_on_line(q.x - line_to_obj.x, q.y - line_to_obj.y);
    Vec_2D to_l1, to_l2;
  
    to_l1 = Vec_2D(seg.l1.x - point_on_line.x, seg.l1.y - point_on_line.y);
    to_l2 = Vec_2D(seg.l2.x - point_on_line.x, seg.l2.y - point_on_line.y);
  
    if (to_l1.dot_prod(to_l2) > 0){ //if both vectors are pointing in the same direction
    //   if(to_l1.length() > to_l2.length()){
// 	return get_point_repulsion(moving_obj, seg.l2);
//       }
//       else{
// 	return get_point_repulsion(moving_obj, seg.l1);
//       }
    }
    else{
      real8 d = q.distance(point_on_line) - moving_obj->get_radius() +1;
      if (d < moving_obj->get_max_speed() * FIELD_RANGE_MULTIPLIER){
	real8 k = REPULSOR_BASELINE / (d * d * line_to_obj.length());
	return line_to_obj.multiply(k);
      }
    }
    return Vec_2D(0,0);
  }
#else
  Vec_2D ST_ForceFieldMove::get_line_repulsion(const Object *moving_obj, const Segment &seg)
  {
    Loc loc;
    moving_obj->get_center(loc.x, loc.y);
    const Loc &line_start = seg.l1;
    const Loc &line_end = seg.l2;

    sint4 line_x = line_end.x - line_start.x;
    sint4 line_y = line_end.y - line_start.y;

    sint4 v_x = loc.x - line_start.x;
    sint4 v_y = loc.y - line_start.y;

    sint4 a = line_x * v_x + line_y * v_y;
    sint4 b = line_x * line_x + line_y * line_y;

    Loc proj(line_start.x + line_x * a / b, line_start.y + line_y * a / b);

    Vec_2D line_to_obj(loc.x - proj.x, loc.y - proj.y);
    Vec_2D to_l1(seg.l1.x - proj.x, seg.l1.y - proj.y);
    Vec_2D to_l2(seg.l2.x - proj.x, seg.l2.y - proj.y);

    if (to_l1.dot_prod(to_l2) > 0){ //if both vectors are pointing in the same direction
    //   if(to_l1.length() > to_l2.length()){
// 	return get_point_repulsion(moving_obj, seg.l2);
//       }
//       else{
// 	return get_point_repulsion(moving_obj, seg.l1);
//       }
    }
    else{
      real8 d = loc.distance(proj) - moving_obj->get_radius() +1;
      if (d < moving_obj->get_max_speed() * FIELD_RANGE_MULTIPLIER){
        real8 k = REPULSOR_BASELINE /  (2 * d * d * line_to_obj.length());
	return line_to_obj.multiply(k);
      }
    }
    return Vec_2D(0,0);
  }
#endif

  //-------------------------------------------------------------------
  bool ST_ForceFieldMove::is_at_location(const Object* obj, const Loc& target)
  {
    Loc position = get_center_loc(obj);
    // The 1-turn look-ahead causes collisions when navigating difficult paths 
    // (like inside the mineral cluster); I'm temporarily dropping it until
    // further testing: delays at waypoints are better then unexpected collisions.
    // -- ddeutscher
    //return position.distance(target) <= obj->get_speed();
    return position == target;
  }

  //-------------------------------------------------------------------
  
  bool ST_ForceFieldMove::is_near_location(const Object *obj, const Loc& target, sint4 range){
    Loc position = get_center_loc(obj);
    return position.distance(target) <= range;
  }

  //-------------------------------------------------------------------

  // st_task param is only needed for the hack below...
  bool ST_ForceFieldMove::is_at_goal(const Object* obj, const Task& task, const ST_Terrain::TIter &st_task)
  {
    // Getting to the goal depends on the definition of the task:
    if (task.goal.target == Goal::OBJ ) {
      switch(task.goal.mode) {
      case Goal::TOUCH :
	return( obj->distance_to(*task.goal.obj) <= 1 );
	// using 1-tick look-ahead:
	//return( obj->distance_to(*task.goal_obj) <= obj->get_speed() );
	break;

      case Goal::VICINITY :	
	return (obj->distance_to(*task.goal.obj) <= task.goal.distance);
	break;

      case Goal::ATTACK :
	//return(is_in_range(obj, task.goal.obj));
        assert(0 && "Not implemented yet");
	break;

      default:
	ERR("Unknown mode");
      }
    } 
    else if (task.goal.target == Goal::LOCATION ) {
      switch(task.goal.mode) {
      case Goal::TOUCH :
        // Temporary hack, until final approach is corrected for "LOCATION" 
        // targets: Since the planned path might finish at a location which is only
        // near the requested goal, compare to the planned goal instead of the 
        // requested goal. (When fixed, is_at_goal should be static again).
        if( st_task->isValid() ) {
          TEST( if( get_center_loc(obj) == st_task->path->path[0] )
                  cout << obj << " arrived at location " << get_center_loc(obj) << ", compared to the last location of the planned path" << endl; );
          return get_center_loc(obj) == st_task->path->path[0];
        }
        // using 1-tick look-ahead; also suffers from low-resolution accuracy problems
        TEST( if( is_at_location(obj, task.goal.loc) )
                cout << obj << " arrived at location " << get_center_loc(obj) << ", compared to the last location of the planned path" << endl; );       
	return is_at_location(obj, task.goal.loc);
	break;

      case Goal::VICINITY :
	return is_near_location(obj, task.goal.loc, task.goal.distance);
	break;

      case Goal::ATTACK :
	//return(is_in_range(obj, task.goal.loc));
        assert(0 && "Not implemented yet");
        break;

      default:
	ERR("Unknown mode");
      }
    } else {
      ERR("Unknown target type");
    }
    return false;
  }
  

  //-------------------------------------------------------------------
#if 0 // This needs coupling with GameObj, which we strive to avoid. As I don't
      // need the "ATTACK" tasks for game1, I commented it out. --ddeutscher
  bool ST_Terrain::is_in_range(const Object *moving_obj, const Loc goal_loc){
    //fixme: does not check if unit can see the target. 
    const GameObj *gob = dynamic_cast<const GameObj*>(moving_obj);
    sint4 range = 1;
    if (gob){
      ScriptObj *weapon = gob->component("weapon");
      const sintptr *r = weapon->get_const_int_ptr("max_ground_range");
      if (r){
	range = *r;
      }
    }

    Loc obj_loc = get_center_loc(moving_obj);
   
    return ( (obj_loc.distance(goal_loc) - moving_obj->get_radius()) <= range );


    
    // this is the check in ShooterAI. Other way makes more intuitive sense though...
    //     real8 rdist = moving_obj->distance_to(*goal_obj);
    //     sint4 sdist = (sint4)floor(rdist);
    //     return (sdist <= range);
  }

  //-------------------------------------------------------------------
  
  bool ST_Terrain::is_in_range(const Object *moving_obj, const Object *goal_obj){
    //fixme: does not check if unit can see each other. 
    const GameObj *gob = dynamic_cast<const GameObj*>(moving_obj);
    sint4 range = 1;
    if (gob){
      ScriptObj *weapon = gob->component("weapon");
      const sintptr *r = weapon->get_const_int_ptr("max_ground_range");
      if (r){
	range = *r;
      }
    }
  
    //if not a GameObj or with no weapon range, should behave exactly
    //like the check for touch.
    return ( moving_obj->distance_to(*goal_obj) <= range );
    
    // this is the check in ShooterAI. Other way makes more intuitive sense though...
    //     real8 rdist = moving_obj->distance_to(*goal_obj);
    //     sint4 sdist = (sint4)floor(rdist);
    //     return (sdist <= range);
  }
#endif

  //-----------------------------------------------------------------------------
  void ST_ForceFieldMove::insert_object(const Object *obj)
  {
    objects[obj] = Loc(-1,-1);
  }

  //-----------------------------------------------------------------------------
  void ST_ForceFieldMove::remove_object(const Object *obj)
  {
    objects.erase(obj);
  }

  //-----------------------------------------------------------------------------
  void ST_ForceFieldMove::update_object(const Object * /*obj*/)
  {
  }

  //-----------------------------------------------------------------------------
  void ST_ForceFieldMove::insert_boundary(const Segment& l)
  {
    segments.push_back(l);
    segment_ends[l.l1] = true;
    segment_ends[l.l2] = true;
  }

  //-----------------------------------------------------------------------
  //------------------------Vec_2D-----------------------------------------
  //-----------------------------------------------------------------------
  Vec_2D Vec_2D::operator+(const Vec_2D &other) const{
    return Vec_2D(x + other.x, y + other.y);
  }
 
  Vec_2D Vec_2D::operator-(const Vec_2D &other) const{
    return Vec_2D(x - other.x, y - other.y);
  }

  sint4 Vec_2D::dot_prod(const Vec_2D &other) const{
    return x*other.x + y*other.y;
  }

  real8 Vec_2D::length() const{
    real8 length = sqrt(((real8)x)*x + y*y);
    return length;
  }

  Vec_2D Vec_2D::project_onto(const Vec_2D &other) const{
    real8 k = ((real8)this->dot_prod(other)) / other.dot_prod(other);
    return other.multiply(k);
  }

  Vec_2D Vec_2D::multiply(real8 k) const{
    return Vec_2D(round(k * x), round(k * y));
  }

  bool Vec_2D::operator==(const Vec_2D &other) const{
    return (x == other.x && y == other.y); 
  }
  
  bool Vec_2D::operator!=(const Vec_2D &other) const{
    return !(*this == other); 
  }

} //end of namespace
