#ifndef GobPtrWrapper_H
#define GobPtrWrapper_H

#include "GameObj.H"
#include "Distance.H"

class GameStateModule;

/// Implement some commonly-used shortcut methods
class GobPtrWrapper
{
protected:
  GameObj* gob;
  GameStateModule &gsm;
  
public:
  GobPtrWrapper(GameObj* gob_, GameStateModule &gsm_) : gob(gob_), gsm(gsm_) {};
  virtual ~GobPtrWrapper() {}; // to create a virtual table

  // access the GameObj - not very safe. we are also forced to provide
  // non-const conversions because ORTS kernel doesn't deal well with 'const GameObj*'
  operator GameObj*()  { return gob; };
  GameObj& operator*() { return *gob; };
  operator const GameObj*()  const  { return gob;  };
  const GameObj& operator*() const  { return *gob; };
  // just another explicit syntax, clearer in some cases
  const GameObj* get_GameObj(void) const { return gob; };

  // give a 'pointer-like' syntax
  GobPtrWrapper* operator->()             { return this; };
  const GobPtrWrapper* operator->() const { return this; };

  // expose GameObj functions that we need
  bool is_pending_action() const {
    return gob->is_pending_action(); 
  };

  // accessories - this is the whole point of the drill
  ScalarPoint getPosition() const;
  bool isMoving() const;
  Scalar distance_to(const GobPtrWrapper& other) const;
  Scalar distance_to(Scalar x, Scalar y) const;

  bool operator<(const GobPtrWrapper& other) { return gob < other.gob; };
};

class Worker : public GobPtrWrapper
{
  static const std::string miningComponent;
  static const std::string miningAction;
  static const std::string returnResourcesAction;

#if 0
  mutable sint4 prev_minerals;      ///< The amount of minerals held by this worker, as last seen by "isMining"
  mutable sint4 last_mining_frame;  ///< The last frame in which this worker was seen mining by "isMining"
-- add to ctor: prev_minerals(0), last_mining_frame(-1000)
#endif

public:
  Worker(GameObj* gob_, GameStateModule &gsm_) : GobPtrWrapper(gob_, gsm_) {};

  // give a 'pointer-like' syntax
  Worker* operator->()              { return this; };
  const Worker* operator->() const  { return this; };

  // accessories - this is the whole point of the drill
  void returnMinerals(const GobPtrWrapper& controlCenter);
  void startMining(const GobPtrWrapper& mineral);
  sint4 getMinerals() const;
  bool isMining() const;
};

REGISTER_TYPEOF(100001, std::set<GobPtrWrapper>::iterator);
REGISTER_TYPEOF(100002, std::set<GobPtrWrapper>::const_iterator);
REGISTER_TYPEOF(100003, Vector<GobPtrWrapper>::iterator);
REGISTER_TYPEOF(100004, Vector<GobPtrWrapper>::const_iterator);

REGISTER_TYPEOF(100005, std::set<Worker>::iterator);
REGISTER_TYPEOF(100006, std::set<Worker>::const_iterator);
REGISTER_TYPEOF(100007, Vector<Worker>::iterator);
REGISTER_TYPEOF(100008, Vector<Worker>::const_iterator);

REGISTER_TYPEOF(100009, Vector<GobPtrWrapper*>::iterator);
REGISTER_TYPEOF(100010, Vector<GobPtrWrapper*>::const_iterator);

REGISTER_TYPEOF(100011, Vector<const GobPtrWrapper*>::iterator);
REGISTER_TYPEOF(100012, Vector<const GobPtrWrapper*>::const_iterator);

#endif
