#include "GAIEventHandler.H"
#include "GameStateModule.H"
#include "GfxModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameTile.H"
#include "ServerObjData.H"
#include "Options.H"
#include "GAISquad.H"
#include "GAITank.H"

using namespace std;

enum { TILE_UNKNOWN=0, TILE_WATER=1, TILE_GROUND=2, TILE_CLIFF=3 };

bool GAIEventHandler::handle_event(const Event& e)
{
	if (e.get_who() == GameStateModule::FROM)
	{
		if (e.get_what() == GameStateModule::STOP_MSG) {
			state.quit = true; return true;
		}    
		if (e.get_what() == GameStateModule::READ_ERROR_MSG) {
      			state.quit = state.error = true; return true;
		}
		if (e.get_what() == GameStateModule::VIEW_MSG) {
			compute_actions();
      			state.gsm->send_actions();
      			Game &game = state.gsm->get_game();	

     
      			if (state.gfxm) state.gfxm->process_changes();
        
      			sint4 vf = game.get_view_frame();
      			sint4 af = game.get_action_frame();

      			if ((vf % 20) == 0) cout << "[frame " << vf << "]" << endl;
	
      			if (state.gfxm) {
				// don't draw if we are far behind
				if (abs(vf-af) < 10) {
	  				state.gfxm->draw();
	  				state.just_drew = true;
				}
      			}
			return true;
    		}
  	}
    	return false;
}

static bool firstTime = 1;
static const int tanksPerSquad = 5;
static const int numberOfSquads = 10;
static const int numberBases = 5;
GAISquad* squads[numberOfSquads];
GameObj* myBases[numberBases];
GameObj* oppBases[numberBases];
GameObj* myTanks[numberOfSquads * tanksPerSquad];
GameObj* oppTanks[numberOfSquads * tanksPerSquad];
int numberTanks = numberOfSquads * tanksPerSquad;

void GAIEventHandler::compute_actions()
{
  	const Game &game = state.gsm->get_game();
  	const sint4 cid = game.get_client_player();
    	const Game::ObjCont &objs = game.get_objs(cid);
  
  	if ( firstTime == 1 )
  	{
		// find my & opponent bases
		int index = 0, index_2 = 0;
		int oid = 0;
		if ( cid == 0 ) oid = 1;
		cout << "I'm player " << cid << ". Opponent is player " << oid << "." << endl;
		const Game::ObjCont &oppobjs = game.get_objs(oid);
		FORALL (objs, it) {
			GameObj *gob = (*it)->get_GameObj();
			if ( gob && (gob->bp_name() == "controlCenter") && *gob->sod.x1 > 0 && gob->sod.in_game ) {
				myBases[index] = gob;
				index++;
			}
			if ( gob && (gob->bp_name() == "tank") ) {
				myTanks[index_2] = gob;
				index_2++;
			}
		}
		for ( int i = 0; i < 5; i++ ) {
			cout << " ControlCenter: [ " << *myBases[i]->sod.x1 << " , " << *myBases[i]->sod.y1 << " ] & [ " << *myBases[i]->sod.x2 << " , " << *myBases[i]->sod.y2 			<< " ]" << endl;}
		for ( int i = 0; i < 15; i++ ) {
			cout << " Tank: [ " << *myTanks[i]->sod.x << " , " << *myTanks[i]->sod.y << " ]" << endl;}
		index = 0;
		index_2 = 0;
		FORALL (oppobjs, it) {
			GameObj *gob = (*it)->get_GameObj();
			if ( gob && (gob->bp_name() == "controlCenter") && *gob->sod.x1 > 0 && gob->sod.in_game ) {
				oppBases[index] = gob;
				index++;
			}
			if ( gob && (gob->bp_name() == "tank") ) {
				oppTanks[index_2] = gob;
				index_2++;
			}	
		}
		// make squads
		for ( int i = 0; i < numberOfSquads; i++ ) {
			if ( i < 5 ) {
				squads[i] = new GAISquad(tanksPerSquad, 1);
			}
			else
			{
				squads[i] = new GAISquad(tanksPerSquad, 0);
			}
		}
		int halfBaseSize = (int)(*myBases[0]->sod.x2 - *myBases[0]->sod.x1) / 2;
		// finding which tanks belong to which squad

		baseTanks1 = new GAITank*[numberTanks / numberBases];
		baseTanks2 = new GAITank*[numberTanks / numberBases];
		baseTanks3 = new GAITank*[numberTanks / numberBases];
		baseTanks4 = new GAITank*[numberTanks / numberBases];
		baseTanks5 = new GAITank*[numberTanks / numberBases];
		int in1 = 0, in2 = 0, in3 = 0, in4 = 0, in5 = 0;
		ScalarPoint base1, base2, base3, base4, base5;

		int t = -1;
		FORS (a, numberTanks) {
			GameObj* it = myTanks[a];
			t++;
			int sq = 0;
			int i = -1;
			int distance = 999999999;
			int x = -1, y = -1;
			FORS (b, numberBases) {
				GameObj* its = myBases[b];
				i++;
				int dx = (*its->sod.x1 + halfBaseSize) - *it->sod.x;
				int dy = (*its->sod.y1 + halfBaseSize) - *it->sod.y;
				int new_distance = dx*dx+dy*dy;
				if ( sqrt(new_distance) < distance ) {
					distance = sqrt(new_distance);
					sq = i;
					x = *its->sod.x1 + halfBaseSize;
					y = *its->sod.y1 + halfBaseSize;
				}
			}
			if ( sq == 0 ) {
				baseTanks1[in1] = new GAITank(it, t);
				in1++;
				base1.x = x;
				base1.y = y;
			}
			else if ( sq == 1 ) {
				baseTanks2[in2] = new GAITank(it, t);
				in2++;
				base2.x = x;
				base2.y = y;
			}
			else if ( sq == 2 ) {
				baseTanks3[in3] = new GAITank(it, t);
				in3++;
				base3.x = x;
				base3.y = y;
			}
			else if ( sq == 3 ) {
				baseTanks4[in4] = new GAITank(it, t);
				in4++;
				base4.x = x;
				base4.y = y;
			}
			else // if ( sq == 4 )
			{
				baseTanks5[in5] = new GAITank(it, t);
				in5++;
				base5.x = x;
				base5.y = y;
			}
		}

		// now find out which tank of each base belongs to which squad
		// base1
		int tanks = numberTanks / numberBases;
		int afstanden[tanks];
		ScalarPoint goal = getTileCoor(base1);
		
			goal.x -= 3;
			goal.y -= 2;
			
			for (int i = 0; i<tanks; i++) {
				ScalarPoint T = getTileCoor(baseTanks1[i]->getPosition());
				ScalarPoint O = goal;
				int dx = T.x - O.x;
				int dy = T.y - O.y;
				if ( dy < 0 && dx < 6 )
					afstanden[i] = 4000 - dx;
				else if ( dx == 0 )
					afstanden[i] = dy;
				else {
					if ( dy >= (dx-1) ) afstanden[i] = 1000 + dx;
					else afstanden[i] = 3000 - dy;
				}
			}
			for ( int t = 0; t < tanks; t++ )
			{
				int min = 999999;
				int min_index = 0;
				for ( int i = 0; i < tanks; i++ )
				{
					if ( afstanden[i] < min )
					{
						min_index = i;
						min = afstanden[i];
					}
				}
				afstanden[min_index] 	= afstanden[t];
				afstanden[t]		= 999999;
				GAITank* temptank = baseTanks1[t];
				baseTanks1[t] = baseTanks1[min_index];
				baseTanks1[min_index] = temptank;
			}			
		
		goal = getTileCoor(base2);
		
			goal.x -= 3;
			goal.y -= 2;
			
			for (int i = 0; i<tanks; i++) {
				ScalarPoint T = getTileCoor(baseTanks2[i]->getPosition());
				ScalarPoint O = goal;
				int dx = T.x - O.x;
				int dy = T.y - O.y;
				if ( dy < 0 && dx < 6 )
					afstanden[i] = 4000 - dx;
				else if ( dx == 0 )
					afstanden[i] = dy;
				else {
					if ( dy >= (dx-1) ) afstanden[i] = 1000 + dx;
					else afstanden[i] = 3000 - dy;
				}
			}
			for ( int t = 0; t < tanks; t++ )
			{
				int min = 999999;
				int min_index = 0;
				for ( int i = 0; i < tanks; i++ )
				{
					if ( afstanden[i] < min )
					{
						min_index = i;
						min = afstanden[i];
					}
				}
				afstanden[min_index] 	= afstanden[t];
				afstanden[t]		= 999999;
				GAITank* temptank = baseTanks2[t];
				baseTanks2[t] = baseTanks2[min_index];
				baseTanks2[min_index] = temptank;
			}			

		goal = getTileCoor(base3);
		
			goal.x -= 3;
			goal.y -= 2;
			
			for (int i = 0; i<tanks; i++) {
				ScalarPoint T = getTileCoor(baseTanks3[i]->getPosition());
				ScalarPoint O = goal;
				int dx = T.x - O.x;
				int dy = T.y - O.y;
				if ( dy < 0 && dx < 6 )
					afstanden[i] = 4000 - dx;
				else if ( dx == 0 )
					afstanden[i] = dy;
				else {
					if ( dy >= (dx-1) ) afstanden[i] = 1000 + dx;
					else afstanden[i] = 3000 - dy;
				}
			}
			for ( int t = 0; t < tanks; t++ )
			{
				int min = 999999;
				int min_index = 0;
				for ( int i = 0; i < tanks; i++ )
				{
					if ( afstanden[i] < min )
					{
						min_index = i;
						min = afstanden[i];
					}
				}
				afstanden[min_index] 	= afstanden[t];
				afstanden[t]		= 999999;
				GAITank* temptank = baseTanks3[t];
				baseTanks3[t] = baseTanks3[min_index];
				baseTanks3[min_index] = temptank;
			}			

		goal = getTileCoor(base4);
		
			goal.x -= 3;
			goal.y -= 2;
			
			for (int i = 0; i<tanks; i++) {
				ScalarPoint T = getTileCoor(baseTanks4[i]->getPosition());
				ScalarPoint O = goal;
				int dx = T.x - O.x;
				int dy = T.y - O.y;
				if ( dy < 0 && dx < 6 )
					afstanden[i] = 4000 - dx;
				else if ( dx == 0 )
					afstanden[i] = dy;
				else {
					if ( dy >= (dx-1) ) afstanden[i] = 1000 + dx;
					else afstanden[i] = 3000 - dy;
				}
			}
			for ( int t = 0; t < tanks; t++ )
			{
				int min = 999999;
				int min_index = 0;
				for ( int i = 0; i < tanks; i++ )
				{
					if ( afstanden[i] < min )
					{
						min_index = i;
						min = afstanden[i];
					}
				}
				afstanden[min_index] 	= afstanden[t];
				afstanden[t]		= 999999;
				GAITank* temptank = baseTanks4[t];
				baseTanks4[t] = baseTanks4[min_index];
				baseTanks4[min_index] = temptank;
			}			

		goal = getTileCoor(base5);
		
			goal.x -= 3;
			goal.y -= 2;
			
			for (int i = 0; i<tanks; i++) {
				ScalarPoint T = getTileCoor(baseTanks5[i]->getPosition());
				ScalarPoint O = goal;
				int dx = T.x - O.x;
				int dy = T.y - O.y;
				if ( dy < 0 && dx < 6 )
					afstanden[i] = 4000 - dx;
				else if ( dx == 0 )
					afstanden[i] = dy;
				else {
					if ( dy >= (dx-1) ) afstanden[i] = 1000 + dx;
					else afstanden[i] = 3000 - dy;
				}
			}
			for ( int t = 0; t < tanks; t++ )
			{
				int min = 999999;
				int min_index = 0;
				for ( int i = 0; i < tanks; i++ )
				{
					if ( afstanden[i] < min )
					{
						min_index = i;
						min = afstanden[i];
					}
				}
				afstanden[min_index] 	= afstanden[t];
				afstanden[t]		= 999999;
				GAITank* temptank = baseTanks5[t];
				baseTanks5[t] = baseTanks5[min_index];
				baseTanks5[min_index] = temptank;
			}			
		
		// now add the tanks in their (fixed) order of appearance
		for ( int i = 0; i < tanks; i++ ) {
			if ( i < tanksPerSquad ) {
				squads[0]->addTank(baseTanks1[i]);
				squads[0]->myBaseCenter = base1;
				squads[1]->addTank(baseTanks2[i]);
				squads[1]->myBaseCenter = base2;
				squads[2]->addTank(baseTanks3[i]);
				squads[2]->myBaseCenter = base3;
				squads[3]->addTank(baseTanks4[i]);
				squads[3]->myBaseCenter = base4;
				squads[4]->addTank(baseTanks5[i]);
				squads[4]->myBaseCenter = base5;
			} else {
				squads[5]->addTank(baseTanks1[i]);
				squads[5]->myBaseCenter = base1;
				squads[6]->addTank(baseTanks2[i]);
				squads[6]->myBaseCenter = base2;
				squads[7]->addTank(baseTanks3[i]);
				squads[7]->myBaseCenter = base3;
				squads[8]->addTank(baseTanks4[i]);
				squads[8]->myBaseCenter = base4;
				squads[9]->addTank(baseTanks5[i]);
				squads[9]->myBaseCenter = base5;
			}
		}

	
		firstTime = 0;
  	}

	FORS ( it , numberOfSquads )
	{
		squads[it]->act(game);
	}
}

ScalarPoint GAIEventHandler::getTileCoor(ScalarPoint x)
{
	x.x = x.x / 16;
	x.y = x.y / 16;
	return x;
}
