#include "GfxGlobal.H"
#include "Options.H"
#include "GfxModule.H"
#include "GameStateModule.H"
#include "ShutDown.H"
#include "SDLinit.H"
#include "Game.H"

#include "GAIEventHandler.H"

static char *VERSION = "gai v0.1";

using namespace std;


// loop is called

class Looper : public LoopFunctor
{
public :

  Looper(GAIState *state_) : state(state_) {}
  
  void loop() const {

    if (state->gfxm) {

      if (state->gfxm->get_quit()) { state->quit = true; exit(0); } // should return

      bool refresh;
      Options::get("-refresh", refresh);
      
      if (refresh) {

	// maximal gfx refresh rate
	
	if (!state->just_drew) 
	  state->gfxm->draw();
	else
	  state->just_drew = false;
      }
    }
  
    // looks for server messages
    // if one or more arrived, calls handlers
    // if not, sleeps for one millisecond
    if (!state->gsm->recv_view()) SDL_Delay(1);
  }

private:
  GAIState *state;
};


int main(int argc, char **argv)
{
  signals_shut_down(true);

  Options opt("GAIClient");
  opt.put("-g", "display world in 3d");
  opt.put("-cinfo", "display game state changes");
  opt.put("-tinfo", "display tile info");
  opt.put("-oinfo", "display object info");
  opt.put("-minfo", "display map info");    
  
  // populate opt with options from various modules
  
  GameStateModule::Options::add();
  GfxModule::Options::add();
  Game::Options::add();
  
  // process command line arguments
  if (Options::process(argc, argv, cerr, VERSION)) exit(20);

  // initialize SDL
  SDLinit::video_init();
  SDLinit::network_init();
  
  // populate option objects with command line values
  // and create modules
  
  GameStateModule::Options gsmo;
  GameStateModule gsm(gsmo);
 
  GfxModule::Options gfxmo;
  GfxModule *gfxm = 0;

  // change seed value to time(0) if 0

  sint4 seed;
  Options::get("-seed", seed);

  if (!seed) seed = time(0);

  Options::set("-seed", seed);
  cout << "seed=" << seed << endl;

  bool use_gfx;
  Options::get("-g", use_gfx);

  if (use_gfx) {
    glutInit(&argc, argv);
    gfxm = new GfxModule;
  }

  // gives event handler access to the state
  GAIState state(&gsm, gfxm);

  GAIEventHandler sev(state, seed);

  // gsm events are handled in sev
  gsm.add_handler(&sev);
  
  // connect game state module to server
  gsm.connect();

  Looper l(&state);

  if (use_gfx) {

    // gfx mode

    // was gfxm->init(gsm, gfxmo);
    gfxm->init(gsm.get_game(), gsm.get_changes(), gsm.get_action_changes(), gfxmo);
    // game loop (could become a thread)
    gfxm->start_loop(&l);

  } else {

    // no gfx
    
    while (!state.quit) l.loop();
  }
  
  delete gfxm;
  return 0;
}
