
/** @file Worlded.C
    World representation for map editor.
*/

// This is an ORTS file
// licensed under the GPL

#include "Worlded.H"
//#include "../../../libs/serverclient/src/RandomIsland.C"
Worlded* Worlded::worldedInstance = NULL;
bool     Worlded::initialized = false;
using namespace std;

struct Point{
  sint4 x;
  sint4 y;
};

void Worlded::add_options()
{
  Options o("World Editor");
  
  o.put("-loadmap", std::string(""),      "map file name to load");
  o.put("-newt",    true,            "new terrain");
  o.put("-debug",   false,           "debug mode");
 
// FRANO
  o.put("-frano",false,"frano's test map");

  o.put("-disp",    "2d display");

  o.put("-m",       std::string("server"),"mode: server|client|game|test|worker");
  o.put("-showv",   false,           "show views");
  o.put("-showl",   false,           "show message length");

  o.put("-hfile",   std::string(""),      "create map from height/ramp file");
  o.put("-out",     std::string(""),      "save map to file");

}

Worlded* Worlded::instance(){
  if(worldedInstance == NULL)
    worldedInstance = new Worlded();

  return worldedInstance;
}


void Worlded::initialize(sint4 t_h, sint4 rad, Tile::Type t)
{
  if(!initialized){
  map_data = new Map<Tile>;
  units = new Vector<Unit>;
  radius = rad;
  tile_height = t_h;

  add_options();
  MapTool::Options::add();

  sint4 x, size = 400;
  bool water = true;
  Options::get("-x", x);
  sint4 y;
  Options::get("-y", y);
  Options::set("-island", size);
  Options::set("-water", water);
  sint4 seed;
  Options::get("-seed", seed);
  rand.srand(seed);
  map_data->allocate(x, y);
  tile_type = t;
  initialized = true;
  }
}

bool Worlded::isInitialized() {
  return initialized;
}


//change tiles around x, y to tile type and increment height by diff
void Worlded::change_tiles(sint4 x, sint4 y){
  sint4 width;
  Options::get("-x", width);
  sint4 height;
  Options::get("-y", height); 
  if((x - radius) < 0 || (y < radius) < 0 || (x + radius) >= width || (y + radius) >= height)
    return;
  x-=radius;
  y-=radius;
  radius+=radius;
  FORS(i, radius){
    FORS(j, radius){
      (*map_data)(x + i, y + j).set_type(tile_type);
      (*map_data)(x + i, y + j).set_flat(tile_height);
    }
  }
}

//Places a unit with given name in the world, no safety checking on x and y

void Worlded::place_unit(std::string n, sint4 x, sint4 y) {
  Unit u;
  u.name = n;
  u.x = x;
  u.y = y;
  units->push_back(u);
}

void Worlded::set_tile(Tile::Type t) { tile_type = t; }
Tile::Type Worlded::get_type() { return tile_type; }

void Worlded::set_map(Map<Tile>* m) { map_data = m;}
Map<Tile>* Worlded::get_map() { return map_data; }

void Worlded::set_height(sint4 h) { tile_height = h; }
sint4 Worlded::get_height() { return tile_height; }

void Worlded::set_radius(sint4 r) { radius = r; }
sint4 Worlded::get_radius() { return radius; }

void Worlded::set_units(Vector<Unit>* u) { units = u; }
Vector<Worlded::Unit>* Worlded::get_units() { return units; }
  
void Worlded::generate_map()
{
  // note: if loading a map, -nplayers values must match
  sint4 num_players;
  
  Options::get("-nplayers", num_players);
 
  sint4 size;
  Options::get("-island", size);
  sint4 tiles_x, tiles_y; 
  Options::get("-x", tiles_x);
  Options::get("-y", tiles_y);
  sint4 tile_n = tiles_x * tiles_y;

  
  map_data->allocate(tiles_x, tiles_y);
  sint2 *o_safe = new sint2[tile_n]; // valid square if 0
  if (!o_safe) ERR("out of memory");
  memset(o_safe, 0, tile_n*sizeof(*o_safe));
  real8 convex = 0;
  Options::get("-convex", convex); 
  sint4 r;
  r = rand.rand_uint4();
  //radius of inner disk
  sint4 rad_i = tiles_x /8;
  //positions of players
  sint4 *x_pos = new sint4[num_players];
  sint4 *y_pos = new sint4[num_players];
  bool *flags = new bool[tile_n];
  bool *checked = new bool[tile_n];
  if (!x_pos || !y_pos || !flags || !checked) ERR("out of memory");
  memset(x_pos, 0, num_players*sizeof(*x_pos));
  memset(y_pos, 0, num_players*sizeof(*y_pos));
  memset(flags, 0, tile_n*sizeof(*flags));
  memset(checked, 0, tile_n*sizeof(*checked));

  Vector<Point> *edges  = new Vector<Point>[num_players];
  Vector<Point> *islands  = new Vector<Point>[num_players];
  if (!edges || !islands) ERR("out of memory");
  Options::set("-seed", r);
  //put players on the disk, doesn't write them to world
  MapTool::place_players(rad_i, x_pos, y_pos, (*opt));
  Point t;
  bool water = false;
  bool cliff = false;
  Options::get("-water", water);
  Options::get("-cliff", cliff); 
  FORS (i , tile_n) {
    if(water)
      (*map_data)(i).set_type(WATER);
    else
      (*map_data)(i).set_type(GROUND);
  }
 
  sint4 level_1 = 0;
  sint4 level_2 = 9;
  MapTool m;
  if(convex == 2)
  FORS(p, 1){
    r = -1;
    while( r <= level_1)
      r = rand.rand_uint4() % 13; //height between 0 and max height
    sint4 i_size = r * 20;
    if (i_size < 100) i_size = 100;
    if (i_size > 500) i_size = 500;
    m.raise_island((*map_data), r, 2, i_size, (*opt), rand.rand_uint4());
  }

  // sint4 level_3 = 12;
  sint4 edge = (sint4)sqrt(((double)size));
  sint4 min = edge;
  sint4 x, y;
  FORS (k, num_players) {
    x = x_pos[k] - min;
    y = y_pos[k] - min;
    //make box
    FORS (z, (min + min)) {
      flags[(y * tiles_x) + x + z] = true;
      flags[((y + min + min) * tiles_x) + x + z] = true;
      flags[((y + z) * tiles_x) + x] = true;
      flags[((y + z) * tiles_x) + x + min + min] = true;
    }
    //set box to UNKNOWN tiles
    x = x_pos[k] - min;
    y = y_pos[k] - min;
    //make box
    if (!convex)
    FORS (z, (min + min)) {
      x = x_pos[k] - min;
      FORS (p, (min + min)) {
	map_data->get_tile(x, y).set_type(Tile::UNKNOWN);
	x++;
      }
      y++;
    }
    //put 8 X 8 island
    x = x_pos[k] - 4;
    y = y_pos[k] - 4;
    sint4 left_edge = x;
    sint4 right_edge = x + 8;
    sint4 bottom = y + 8;
    sint4 top = y;

    for (;y <= bottom; ++y) {
      for (x = x_pos[k] - 4; x <= right_edge; ++x) {

        Tile &tile = map_data->get_tile(x, y);
	tile.set_flat(level_2);
	tile.set_type(GROUND);
	flags[(y * tiles_x) + x] = true;
	if(x == left_edge){
	  t.x = x - 1;
	  t.y = y;
	  edges[k].push_back(t);
	  flags[(tiles_x * t.y) + t.x] = true;
	}
	if(x == right_edge){
	  t.x = x + 1;
	  t.y = y;
	  edges[k].push_back(t);
	  flags[(tiles_x * t.y) + t.x] = true;
	}
	if(y == top){
	  t.y = y - 1;
	  t.x = x;
	  edges[k].push_back(t);
	  flags[(tiles_x * t.y) + t.x] = true;
	}
	if(y == bottom){
	  t.y = y + 1;
	  t.x = x;
	  edges[k].push_back(t);
	  flags[(tiles_x * t.y) + t.x] = true;
	}
      }
    }

  }
     
  FORS (i, size - 8*8) {
    FORS (j, num_players) {
      r = rand.rand_uint4();
      r = abs(r);
      MapTool::add_tile(r, edges[j], (*map_data), flags, tiles_x, tiles_y, level_2);
    }
  }

  FORS (k, num_players) {
    x = x_pos[k] - min;
    y = y_pos[k] - min;
    sint4 left_edge = x;
    sint4 right_edge = x_pos[k] + min;
    sint4 bottom = y_pos[k] + min;
    sint4 top = y;
    //flood
    if(!convex){
    while(y <= bottom){
      while(x <= right_edge && map_data->get_tile(x, y).get_height(0) != level_2){
	if(water)
          (*map_data)(x, y).set_type(WATER);
	else
	  (*map_data)(x, y).set_type(GROUND);
        x++;
      }
      x = right_edge - 1;
      while(x >= left_edge && (*map_data)(x, y).get_height(0) != level_2){
        if(water)
          (*map_data)(x, y).set_type(WATER);
	else
	  (*map_data)(x, y).set_type(GROUND); 
        x--;
      }
      y++;
      x = left_edge;
    }
    
    x = left_edge;
    y = top;
    while(x <= right_edge){
      while(y <= bottom && (*map_data)(x, y).get_height(0) != level_2){
	if(water)
          (*map_data)(x, y).set_type(WATER);
	else
	  (*map_data)(x, y).set_type(GROUND);
        y++;
        }
      y = bottom - 1;
      while(y >= top && (*map_data)(x, y).get_height(0) != level_2){
	if(water)
          (*map_data)(x, y).set_type(WATER);
	else
	  (*map_data)(x, y).set_type(GROUND);
        y--;
      }
      x++;
      y = top;
    }
   
   
    }

    // set holes to ground tiles
        y = y_pos[k] - min;
    //make box
    if(!convex)
    FORS(z, (min + min)){
      x = x_pos[k] - min;
      FORS(p, (min + min)){
	if((*map_data)(x, y).get_typeW() == Tile::UNKNOWN){
	  (*map_data)(x, y).set_type(GROUND);
	  (*map_data)(x, y).set_flat(level_2);
	}
	x++;
      }
      y++;
    }  
  
  }
  //convex hull algo
  if(convex)
  FORS(k, num_players){
    std::stack<Point> all;
    std::stack<Point> u_edge;
    std::stack<Point> l_edge;
    x = x_pos[k] + min - 1;
    y = y_pos[k] + min - 1;
    sint4 left_edge = x_pos[k] - min;
    sint4 top = y_pos[k] - min;
    //get island tiles
    while(x > left_edge){
      y = y_pos[k] + min - 1;
      while(y > top){
	if((*map_data)(x, y).get_height(0) == level_2){
	  t.x = x;
	  t.y = y;
	  all.push(t);

	}
	y--;
      }
      x--;
    }
    MapTool::upper_edge(all, u_edge, convex);
    MapTool::lower_edge(all, l_edge, convex);
    
    // set points in convex hull to island ground tiles
    x = x_pos[k] + min - 1;
    y = y_pos[k] + min - 1;
    Vector<Point> poly;
    while(!u_edge.empty()){
      poly.push_back(u_edge.top());
      u_edge.pop();
    }
    while(!l_edge.empty()){
      poly.push_back(l_edge.top());
      l_edge.pop();
    }

    while(x > left_edge){
      y = y_pos[k] + min - 1;
      while(y > top){
	  t.x = x;
	  t.y = y;	
	if(MapTool::inside_polygon(poly, t)){
	  (*map_data)(x, y).set_type(GROUND);
	  (*map_data)(x, y).set_flat(level_2);
	}
	y--;
      }
      x--;
    }
   
 
  }

  y = 1;
  while (y < tiles_y - 1){
    x = 1;
    while(x < tiles_x - 1){
      MapTool::smooth_edge(x, y, (*map_data), (*opt), level_2 - level_1);
      x++;
    }
    y++;
  }
  

 
      
  MapTool::sync_corner((*map_data), tiles_x, tiles_y);
  if (cliff) MapTool::cliff_tiles((*map_data), tiles_x, tiles_y);

  FORS(k, num_players){
    bool ramp = false;
    x = x_pos[k] - min;
    y = y_pos[k] - min;
    while (!ramp) {
      ramp = MapTool::place_ramp((*map_data), x, y,
                                 (min + min) * 5 / 6,
                                 rand.rand_uint4() % 8,
                                 level_1, level_2, tiles_x, tiles_y);
    }
  }

  MapTool::sync_corner((*map_data), tiles_x, tiles_y);
  sint4 nobjs;
  Options::get("-nobjs", nobjs); 
  // create random clusters of minerals
  Unit u;

#if 1
  Point point;
  FORS(i, num_players){
    FORS(t, nobjs / num_players / 3){
      point = MapTool::place_object(x_pos[i], y_pos[i], 3, edge/3, rand);
      u.name =  "mineral";
      u.x = point.x;
      u.y = point.y;
    }
  }
#endif

  FORS (p, num_players) {
    u.name = "worker";
    u.x = x_pos[p];
    u.y = y_pos[p];
  }
  set_map(map_data); 
  delete [] o_safe;
}
  

/*
  //returns worst ratio of n player scores
  real4 get_worst(int* player_score, int n){
  }
*/

