/** @file ScriptModel.C
    @see ScriptModel.H

    $Id: ScriptModel.C,v 1.21 2006/12/17 22:20:48 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/gfxclient/src/ScriptModel.C,v $
*/

// This is an ORTS file (c) Tim Furtak

#include "Global.H"
#include "ScriptModel.H"
#include "GfxModule.H"
#include "GfxAnimesh.H"
#include "Game.H"
#include "Profiler.H"

using namespace std;

GfxModule* ScriptModel::gfxm;
const sint4 ScriptModel::STYPE_GFXANIMESH_ID = stype::next_id();

//===================================================================

ScriptModel::ScriptModel()
{
  add("load",           &ScriptModel::load_model,     0, 4, 0);
  add("set_shadow",     &ScriptModel::set_shadow,     0, 1, 0);
  add("enable_shadow",  &ScriptModel::enable_shadow,  0, 1, 0);

  add("position",       &ScriptModel::position,       0, 3, 0);
  add("rotate",         &ScriptModel::rotate,         0, 1, 0);
  add("tilt",           &ScriptModel::tilt,           0, 1, 0);
  add("scale",          &ScriptModel::scale,          0, 1, 0);

  add("draw_2d",        &ScriptModel::draw_2d,        0, 3, 0);
  add("draw_overlay",   &ScriptModel::draw_overlay,   0, 5, 0);
  add("idle_anim",      &ScriptModel::idle_animation, 0,-1, 0);
  add("play_anim",      &ScriptModel::play_animation, 0,-1, 0);
  add("set_frame",      &ScriptModel::set_frame,      0, 2, 0);
  add("play_custom",    &ScriptModel::play_custom,    0,-1, 0);

  add("set_parent",     &ScriptModel::set_parent,     1,-1, 0);
  
  mesh = 0;
}

ScriptModel::~ScriptModel()
{
}

//-------------------------------------------------------------------

void ScriptModel::add(const std::string &name, myfunc f, sint4 objs, sint4 p, sint4 p2)
{
  ObjActionDefinition def(name, objs, p, p2);
  ObjAction *a = new ObjAction(static_cast<ObjFunction>(f), def); // fixme: nonportable cast
  actions.push_back(a);
  act_map[name] = a;
}

//===================================================================

stype ScriptModel::load_model(const SActParam &p)
{
  PROFILE("ScriptModel::load_model");
  VALN_CHECK(4);

  if (mesh) {
    delete mesh;
    mesh = 0;
  }

  GameObj *gob = this->parent->get_GameObj();
  SCRIPT_ASSERT(gob && "GameObj ptr 0");
  
  string path = GUI_gfxm->opt.gfx_path;
  //  string path = ObjAction::get_pwd();

  if (path.size() && path[path.size()-1] != PATH_SEP) path = path + PATH_SEP;

  string mpath = sanitize_path(path + "/" + p.val[0].to_cstr());
  string tpath = sanitize_path(path + "/" + p.val[1].to_cstr());

  GfxAnimesh *m = new GfxAnimesh(*GUI_gfxm, mpath, tpath, gob->get_int("owner"),
                                 p.val[2].to_int()/100.0, p.val[3].to_int()/100.0);

  GfxGameObject *ggo = GUI_gfxm->game2gfx[gob];
  if (!ggo->gfx_obj) {
    ggo->attach_gfxobj(m);
    GUI_gfxm->add_GfxObject(m);
  }

  mesh = m;

  return stype();
}

//-------------------------------------------------------------------

stype ScriptModel::set_shadow(const SActParam &p)
{
  //  assert(p.val.size() == 1 || p.val.size() == 2);
  VALN_CHECK(1);
  
  real4 scale = 0.0;
  if (p.val.size() >= 2) scale = p.val[1].to_int() / 100.0;

  SCRIPT_ASSERT(mesh && "mesh ptr 0");
  mesh->set_shadow_tex(p.val[0].to_int(), scale);

  return stype();
}

//-------------------------------------------------------------------

stype ScriptModel::enable_shadow(const SActParam &p)
{
  VALN_CHECK(1);
  SCRIPT_ASSERT(mesh && "mesh ptr 0");
  mesh->enable_shadow(p.val[0].to_int());
  return stype();
}

//-------------------------------------------------------------------

stype ScriptModel::position(const SActParam &p)
{
  VALN_CHECK(3);
  SCRIPT_ASSERT(mesh && "mesh ptr 0");
  
  real4 x = p.val[0].to_int()/100.0;
  real4 y = p.val[1].to_int()/100.0;
  real4 z = p.val[2].to_int()/100.0;

  mesh->set_position(x, y, z);
  return stype();
}

//-------------------------------------------------------------------

stype ScriptModel::rotate(const SActParam &p)
{
  VALN_CHECK(1);
  SCRIPT_ASSERT(mesh && "mesh ptr 0");
  real4 y = p.val[0].to_int();
  mesh->set_rotation(y);
  return stype();
}

//-------------------------------------------------------------------

stype ScriptModel::tilt(const SActParam &p)
{
  VALN_CHECK(1);
  SCRIPT_ASSERT(mesh && "mesh ptr 0");
  real4 y = p.val[0].to_int();
  mesh->set_pitch(y);
  return stype();
}
//-------------------------------------------------------------------

stype ScriptModel::scale(const SActParam &p)
{
  VALN_CHECK(1);
  SCRIPT_ASSERT(mesh && "mesh ptr 0");
  real4 s = p.val[0].to_int();
  mesh->set_scale(s);
  return stype();
}

//-------------------------------------------------------------------

stype ScriptModel::draw_2d(const SActParam &p)
{
  VALN_CHECK(3);
  SCRIPT_ASSERT(mesh && "mesh ptr 0");

  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
  glEnable(GL_COLOR_MATERIAL);

  GfxAnimesh::beginDraw();

  glClear(GL_DEPTH_BUFFER_BIT);
  GfxObject *gob = mesh;
  sint4 x0 = p.val[0].to_int();
  sint4 y0 = p.val[1].to_int();
  // negate object scale (called in draw())
  // normalize size by dividing by radius
  real4 s = 0.6 * p.val[2].to_int() / gob->pose.scale / gob->radius;
  
  glPushMatrix();

  glTranslatef(x0, y0, 0);

  // match perspective of the active view
  glRotatef(GUI_gfxm->views[GUI_gfxm->active_view].x_angle, 1.0, 0.0, 0.0);
  glRotatef(GUI_gfxm->views[GUI_gfxm->active_view].y_angle, 0.0, 1.0, 0.0);

  glScalef(s,s,s);

  glTranslatef(-gob->pose.center.x, -gob->pose.center.y, -gob->pose.center.z);
  real4 foo = gob->pose.scale;
  glRotatef(gob->pose.y_rot,0,1,0);
  glTranslatef(-gob->centroid.x * foo, -gob->centroid.y * foo, -gob->centroid.z * foo);
  glRotatef(-gob->pose.y_rot,0,1,0);
  gob->draw();
  glPopMatrix();

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); // particle effects could have changed this

  //  GfxAnimesh::endDraw();

  glDisable(GL_LIGHTING);
  glCullFace(GL_BACK);

  glDisable(GL_COLOR_MATERIAL);
  glDisable(GL_CULL_FACE);
  glDisable(GL_DEPTH_TEST);

  return stype();
}

//-------------------------------------------------------------------

stype ScriptModel::draw_overlay(const SActParam &p)
{
  VALN_CHECK(5);
  SCRIPT_ASSERT(mesh && "mesh ptr 0");
  
  real4 x1 = p.val[1].to_int();
  real4 y1 = p.val[2].to_int();
  real4 x2 = p.val[3].to_int();
  real4 y2 = p.val[4].to_int();

  glBindTexture(GL_TEXTURE_2D, p.val[0].to_int());

  GfxObject *gob = mesh;

  glPushMatrix();
  glTranslatef(gob->pose.center.x, gob->pose.center.y, gob->pose.center.z);

  GUI_gfxm->overlay_mode(); // remove rotation

  glBegin(GL_QUADS);

  //  glEnable(GL_POLYGON_SMOOTH);
  glTexCoord2f(0, 1); glVertex3f(x1, y1, 0);
  glTexCoord2f(1, 1); glVertex3f(x2, y1, 0);
  glTexCoord2f(1, 0); glVertex3f(x2, y2, 0);
  glTexCoord2f(0, 0); glVertex3f(x1, y2, 0);

  // hack to draw over the crease from the smoothed quad
  /*
  glTexCoord2f(0, 0); glVertex3f(x1, y2, 0);
  glTexCoord2f(0, 1); glVertex3f(x1, y1, 0);
  glTexCoord2f(1, 1); glVertex3f(x2, y1, 0);
  glTexCoord2f(1, 0); glVertex3f(x2, y2, 0);
  */
  //  glDisable(GL_POLYGON_SMOOTH);

  glEnd();

  glPopMatrix();
//  overlay_mode_end(); // pops the modelview

  return stype(1);
}

//-------------------------------------------------------------------

stype ScriptModel::idle_animation(const SActParam &p)
{
  SCRIPT_ASSERT(mesh && "no mesh");
  sint4 s = p.val.size();

  if (s < 3 || s > 4) {
    SCRIPT_ERR("idle_animation parameter size mismatch");
  }

  string name = p.val[0].to_cstr();
  real4 fps = (real4)(p.val[1].to_int());
  FrameSet::PLAY_TYPE pt = static_cast<FrameSet::PLAY_TYPE>(p.val[2].to_int());

  if (s == 3) mesh->set_idle(name, fps, pt);
  if (s == 4) mesh->set_idle(name, fps, pt, p.val[3].to_int()); // looping frames
  return stype();
}

//-------------------------------------------------------------------

stype ScriptModel::play_animation(const SActParam &p)
{
  SCRIPT_ASSERT(mesh && "no mesh");
  sint4 s = p.val.size();

  if (s < 3 || s > 4) {
    SCRIPT_ERR("play_animation parameter size mismatch");
  }

  string name = p.val[0].to_cstr();
  real4 fps = (real4)p.val[1].to_int();
  FrameSet::PLAY_TYPE pt = static_cast<FrameSet::PLAY_TYPE>(p.val[2].to_int());

  if (s == 3) mesh->play(name, fps, pt);
  if (s == 4) mesh->play(name, fps, pt, p.val[3].to_int()); // looping frames
  return stype();
}

//-------------------------------------------------------------------

stype ScriptModel::play_custom(const SActParam &p)
{
  SCRIPT_ASSERT(mesh && "no mesh");
  VALGEN_CHECK(3);

  sint4 s = p.val.size();
  FrameSet::PLAY_TYPE pt = static_cast<FrameSet::PLAY_TYPE>(p.val[s-1].to_int());
  real4 fps = (real4)p.val[s-2].to_int();

  Vector<sint4> frames;
  FORU (i, s-2) frames.push_back(p.val[i].to_int());

  mesh->play_custom(frames, fps, pt);
  return stype();
}

//-------------------------------------------------------------------

stype ScriptModel::set_frame(const SActParam &p)
{
  SCRIPT_ASSERT(mesh && "no mesh");
  VALGEN_CHECK(2);
  mesh->set_frame(p.val[0].to_cstr(), p.val[1].to_int());
  return stype();
}

//-------------------------------------------------------------------

stype ScriptModel::set_parent(const SActParam &p)
{
  OBJN_CHECK(1);
  SCRIPT_ASSERT(mesh && "no mesh");

  ScriptModel *s = dynamic_cast<ScriptModel*>((ScriptObj*)p.obj[0]);
  SCRIPT_ASSERT(s && "no ScriptModel");
  
  GfxAnimesh *o = s->mesh;

  if (p.val.size() == 1) {
    string tag_name = p.val[0].to_cstr();
    o->attachGfxObject_tag(mesh, tag_name);
  } else {
    o->attachGfxObject(mesh);
  }
  mesh->match_id(o->id());
  return stype_from_GfxMesh(mesh);
}

//===================================================================

ObjAction* ScriptModel::action(const std::string &name) const
{
  FIND (act_map, it, name);
  if (it != act_map.end()) return it->second;
  return 0;
}

//===================================================================

void ScriptModel::destruct()
{
  if (mesh) mesh->kill();
  mesh = 0;
}

//===================================================================
