/** @file Options.C
    @see Options.H.

    $Id: Options.C,v 1.10 2006/04/03 00:41:32 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/Options.C,v $
*/

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Options.H"

using namespace std;

Options::~Options() {
  //  FORALL (*omap, i) {
  //    if (i->second) { delete i->second; i->second = 0; }
  //  }
}

template <> string Options::Type<bool>::name  = "bool";
template <> string Options::Type<sint4>::name = "sint4";
template <> string Options::Type<real8>::name = "real8";
template <> string Options::Type<string>::name = "string";
template <> string Options::Type< Vector<string> >::name = "svector";

Vector<string> Options::heading_name;
Vector<Options::OMap> Options::options;
Options::OMap Options::s_omap;

//-------------------------------------------------------------------

Options::Options(string heading)
{
  sint4 s = options.size();
  options.resize(s+1);
  omap = &options[s];
  heading_name.push_back(heading);
}

//-------------------------------------------------------------------

template <> bool OptT<bool>::from_string(const std::string &s)
{
  if (s.empty()) return true;
  if      (s[0] == '0') value = false;
  else if (s[0] == '1') value = true;
  else return true;
  return false;
}

template <> string OptT<bool>::to_string() const
{
  stringstream s;
  if (param)
    s << "[bool/" << value << "]";
  return s.str();
}

template <> string OptT<bool>::get_type_name() const { return "bool"; }

//-------------------------------------------------------------------

template <> bool OptT<sint4>::from_string(const std::string &s)
{
  int v;
  sint4 r = sscanf(s.c_str(), "%d", &v);
  value = v;
  assert(sizeof(int) >= 4);
  return r == 0 || r == EOF;
}

template <> string OptT<sint4>::to_string() const
{
  stringstream s;
  s << "[sint4/" << value << "]";
  return s.str();
}

template <> string OptT<sint4>::get_type_name() const { return "sint4"; }

//-------------------------------------------------------------------

template <> bool OptT<real8>::from_string(const std::string &s)
{
  sint4 r = sscanf(s.c_str(), "%lf", &value);
  return r == 0 || r == EOF;
}

template <> string OptT<real8>::to_string() const
{
  stringstream s;
  s << "[real8/" << value << "]";
  return s.str();
}

template <> string OptT<real8>::get_type_name() const { return "real8"; }

//-------------------------------------------------------------------

template <> bool OptT<string>::from_string(const std::string &s)
{
  value = s;
  return false;
}

template <> string OptT<string>::to_string() const
{
  stringstream s;
  s << "[string/\"" << value << "\"]";
  return s.str();
}

template <> string OptT<string>::get_type_name() const { return "string"; }

//-------------------------------------------------------------------

template <> bool OptT< Vector<string> >::from_string(const std::string &s)
{
  value.push_back(s);
  return false;
}

template <> string OptT< Vector<string> >::to_string() const
{
  stringstream s;
  s << "[DEFINE/\"";
  FORS (i, value.size()-1) {
    s << value[i] << ",";
  }
  if (!value.empty())
    s << value[value.size()-1];
  s << "\"]";
  return s.str();
}

template <> string OptT< Vector<string> >::get_type_name() const { return "define list"; }

//-------------------------------------------------------------------

bool Options::process(sint4 argc, char **argv, ostream &oserr, const char *remark)
{
  for (sint4 i=1; i < argc; i += 2) {

    string key = argv[i];

    OMapIt oit = s_omap.find(key);

    if (oit == s_omap.end()) {
      if (remark) oserr << remark << endl;
      oserr << "options: unknown option " << key << endl;
      oserr << "choices are:" << endl;
      FORALL (heading_name, i) {
        oserr << *i << ":" << endl;
        FORALL (options[i-heading_name.begin()], it) {
          form(oserr, "%14s : %s ", (*it).first.c_str(), (*it).second->remark.c_str());
          oserr << (*it).second->to_string() << endl;
        }
        oserr << endl;
      }
      return true;
    }

    Opt *opt = oit->second;
    if (!opt->param) {
      opt->from_string("1");
      --i;
      continue;
    }

    if (i == argc-1) { oserr << "options: missing value" << endl; return true; }

    string val = argv[i+1];

    if (opt->from_string(val)) {
      oserr << "options: value syntax error for option "
            << key << " (" << val << ")" << endl;
      return true;
    }
  }

  return false;
}

bool Options::process(const string &args, ostream &oserr, const char *remark)
{
  stringstream ss(args);
  string key;
  
  while (ss >> key) {

    OMapIt oit = s_omap.find(key);

    if (oit == s_omap.end()) {
      if (remark) oserr << remark << endl;
      oserr << "options: unknown option " << key << endl;
      oserr << "choices are:" << endl;
      FORALL (heading_name, i) {
        oserr << *i << ":" << endl;
        FORALL (options[i-heading_name.begin()], it) {
          form(oserr, "%14s : %s ", (*it).first.c_str(), (*it).second->remark.c_str());
          oserr << (*it).second->to_string() << endl;
        }
      }
      return true;
    }

    Opt *opt = oit->second;
    if (!opt->param) {
      opt->from_string("1");
      continue;
    }

    if (ss.eof()) { oserr << "options: missing value" << endl; return true; }

    string val;
    ss >> val;

    if (opt->from_string(val)) {
      oserr << "options: value syntax error for option "
            << key << " (" << val << ")" << endl;
      return true;
    }
  }

  return false;
}



// Local Variables:
// c-file-style:"gnu"
// c-basic-offset:2
// indent-tabs-mode:nil
// tab-width:4
// End:
