// $Id: OSL_Parser.H,v 1.2 2004/10/29 16:58:40 orts_jyip Exp $

// (c) Jonathan Yip, licensed under the GPL

#ifndef OSL_PARSER_H
#define OSL_PARSER_H

#include "Global.H"
#include "OSL.H"
#include "Lexer.H"

namespace OSL {

  class Parser {
  public:
    Parser(Program &p, const std::string &file_name);

    void parse();

  private:
    Program& prog;
    RecStream rs;
    Lexer* lex;

    Class* curr_class;
    FuncMember* curr_func;
    bool has_return;
    bool curr_static;

    //    enum Modifier { MOD_PRIVATE, MOD_PROTECTED, MOD_PUBLIC };
    Modifier curr_mod;

    //    bool attributes;
    //    vector<string> attributes;

    enum VarType { VAR_UNKNOWN, VAR_LOCAL, VAR_PARAM, VAR_INST_MEMBER,
		   VAR_STATIC_MEMBER };
    typedef std::pair<VarType, uint4> var_location;
    typedef std::pair<std::string, uint4> var_info;
    uint4 curr_local;
    std::vector<var_info> local_vars;
    std::stack<uint4> local_scopes;
    std::vector<var_info> param_vars;
    std::vector<var_info> member_vars;

    std::stack<uint4> break_addrs;
    std::stack<uint4> cont_addrs;

    struct TokenInfo {
      Token tk;
      uint4 line_num;
      uint4 pos_num; // fixme: need to add
      std::string file;
      
      TokenInfo(Token tk_, uint4 line_num_,
		uint4 pos_num_, std::string file_) :
	tk(tk_), line_num(line_num_), pos_num(pos_num_), file(file_) {}
    };

    std::vector<TokenInfo> token_stream;
    uint4 token_stream_ptr;

    const Token get_token();
    const Token peek_token();
    const Token get_token_info();
    const Token peek_token_info();

    var_location find_variable(const std::string var_name);
    void add_block();
    void del_block();

    void parse_declarations();
    void parse_definitions();

    void parse_func_headers();

    void parse_var_or_array(const Token &type, const bool is_static);
    //    void parse_static();
    void parse_var_dec(const Token &type, const bool is_static);
    void parse_array_dec(const Token &type, const bool is_static);
    void parse_void_func(const bool is_static);
    void parse_func_dec(const Token &return_type, 
			const std::string &func_name,
			const bool is_static,
			const bool is_constructor);
    void parse_func_dec2(const Token &return_type, 
			 const std::string &func_name,
			 const bool is_static,
			 const bool is_constructor);
    void parse_func_def();
    
    
    void parse_class_dec();
    void parse_class_def();
    void parse_constructor(const Token &type);

    void parse_stmt_list(const TokenType &expected_end_marker);
    void parse_stmt();

    void parse_local_array(const TokenType &type);
    void parse_local_var(const TokenType &type);    
    void parse_local_obj(const std::string &class_name);

    TypeIndex parse_func_call(const std::string &func_name, const bool is_stmt,
			 const sint4 class_index);

    void parse_extf_call(const uint4 func_index, const bool is_stmt);

    void parse_assign(const std::string &var_name);
    void parse_if();
    void parse_while();
    void parse_for();
    void parse_switch();
    void parse_break();
    void parse_cont();
    void parse_return();

    void parse_exp(const TokenType &expected_end_marker,
		   const TypeIndex &expected_type);
    void parse_till_low_op(sint4 prec, const TypeIndex &expected_type);
    void get_term(TypeIndex &type);
    
    void parse_class_var(uint4 loc, TypeIndex &expected_type);

    sint4 get_precedence_of(const sint4 &op);

    bool valid_type(const TokenType &t);
    bool valid_sint1(const std::string &s);
    bool valid_uint1(const std::string &s);
    bool valid_sint2(const std::string &s);
    bool valid_uint2(const std::string &s);
    bool valid_sint4(const std::string &s);
    bool valid_uint4(const std::string &s);
    bool valid_sint8(const std::string &s);
    bool valid_uint8(const std::string &s);
  };


}

#endif
