// $Id: ForceFieldPE.C,v 1.1 2006/12/31 04:09:40 orts_furtak Exp $
// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "ForceFieldPE.H"
#include "GameObj.H"

//===============================================================

ForceFieldPE::ForceFieldPE(){

}

//---------------------------------------------------------------
void ForceFieldPE::add_obj(const Object *obj){
  ObstacleData od;
  od.team = get_team(obj);
  obj2obstacle[obj] = od;
  obj2obstacle[obj].em = create_emitter(obj, REPULSE_STRENGTH); //done here so aas to prevent deletion of em when od destructor called
}

//---------------------------------------------------------------
void ForceFieldPE::remove_obj(const Object* obj){
  obj2obstacle.erase(obj);
}

//---------------------------------------------------------------
void ForceFieldPE::update_obj(const Object* obj){
  //do nothing
  obj2obstacle[obj].team = get_team(obj); //fixme: does TerrainModule send an update if the team of an object changed?
  delete obj2obstacle[obj].em;
  obj2obstacle[obj].em = create_emitter(obj, REPULSE_STRENGTH);
}
//---------------------------------------------------------------
void ForceFieldPE::add_segments(const Vector<TerrainBase::Segment>& segs){
  FORALL(segs, it){
    wallForces.add_emitter(new VecField::OrthoLineEmitter(it->l1.x, it->l1.y, it->l2.x, it->l2.y, 0, 16, WALL_REPULSE_STRENGTH)); 
  }
}
//---------------------------------------------------------------
void ForceFieldPE::execute_tasks(std::list<PathfindTask> &tasks, Vector<TerrainBase::MoveCmd> &cmds, Vector<TerrainBase::StatusMsg> &msgs){
  
  attractors.clear();
  // VecField forces;

//   FORALL(obj2obstacle, it){
//     it->second.em = create_emitter(it->first, REPULSE_STRENGTH);
//     if (it->second.em)
//       forces.add_emitter(it->second.em);
//   }

//   FORALL(walls, it){
//     Segment &s = *it;
//     forces.add_emitter(new VecField::OrthoLineEmitter(s.l1.x, s.l1.y, s.l2.x, s.l2.y, 0, 16,1.0));
//   }

  FORALL(tasks, task_it){
    assert(!task_it->task.objs.empty());
   
    //Attach a TaskData object tasks we haven't seen before
    if (!task_it->executor_data){
      task_it->executor_data = new FFTaskData();
    }

    //fixme: does not work for tasks with more than one object
    const Object* obj = *(task_it->task.objs.begin());
    
    //if no path is found, add a message for that task
    if (task_it->noPathFound ) {
      TerrainBase::StatusMsg s;
      // fixme: this is not robust to future changes in the meaning of TaskId.
      s.task_id = (TerrainBase::TaskId)&(*task_it);
      s.obj = obj;
      s.type = TerrainBase::StatusMsg::NO_PATH_FAILURE;
      msgs.push_back(s);
      continue;
    }
    //skip tasks that are pending
    if (task_it->isPending) continue;
    
    Loc obj_center = get_center_loc(obj);

    bool finished_path = (task_it->current_wp == (static_cast<sint4>(task_it->path.locs.size())-1)
			  && obj_center.distance(task_it->path.locs.at(task_it->current_wp)) <= 1);

    //check to see if object is at the goal,and add a message if it is
    if (finished_path || is_at_goal(obj, task_it->task)){
      TerrainBase::StatusMsg s;
      // fixme: this is not robust to future changes in the meaning of TaskId.
      s.task_id = (TerrainBase::TaskId)&(*task_it);
      s.obj = obj;
      s.type = TerrainBase::StatusMsg::ARRIVED;
      msgs.push_back(s);
      add_stop_command(obj, cmds);      
      std::cerr << "Path Success\n";
      continue;
    }

    //Check progress alone path, and pass a Path Failure message if required.
    FFTaskData* pData = dynamic_cast<FFTaskData*>(task_it->executor_data);
    assert(pData);
    
    Loc current_loc = get_center_loc(obj);
    
    if (pData->prev_loc == Loc(-1, -1) || current_loc.distance(task_it->path.locs[task_it->current_wp]) <
	pData->prev_loc.distance(task_it->path.locs[task_it->current_wp])){
      pData->progress_count = 0;
      pData->prev_loc = current_loc;
    }
    else{
      pData->progress_count++;
    }
    
   

    if (pData->progress_count >= PROGRESS_CUTOFF){
      TerrainBase::StatusMsg s;
      // fixme: this is not robust to future changes in the meaning of TaskId.
      s.task_id = (TerrainBase::TaskId)&(*task_it);
      s.obj = obj;
      s.type = TerrainBase::StatusMsg::MOVEMENT_FAILURE;
      msgs.push_back(s);
      add_stop_command(obj, cmds);
      continue;
    }
    
    //If not at goal, then calculate the next move:
    calculate_move(*task_it, cmds);
    
  }

//   FORALL(obj2obstacle, it){
//     it->second.em = 0;
    
//   }
}
//---------------------------------------------------------------

void ForceFieldPE::calculate_move(PathfindTask &pt, Vector<TerrainBase::MoveCmd> &cmds ){
  const Object* obj = *(pt.task.objs.begin());
  Loc obj_center = get_center_loc(obj);
    
  //calculate_attractor:
  Vec2<real4> attractor_loc = get_attractor(pt);
  
  attractors.push_back(attractor_loc);
  
  VecField::PointEmitter attractor = VecField::PointEmitter(attractor_loc.x, attractor_loc.y, 100000, ATTRACT_STRENGTH); //radius set to large number... better to write a new Emitter?
  
  //forces.add_emitter(attractor);
  
  Vec2<real4> d(attractor_loc.x - obj_center.x, attractor_loc.y - obj_center.y);

  Vec2<real4> forces = attractor.field(obj_center.x, obj_center.y, d.x, d.y);
  FORALL (obj2obstacle, it){
    if (it->first == obj) continue;
    if (pt.task.goal.target == TerrainBase::Goal::OBJ && pt.task.goal.obj == it->first) continue;
    forces += it->second.em->field(obj_center.x, obj_center.y, d.x, d.y);
  }

  forces += wallForces.field(obj_center.x, obj_center.y, d.x,d.y);

  // d = forces.field(obj_center.x, obj_center.y, d.x, d.y) + attractor.field(obj_center.x, obj_center.y, d.x, d.y);

  //if moving to an object, counter that object's repulsion
  //   if (pt.task.goal.target == TerrainBase::Goal::OBJ){
  //     VecField::Emitter *em = create_emitter(pt.task.goal.obj, 0 - REPULSE_STRENGTH);
  //     d = d + em->field(obj_center.x, obj_center.y, d.x, d.y);
  //     delete em;
  //     em = 0;
  //   }

  Loc target;
  target.x = round(obj_center.x + forces.x);
  target.y = round(obj_center.y + forces.y);
  add_move_command(obj, target, cmds);

  //update emitter for this object

//   d.normalize();
//   d = d * obj->get_max_speed();
//   Vec2<real8> new_pos(d.x + obj_center.x, d.y + obj_center.y);
  
//   forces.remove_emitter(obj2obstacle[obj].em);
//   obj2obstacle[obj].em = new VecField::OrthoEmitter(new_pos.x, new_pos.y, obj->get_radius() + 4, obj->get_radius() + 16, REPULSE_STRENGTH);

//   forces.add_emitter(obj2obstacle[obj].em); 
}

//---------------------------------------------------------------
bool ForceFieldPE::is_at_goal(const Object* obj, const TerrainBase::Task &task){
  
  Loc obj_loc;
  if (task.goal.target == TerrainBase::Goal::OBJ ) {
    switch(task.goal.mode) {
    case TerrainBase::Goal::TOUCH :
      return( obj->distance_to(*task.goal.obj) <= 4 );
      break;
 
    case TerrainBase::Goal::VICINITY :
      return (obj->distance_to(*task.goal.obj) <= task.goal.distance);
      //assert(0 && "Not implemented yet");
      break;
      
    case TerrainBase::Goal::ATTACK :
      ERR("Not implemented yet");
      break;
	
    default:
      ERR("Unknown mode");
    }
  } else if (task.goal.target == TerrainBase::Goal::LOCATION ) {
    switch(task.goal.mode) {
    case TerrainBase::Goal::TOUCH :
      obj_loc = get_center_loc(obj);
      return (obj_loc.distance(task.goal.loc) < TOUCH_THRESHOLD);
      break;

    case TerrainBase::Goal::VICINITY :
      obj_loc = get_center_loc(obj);
      return (obj_loc.distance(task.goal.loc) < task.goal.distance);
      break;

    case TerrainBase::Goal::ATTACK :
      ERR("unimplemented");
      break;

    default:
      ERR("Unknown mode");
    }
  } else {
    ERR("Unknown target type");
  }
  return false;
}
//---------------------------------------------------------------
Loc ForceFieldPE::get_center_loc(const Object* obj){
  Loc obj_loc;
  obj->get_center(obj_loc.x, obj_loc.y);
  return obj_loc;
}
//---------------------------------------------------------------

Vec2<real4> ForceFieldPE::get_attractor(PathfindTask &pt){
  Loc object_location;
  Vec2<real4> to_wp, scaled_to_wp, obj_center, point_on_line, next_wp;
  const Object* obj;

  obj = *(pt.task.objs.begin());
  object_location = get_center_loc(obj);
  obj_center = Vec2<real4>(object_location.x, object_location.y);
  
  // std::cerr << "Current WP: " << pt.current_wp << " Size: " << pt.path.locs.size() << std::endl;
  //std::cerr << "Current WP: " << pt.path.locs.at(pt.current_wp) << std::endl;

  assert(pt.current_wp >= 0 && pt.current_wp < static_cast<sint4>(pt.path.locs.size()));
  if (pt.current_wp <= 0){ 
    point_on_line = obj_center;
  }
  else{
    point_on_line = find_point_on_line(obj_center, pt.path.locs.at(pt.current_wp),
				       pt.path.locs.at(pt.current_wp-1));
  }
  
  real4 offset = obj->get_radius() + PATH_LOOK_AHEAD;
  next_wp = Vec2<real4>(pt.path.locs.at(pt.current_wp).x, pt.path.locs.at(pt.current_wp).y);
  scaled_to_wp = to_wp = next_wp - point_on_line;
  if (scaled_to_wp.x != 0 || scaled_to_wp.y != 0){
    scaled_to_wp.normalize();
    scaled_to_wp = scaled_to_wp * offset;
  }

  //std::cerr << "to_wp: " << to_wp.length() << " scaled: " << scaled_to_wp.length() << std::endl;

  //If wp_attract is set, set the attractor to the wp unless you're near it.
  if (wp_attract && scaled_to_wp.length() < to_wp.length()){
    return next_wp;
  }
  while (scaled_to_wp.length() >= to_wp.length()){
    offset = offset - to_wp.length();
    point_on_line = next_wp;
    if (pt.current_wp >= (static_cast<sint4>(pt.path.locs.size()) - 1)){
      int i = pt.path.locs.size() -1;
      return Vec2<real4>(pt.path.locs.at(i).x, pt.path.locs.at(i).y); //the last wp (the goal) is returned
    }
    pt.current_wp++;
    //std::cout << "new wp: " << pt.path.locs.at(pt.current_wp).x
    //		<< ", " << pt.path.locs.at(ptcurrent_wp).y << "\n";
    next_wp = Vec2<real4>(pt.path.locs.at(pt.current_wp).x, pt.path.locs.at(pt.current_wp).y);
    scaled_to_wp = to_wp = next_wp - point_on_line;
    
    if (scaled_to_wp.x != 0 || scaled_to_wp.y != 0){
      scaled_to_wp.normalize();
      scaled_to_wp = scaled_to_wp * offset;
    }
  }
  Vec2<real4> r = point_on_line + scaled_to_wp;
  assert(fabs(point_on_line.x) <= 1.84391e19);
  assert(fabs(scaled_to_wp.x) <= 1.84391e19);
  return point_on_line + scaled_to_wp;
}

//---------------------------------------------------------------
Vec2<real4> ForceFieldPE::find_point_on_line(Vec2<real4> point, Loc line_start, Loc line_end)
{
  sint4 line_x = line_end.x - line_start.x;
  sint4 line_y = line_end.y - line_start.y;

  real4 v_x = point.x - line_start.x;
  real4 v_y = point.y - line_start.y;

  real4 a = line_x * v_x + line_y * v_y;
  real4 b = line_x * line_x + line_y * line_y;

  Vec2<real4> proj(line_start.x + line_x * a / b, line_start.y + line_y * a / b);

  return proj;
}

//---------------------------------------------------------------
void ForceFieldPE::add_move_command(const Object *obj, Loc target, Vector<TerrainBase::MoveCmd> &cmds){
  TerrainBase::MoveCmd m;
  m.obj = obj;
  m.next_loc = target;
  m.speed = obj->get_max_speed();
  cmds.push_back(m);
}

//---------------------------------------------------------------

void ForceFieldPE::add_stop_command(const Object *obj, Vector<TerrainBase::MoveCmd> &cmds){
  TerrainBase::MoveCmd m;
  m.obj = obj;
  m.next_loc = get_center_loc(obj);
  m.speed = 0;
  cmds.push_back(m);
}

//---------------------------------------------------------------
sint4 ForceFieldPE::get_team(const Object *obj){
  //fixme: this is a hack. There should be a better way to get the team
  const GameObj* p = dynamic_cast<const GameObj*>(obj);
  if (p){
    return *(p->sod.owner);
  }
  else
    return -1;
}

//---------------------------------------------------------------

VecField::Emitter* ForceFieldPE::create_emitter(const Object *obj, real4 mag){
  VecField::Emitter* em = 0;
  if (obj->get_shape() == Object::CIRCLE){
    Loc obj_center = get_center_loc(obj);
    real4 tx = obj_center.x;
    real4 ty = obj_center.y;
    em = new VecField::PointEmitter2(tx, ty, obj->get_radius(), obj->get_radius() + 16, mag);
    em = new VecField::Randomizer(0.5, em);
  }
  else if (obj->get_shape() == Object::RECTANGLE){
    Loc p1, p2;
    obj->get_p1(p1.x, p1.y);
    obj->get_p2(p2.x, p2.y);
    real4 tx1 = p1.x;
    real4 tx2 = p2.x;
    real4 ty1 = p1.y;
    real4 ty2 = p2.y;
    em = new VecField::RectangleEmitter(tx1, ty1, tx2, ty2, 0, 12, mag);
  }
  return em;
}
//---------------------------------------------------------------
