#ifndef Bucket_H
#define Bucket_H

// $Id: Bucket.H,v 1.1 2006/12/31 04:09:40 orts_furtak Exp $
      
// This is an ORTS file (c) Michael Buro, licensed under the GPL

/**
 * A bucket provides noth,south,east, and west faces for the object
 * it contains. This is used to optimize search though the grid.
 * So if a move to the north is being concidered, having a clear path
 * for the north face implise that one exists for the entire cirlce to 
 * move north. That way only the appropriet face needs to be checked.
 * */

#include "Path.H"
#include "SimpleMap.H"

class Bucket {

public:
  Bucket() {}
  Bucket(const Vector<Point2i> &points_);
  ~Bucket();

  const Vector<Point2i> &get_points() const { return points; }
  const Point2i &get_center() const { return center; }

  const Vector<Point2i> &get_n() const { return pointsN; }
  const Vector<Point2i> &get_e() const { return pointsE; }
  const Vector<Point2i> &get_s() const { return pointsS; }
  const Vector<Point2i> &get_w() const { return pointsW; }
 
  void set_center(const Point2i &c) { center = c; }

  void set_n(const Vector<Point2i> &p) { pointsN = p; }
  void set_e(const Vector<Point2i> &p) { pointsE = p; }
  void set_s(const Vector<Point2i> &p) { pointsS = p; }
  void set_w(const Vector<Point2i> &p) { pointsW = p; }

  /**
   * check the appropriet face(s) to move in the give direction,
   * place the cost of movement in that direction in cost
   * */
  bool check_direction(const SimpleMap<SizeCell> &map, sint4 d, sint4 &cost);
	  
  /**
   * check that there is open space for the movement of the squares in the
   * given vector in the given direction (using x and y displacements).
   * place the cost of the movement in cost.
   * */
  bool check_vector(const SimpleMap<SizeCell> &map, const Vector<Point2i> &points,
		    sint4 dx, sint4 dy, sint4 &cost);

private:

  Point2i center;
  
  Vector<Point2i> points; // points of the border of the bucket
  
  Vector<Point2i> pointsN; // points of the north face of the bucket
  Vector<Point2i> pointsE; // points of the east face of the bucket
  Vector<Point2i> pointsS; // points of the south face of the bucket
  Vector<Point2i> pointsW; // points of the west face of the bucket
  
};

#endif //_BUCKET_H_
