// $Id: PathProcessor.C,v 1.1 2006/12/31 04:09:40 orts_furtak Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"
#include "PathProcessor.H"

using namespace std;

const sint4 PathProcessor::FROM = EventFactory::new_who();
const sint4 PathProcessor::PATH_SUCCESS = EventFactory::new_what();
const sint4 PathProcessor::PATH_FAILURE = EventFactory::new_what();

//===================================================================

void PathProcessor::remove_path(GameObj *obj)
{
  // fixme: SLOW!
  FORALL (paths, i) {
    UnitPath &cur_path = *i;
    if (cur_path.unit == obj) {
      paths.erase(i);
      break;
    }
  }
}

//===================================================================

void PathProcessor::process_paths()
{
  remove_dead_paths();

  for (Vector<UnitPath>::iterator j = paths.begin(); j != paths.end();) {
    Vector<UnitPath>::iterator i = j++;
    UnitPath &cur_path = *i;
    
    if (cur_path.sleeping()) continue;

    if (*cur_path.unit->sod.is_moving != 0) continue;

    sint4 x = *cur_path.unit->sod.x;
    sint4 y = *cur_path.unit->sod.y;

    Point2i p(x, y);
    Point2i cur_goal = cur_path.path.at(cur_path.current_target);
    
    if (p == cur_goal || cur_path.tries == -1) {

      // reached way-point || just getting started

      //      cout << "active = " << cur_path.active << endl;
      cur_path.tries = 0;

      if (cur_path.current_target == 0) { // path is in reverse order (goal first)
	// reached the goal
	j = paths.erase(i);
        call_handlers(PathStatusEvent(FROM, PATH_SUCCESS, cur_path.unit));
	continue;
      }

      // move to next waypoint
      if (p == cur_goal) cur_path.current_target--;

      Vector<sint4> params;
      Point2i goal = cur_path.path.at(cur_path.current_target);
      params.push_back(goal.x);
      params.push_back(goal.y);
      params.push_back(*cur_path.unit->sod.max_speed);

      cur_path.unit->set_action("move", params);
      continue;
    }
    
    // unit is not moving
    // not at waypoint/goal

    if (cur_path.spot != p) {
      //      cur_path.tries = 0;
      cur_path.mark_spot(p);
    } else {
      cur_path.tries++;
    }
    
    if (cur_path.tries > 1) {
      // unexpected collition... try to repair

      cout << "collision " << cur_path.tries << endl;

      if (cur_path.tries <  3) {
	int delay = 500 + (p.x-cur_goal.x)+ (p.y-cur_goal.y) % 500;
	cur_path.sleep(delay);
	cur_path.tries++;
	continue;
      }

      cout << "failure" << endl;
  
      SActParam param;
      param.obj.push_back(cur_path.unit);
      
      const char *c = "WOW!";
      param.val.push_back(stype(c));

      ObjAction *d = cur_path.unit->action("say");
      // if (d) d->Act(param);
      param.val.clear();

      if (cur_path.tries > 5) {
        cout << "giving up" << endl;
	c = "Giving up!";
        call_handlers(PathStatusEvent(FROM, PATH_FAILURE, cur_path.unit));
	cur_path.kill();
	continue;
      }

      if (cur_path.tries >= 3 && cur_path.tries <= 5) {
	failed.push_back(cur_path);
	c = "Repathing...";
      }
	  
      param.val.push_back(stype(c));
      d = cur_path.unit->action("say");
      // if (d) d->Act(param);
      
      continue;
    }  
  }
}

//===================================================================
