// $Id: SimpleGlobalPFOverlay.C,v 1.1 2006/12/31 04:09:40 orts_furtak Exp $

#include "SimpleGlobalPFOverlay.H"
#include "GfxModule.H"
#include "GfxGlobal.H"
#include "Game.H"

//===================================================================
SimpleGlobalPFOverlay::SimpleGlobalPFOverlay(SimplePFAdapter &pf_, GfxModule &gfxm_)
  : pf(pf_), dot(gfxm_), enabled(true), gfxm(&gfxm_)
{
}

//===================================================================
bool SimpleGlobalPFOverlay::handle_event(const Event &e)
{
  if (e.get_who() == GfxModule::FROM) {
    if (e.get_what() == GfxModule::DRAW_MSG) {
      draw();
      return true;
    }
  }
  return false;
}

//===================================================================
void SimpleGlobalPFOverlay::draw()
{
  if (!enabled) return;
  
  if (!pf.initiated) return;
  
  boost::shared_ptr<const SimpleTerrain::SimpleGlobalPF::Map> map;
  
  map = pf.pfEngine->get_ground_map();

  sint4 map_width = map->get_w();
  sint4 map_height = map->get_h();
  
  sint4 gw = gfxm->get_game().get_map().get_width();
  sint4 gh = gfxm->get_game().get_map().get_height();

  real4 dx = (real4)gw / (real4)(map_width-1);
  real4 dy = (real4)gh / (real4)(map_height-1);

  dot.begin_squares();
  FORS (xi, map_width) {
    FORS (yi, map_height) {
      if ((*map)(xi, yi).cost != 0 ) {
        
	Vec3<real4> c;
	//real4 v = (*map)(xi, yi).max_size / 20.0;
	c = Vec3<real4>(0, 1, 1);
	//          if ((*map)(xi, yi).max_size < 1) c = Vec3<real4>(1, 0, 0);
	//          if ((*map)(xi, yi).cost > 0)     c = Vec3<real4>(0, 1, 1);

	real4 x1 = (xi-0.5) * dx;
	real4 x2 = (xi+0.5) * dx;
	real4 y1 = (yi-0.5) * dy;
	real4 y2 = (yi+0.5) * dy;
        
	dot.draw_square(x1, y1, x2, y2, c);
      }
      if ((*map)(xi, yi).max_size <= 0){
	Vec3<real4> c;
	c = Vec3<real4>(0, 1, 0);
	real4 x1 = (xi-0.5) * dx;
	real4 x2 = (xi+0.5) * dx;
	real4 y1 = (yi-0.5) * dy;
	real4 y2 = (yi+0.5) * dy;
	dot.draw_square(x1, y1, x2, y2, c);
      }
      if ((*map)(xi, yi).marked > 0){
	Vec3<real4> c;
	c = Vec3<real4>(1, 0, 0);
	real4 x1 = (xi-0.5) * dx;
	real4 x2 = (xi+0.5) * dx;
	real4 y1 = (yi-0.5) * dy;
	real4 y2 = (yi+0.5) * dy;
	dot.draw_square(x1, y1, x2, y2, c);
      }
    }
  }
  dot.end_squares();

}
