#include "Base.h"
#include "Global.H"
#include "TR_PFEngine.H"


//DJD: new defines and declarations of static variables in TraFace {

//the epsilon to use for the DCDT
#define EPSILON 0.00001f
//the number of sections of a circle for the arcs around vertices
#define SECTIONS 20
//a float definition for pi
#define PI 3.1415926535897932384626433832795f
//a difference of less than this much influence doesn't matter
//#define NO_INFLUENCE 0.01f
//number of sides of a polygon to put in for a circular object
#define SIDES 8
//fraction of a circle to move when trying a spot adjacent to an obstacle
#define TRIES 10

//SrArray<TraFace *> TraFace::processing;
//int TraFace::current = 0;

//new defines and declarations of static variables in TraFace }


TR_PFEngine::TR_PFEngine(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, sint4 client_pID_, sint4 neutral_pID_) : GlobalPathfinder(){
  client_pID = client_pID_;
  neutral_pID = neutral_pID_;

  //DJD: initialize traignulator with the boundary of the environment {
  float width = (float)(tiles_x_ * tile_points_);
  float height = (float)(tiles_y_ * tile_points_);
  SrPolygon boundary;
  boundary.push().set(0.0f, 0.0f);
  boundary.push().set(width, 0.0f);
  boundary.push().set(width, height);
  boundary.push().set(0.0f, height);
  pfDcdt.init(boundary, EPSILON, 1);
  //initialize traignulator with the boundary of the environment }
}

//DJD: function definitions for updating the terrain for obstacles {

//adds an obstacle to the environment representation
void TR_PFEngine::AddObstacle(GameObj *gob)
{
	//create a closed polygon for the obstacle
	SrPolygon poly;
	poly.open(false);
	//check if the obstacle is rectangular
	if (gob->get_int("shape") == Object::RECTANGLE)
	{
		//get its corners
		sint4 x1, y1, x2, y2;
		gob->get_p1(x1, y1);
		gob->get_p2(x2, y2);
		//create a polygon using them
		poly.push().set((float)x1, (float)y1);
		poly.push().set((float)x2, (float)y1);
		poly.push().set((float)x2, (float)y2);
		poly.push().set((float)x1, (float)y2);
	}
	//check if the obstacle is circular
	else if (gob->get_int("shape") == Object::CIRCLE)
	{
		//get its center and radius
		sint4 cx, cy;
		gob->get_center(cx, cy);
		sint4 r = gob->get_radius();
		float x, y;
		float angle = 0.0f;
		float inc = 2.0f * PI / (float)SIDES;
		//create a certain number of vertices to approximate the circle
		for (int i = 0; i < SIDES; i++)
		{
			//calculate the position of the vertex
			x = (float)cx + cos(angle) * (float)r;
			y = (float)cy + sin(angle) * (float)r;
			//add it to the polygon
			poly.push().set(x, y);
			//and increment the angle to the next vertex
			angle += inc;
		}
	}
	else
	{
		//we only deal with circular and rectangular obstacles
		return;
	}
	//get the obstacle's ID within ORTS
	ObjId ortsId = GameObj::get_obj_id(gob);
	//insert the obstacle's polygon into the triangulation and record its ID there
	int dcdtId = pfDcdt.insert_polygon(poly);
	//add this to our list of obstacles
	obstacles.push_back(TR_Obstacle(ortsId, dcdtId));
}

//removes an obstacle from the environment representation
void TR_PFEngine::RemoveObstacle(GameObj *gob)
{
	//get the ID of the obstacle within ORTS
	ObjId ortsId = GameObj::get_obj_id(gob);
	//go through our list of obstacles
	for (std::vector<TR_Obstacle>::iterator i = obstacles.begin(); i != obstacles.end(); i++)
	{
		//when we find one with a matching ID,
		if (ortsId == i->ortsId)
		{
			//if the object was successfully added to the triangulation,
			if (i->dcdtId != -1)
			{
				//remove it
				pfDcdt.remove_polygon(i->dcdtId);
			}
			//remove it from the list of obstacles
			obstacles.erase(i);
			break;
		}
	}
}

//updates the abstraction based on the changed triangulation
void TR_PFEngine::UpdateAbstraction()
{
	//delete the abstraction
	pfDcdt.DeleteAbstraction();
	//rebuild it
	pfDcdt.Abstract();
	//repair the sectors that have been invalidated
	pfDcdt.InitializeSectors();
	//and record that the new faces have been processed
	TraFace::Clear();
}

//function definitions for updating the terrain for obstacles }

void TR_PFEngine::insert_object(const Object *obj){
//DJD: add logic for keeping track of obstacles in the triangulation {
	GameObj *gob = (GameObj *)obj;
	//if the object is an obstacle,
	if (TR_Obstacle::IsObstacle(gob))
	{
		//add it to the representation
		AddObstacle(gob);
		UpdateAbstraction();
		//and don't consider it for influence
		return;
	}
//add logic for keeping track of obstacles in the triangulation }
#if 0
  //DJD: add logic for keeping track of enemy unit influence in the triangulation {
  //only processes enemy units
  GameObj *gob = (GameObj *)obj;
  const sintptr *owner = gob->get_const_int_ptr("owner");
  if ((*owner == client_pID) || (*owner == neutral_pID))
    {
      return;
    }
  //locates the triangle in which the object's center lies
  TraBase *s = NULL;
  sint4 x, y;
  obj->get_center(x, y);
  //sets the object's triangle to null if not found
  if (pfDcdt.LocatePoint(x, y, s) == SeTriangulator::NotFound)
    {
      objects.push_back(TR_Object(gob));
    }
  else
    {
      //otherwise sets it to the triangle face found
      TraFace *face = (TraFace *)s->fac();
      TR_Object trobj(gob, face);
      objects.push_back(trobj);
      //and updates the unit's influence in the triangulation
      pfDcdt.UpdateInfluence(face, trobj.influence);
    }
  //add logic for keeping track of enemy unit influence in the triangulation }
#endif
}

void TR_PFEngine::remove_object(const Object *obj){
//DJD: add logic for removing dead or vanished obstacles from the triangulation {
	GameObj *gob = (GameObj *)obj;
	//if the object is an obstacle,
	if (TR_Obstacle::IsObstacle(gob))
	{
		//remove it from the representation
		RemoveObstacle(gob);
		UpdateAbstraction();
		//and don't consider it for influence
		return;
	}	
//add logic for removing dead or vanished obstacles from the triangulation }
#if 0
  //DJD: add logic for removing influence of vanished or dead enemy units {
	//only processes enemy units
	const sintptr *owner = ((GameObj *)obj)->get_const_int_ptr("owner");
	if ((*owner == client_pID) || (*owner == neutral_pID))
	{
		return;
	}
	ObjId id = GameObj::get_obj_id((GameObj *)obj);
	//goes through the list of objects
	for (std::vector<TR_Object>::iterator i = objects.begin(); i != objects.end(); i++)
	{
		//upon finding the matching object
		if (i->id == id)
		{
		  pfDcdt.UpdateInfluence(i->face, -(i->influence));
		  //removes it from the list
		  objects.erase(i);
		  break;
		}
	}
	//add logic for removing influence of vanished or dead enemy units }
#endif
}

void TR_PFEngine::update_object(const Object *obj){
//DJD: add logic for updating positions of obstacles in the triangulation {
	GameObj *gob = (GameObj *)obj;
	//if the object is an obstacle,
	if (TR_Obstacle::IsObstacle(gob))
	{
		//update it within the representation
		RemoveObstacle(gob);
		AddObstacle(gob);
		UpdateAbstraction();
		//and don't consider it for influence
		return;
	}	
//add logic for updating positions of obstacles in the triangulation }
#if 0
  //DJD: add logic for updating enemy position or influence in the triangulation {
  //only processes enemy units
  const sintptr *owner = ((GameObj *)obj)->get_const_int_ptr("owner");
  if ((*owner == client_pID) || (*owner == neutral_pID))
    {
      return;
    }
  ObjId id = GameObj::get_obj_id((GameObj *)obj);
  //goes through the objects
  for (uint4 i = 0; i < objects.size(); i++)
    {
      //until it finds the desired object
      if (objects[i].id == id)
	{
	  //record the object's old triangle and influence
	  TraFace *oldFace = objects[i].face;
	  float oldInfluence = objects[i].influence;
	  //calculate the object's new influence
	  objects[i].CalculateInfluence((GameObj *)obj);
	  //remove the object's influence from it last triangle
	  if (objects[i].face != NULL)
	    {
	      pfDcdt.UpdateInfluence(objects[i].face, -(objects[i].influence));
	    }
	  TraBase *s = NULL;
	  SeTriangulator::LocateResult result;
	  sint4 x, y;
	  obj->get_center(x, y);
	  //if possible, uses the last triangle as the starting point
	  //otherwise, uses the triangle around the nearest sector midpoint
	  //then finds the triangle in which the object is currently centered
	  if (objects[i].face == NULL)
	    {
	      result = pfDcdt.LocatePoint(x, y, s);
	    }
	  else
	    {
	      result = pfDcdt.LocatePoint(objects[i].face, x, y, s);
	    }
	  //sets the object's face to NULL if it wasn't found
	  if (result == SeTriangulator::NotFound)
	    {
	      objects[i].face = NULL;
	    }
	  else
	    {
	      //otherwise, sets it appropriately
	      objects[i].face = (TraFace *)s->fac();
	    }
	  //checks if the object has moved triangles or if it has changed influence
	  if ((oldFace != objects[i].face) || (fabs(oldInfluence - objects[i].influence) > NO_INFLUENCE))
	    {
	      //if it has and the old triangle was valid, subtract its influence from the old triangle
	      if (oldFace != NULL)
		{
		  pfDcdt.UpdateInfluence(oldFace, -oldInfluence);
		}
	      //if the new triangle is valid, add the object's influence to the new triangle
	      if (objects[i].face != NULL)
		{
		  //and updates the object's new influence on the map
		  pfDcdt.UpdateInfluence(objects[i].face, objects[i].influence);
		}
	    }
	  break;
	}
    }
  //add logic for updating enemy position or influence in the triangulation }
#endif
}

void TR_PFEngine::insert_boundaries(const Vector<TerrainBase::Segment> &segs){
 
  //DJD: combine new segments with existing ones and update the triangulation {
  //TODO: keep track of buildings, as in add/change/remove objects
  //TODO: insert and remove segments representing their boundaries accordingly
  sm.clear_delta();
  // send all segments to merge class
  FORALL (segs, s)
    {
      DCDT::Segment s1(s->l1.x, s->l1.y, s->l2.x, s->l2.y);
      sm.new_segment(s1);
    }
  // process all changes
  bool changed = false;
  sint4 add_n = 0, del_n = 0;
  FORS (i, 4)
    {
      del_n += sm.del[i].size();
      FORALL (sm.del[i], it)
	{
	  pfDcdt.remove_polygon(*it);
	  changed = true;
	}
    }
  sint4 total = 0;
  FORS (i, 4)
    {
      total += sm.all[i].size();
      add_n += sm.add[i].size();
      FORALL (sm.add[i], it)
	{
	  const SortSegment &ss = **it;
	  SrPolygon pol;
	  pol.push().set(ss.s.x1, ss.s.y1);
	  pol.push().set(ss.s.x2, ss.s.y2);
	  ss.label = pfDcdt.insert_polygon(pol);
	  changed = true;
	}
    }
  //if the triangulation was changed,
  if (changed)
    {
      //delete the abstraction
      pfDcdt.DeleteAbstraction();
      //rebuild it
      pfDcdt.Abstract();
      //repair the sectors that have been invalidated
      pfDcdt.InitializeSectors();
      //and record that the new faces have been processed
      TraFace::Clear();
    }
  //combine new segments with existing ones and update the triangulation }

}

bool TR_PFEngine::find_path(const Object *obj, const Loc &goal, Path& output, ConsiderObjects consider){

  //DJD: find a path from an object to some goal {
  //gets the coordinates of the unit
  Loc start;
  obj->get_center(start.x, start.y);
  //if the pathfinding target is an area,
  
  //find a path there
  find_path(start, goal, obj->get_radius(), output, consider);

  if (output.locs.size() > 0){
    return true;
  }
  else{
    return false;
  }
 
  //find a path from an object to some goal }

}

bool TR_PFEngine::find_path(const Object *obj, const Object *goal, Path& output, ConsiderObjects consider){

//DJD: find a path from an object to some goal {
	//gets the current time
	Uint32 startTime = SDL_GetTicks();
	//gets the coordinates of the unit
	Loc start;
	obj->get_center(start.x, start.y);
		//if it is an object, determine its position
		Loc dest;
		GameObj *gob = (GameObj *)goal;
		//if the object is not an obstacle, we can go to its center
		if (!TR_Obstacle::IsObstacle(gob))
		{
			if (gob->get_int("shape") == Object::RECTANGLE)
			{
				//if the destination object is rectangular,
				sint4 x1, y1, x2, y2;
				gob->get_p1(x1, y1);
				gob->get_p2(x2, y2);
				//calculate it center and go there
				dest.x = x1 + x2 / 2;
				dest.y = y1 + y2 / 2;
			}
			else if (gob->get_int("shape") == Object::CIRCLE)
			{
				//if the destination object is circular, go to its center
				gob->get_center(dest.x, dest.y);
			}
			else
			{
				//we're only dealing with rectangles and circles for now
				return -1;
			}
			//find a path there
			return find_path(start, dest, obj->get_radius(), output, consider);
		}
		//if it is an obstacle, we must set the destination to be outside it
		else
		{
			//calculate the center and "radius" of the obstacle
			sint4 x, y, r;
			if (gob->get_int("shape") == Object::RECTANGLE)
			{
				sint4 x1, y1, x2, y2;
				gob->get_p1(x1, y1);
				gob->get_p2(x2, y2);
				x = x2 - x1;
				y = y2 - y1;
				r = (sint4)(sqrt((float)(x * x + y * y)) / 2.0f) + obj->get_radius();
				x = x1 + x2 / 2;
				y = y1 + y2 / 2;
			}
			else if (gob->get_int("shape") == Object::CIRCLE)
			{
				gob->get_center(x, y);
				r = gob->get_radius() + obj->get_radius();
			}
			else
			{
				//we're only dealing with rectangles and circles for now
				return -1;
			}
			std::cout << "Pathing to (" << x << ", " << y << ")\n";
			float base = atan2((float)start.y - (float)y, (float)start.x - (float)x);
			float diff = PI / (float)TRIES;
			for (int i = 0; i < TRIES; i++)
			{
				float theta = base + diff * (float)i;
				diff = -diff;
				dest.x = x + (sint4)(cos(theta) * (float)r);
				dest.y = y + (sint4)(sin(theta) * (float)r);
				float result = find_path(start, dest, obj->get_radius(), output, consider);
				if ((result != -1) || (SDL_GetTicks() >= startTime + 10))

				{
				  std::cout << "Returning " << ((result == -1) ? "failed" : "successful") << " result after " << (SDL_GetTicks() - startTime) << "ms\n";
					return result;
				}
			}
			return -1;
		}
//find a path from an object to some goal }
#if 0
//DJD: find a path from an object to some goal {
  //gets the coordinates of the unit
  Loc start;
  obj->get_center(start.x, start.y);
  //if the pathfinding target is an area,
  
  //if it is an object, determine its position
  Loc dest;
  goal->get_center(dest.x, dest.y);
  //and find a path there
  return find_path(start, dest, obj->get_radius(), output, consider);
  //TODO: if it's a building, find a path to get near it instead of inside it
    
//find a path from an object to some goal }
#endif
}

bool TR_PFEngine::find_path(const Loc &l1, const Loc &l2, sint4 radius, Path &output, ConsiderObjects /*consider*/){
  radius += 2;
  //DJD: find a path from one location to another with given radius {
  //get the current time
  Uint32 time = SDL_GetTicks();
  std::cout << "Going from (" << l1.x << ", " << l1.y
	    << ") to (" << l2.x << ", " << l2.y << ")\n";
  //find the path with soft timeout of 2 milliseconds and hard timeout of 5 milliseconds
  if (!pfDcdt.SearchPathFast((float)l1.x + 0.5f, (float)l1.y + 0.5f, (float)l2.x + 0.5f, (float)l2.y + 0.5f,
			      (float)radius, time + 2, time + 5))
    {
      return false;
    }
  //get the path and record the first waypoint
  SrPolygon CurPath = pfDcdt.GetPath();
  output.locs.push_back(Loc((sint4)CurPath[0].x, (sint4)CurPath[0].y));
  //go through the straight segments of the path
  for (int i = 0; i < CurPath.size() - 1; i += 2)
    {
      //if another follows it, construct the curved segment in between
      if (i < CurPath.size() - 3)
	{
	  //determine the angles of both adjoining segments
	  float angle1 = atan2(CurPath[i + 1].y - CurPath[i].y, CurPath[i + 1].x - CurPath[i].x);
	  float angle2 = atan2(CurPath[i + 3].y - CurPath[i + 2].y, CurPath[i + 3].x - CurPath[i + 2].x);
	  //calculate the difference between the two
	  float diff = angle1 - angle2;
	  diff += (diff <= -PI) ? 2.0f * PI : (diff > PI) ? -2.0f * PI : 0.0f;
	  //calculate the angle to the center of curvature
	  float start = angle1 + ((diff < 0.0f) ? PI / -2.0f : PI / 2.0f);
	  //followed by its coordinates
	  float cornerX = CurPath[i + 1].x - cos(start) * radius;
	  float cornerY = CurPath[i + 1].y - sin(start) * radius;
	  //calculate the number of segments to draw based on the angle
	  int number = (int)(fabs(diff) / (2.0f * PI) * (float)SECTIONS);
	  float theta = -diff / (float)number;
	  float r = radius / fabs(cos(theta * 0.5f));
	  start += theta * 0.5f;
	  //and go through the segments
	  for (int j = 0; j < number; j++)
	    {
	      //calculate the next vertex in the line strip
	      Loc l((sint4)(cornerX + cos(start) * r), (sint4)(cornerY + sin(start) * r));
	      //only insert each vertex if it's different than the last
	      if ((l.x != output.locs.back().x) && (l.y != output.locs.back().y))
		{
		  output.locs.push_back(l);
		}
	      //incrementing the angle each time
	      start += theta;
	    }
	}
    }
  //calculate and place the last vertex on the curve
  Loc l((sint4)CurPath[CurPath.size() - 1].x, (sint4)CurPath[CurPath.size() - 1].y);
  if ((l.x != output.locs.back().x) && (l.y != output.locs.back().y))
    {
      output.locs.push_back(l);
    }
  
  //reverse order
  for (uint i = 0; i <= (output.locs.size()-1)/2; i++){
    Loc temp;
    temp = output.locs[i];
    output.locs[i] = output.locs[output.locs.size()-1-i];
    output.locs[output.locs.size()-1-i] = temp;
  }


  return true;
  //find a path from one location to another with given radius }

}
