// $Id: SI_Strategies.cpp,v 1.1 2006/08/10 17:16:02 nicholas Exp $


#include "SI_Strategies.h"

#include "SI_SimulationAI.H"
#include "SI_SimulationState.h"

#include "GameStateModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameObj.H"
//#include "GfxModule.H"
#include "TerrainModule.H"
#include "TerrainClassifier.H"

namespace Simul {

#define DEFINE_TASK(name)						\
  name::name(uint4 playerID,SimulationState* state,ObjectVector& origObjects,	\
	     TaskCreateFunctionPtr f,std::string n,real8 timeout,MoveStruct* extraData) : \
    Task(playerID,state,origObjects,f,n,timeout,extraData){	\
    exec();								\
  }									\
    name::name(Task* parent,ObjectVector& origObjects,TaskCreateFunctionPtr function,std::string n, \
	       real8 timeout,MoveStruct* extraData) :	\
      Task(parent,origObjects,function,n,timeout,extraData){ \
      exec();								\
    }									\
    name::name(const name& t) : Task(t){}				\
    Task* name::clone(SimulationState* state,Task* newParent){		\
      name* newTask = new name(*this);					\
      newTask->parent = newParent;					\
      newTask->state = state;						\
      FORALL(children,it){						\
	newTask->children.push_back((*it)->clone(state,newTask));	\
      }									\
      return newTask;							\
    }									\
    bool name::exec()


  DEFINE_TASK(SpreadAttack) {
    ObjectVector temp;

    switch(lineNum) {
    case 0:
      // need a copy of this vector, since the constructor for tasks modifies the original by removing objects from it
      temp = objects;

      FORALL(temp, it) {
        ObjectVector object;
	object.push_back(*it);

        GameObj* nearestBase = state->getNearestEnemyBaseOfGroup(playerID,object);
	if(nearestBase != NULL) {
	  new Attack(this,object,function,name,0.0,new MoveStruct(nearestBase));
	}
      }
      default:
        return false;
    }
    return false;
  }

  DEFINE_TASK(MassAttack) {
    // ObjectVector temp;
    switch(lineNum){

    case 0:
      if(state->smartJoin){
        new SmartJoin(this, objects, function, name, 0.0);
      }else{
        new Join(this, objects, function, name, 0.0);
      }

      ++lineNum;
      break;
    case 1:{
    // need a copy of this vector, since the constructor for tasks modifies the original by removing objects from it
	GameObj* nearestBase = state->getNearestEnemyBaseOfGroup(playerID,objects);
        //temp = objects;

	if(nearestBase != NULL){
	    new Attack(this,objects,function,name,0.0,new MoveStruct(nearestBase));
	}
        break;
      }
      default:
        return false;
    }
    return false;
  }

  DEFINE_TASK(Join) {
    assert(extraData == NULL); // extradata should be NULL for this task

    switch(lineNum) {
    case 0:{
      GameObj* centerObject = state->getNearestObject(objects, state->getCenterOfObjects(objects));
      assert(centerObject != NULL && centerObject->get_shape() == Object::CIRCLE);// takje this assert OUT later
      TerrainBase::Loc center_loc;
      centerObject->get_center(center_loc.x, center_loc.y);
      
      if(state->findspot){
	GameStateModule & gsm = state->simAI->get_gsm();
	center_loc = find_a_spot(center_loc, gsm);
      }

      new MoveToLocation(this,objects,function,name,0.0,new MoveStruct(center_loc));
      ++lineNum;
      break;
    }
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(Attack) {
    assert(extraData != NULL);

    switch(lineNum) {
    case 0:
      // everyone should arrive at the same time
      FORALL(objects,it) {
        TerrainBase::Task task(TerrainBase::Goal::OBJ, TerrainBase::Goal::TOUCH,
                               TerrainBase::Task::INDIVIDUAL);


        task.goal.obj = extraData->obj;

        GameObj* currObj = *it;

        ExtraObjectData newData;
        newData.target = extraData->obj;
        newData.state = ExtraObjectData::ATTACK;

        Player* currPlayer = state->getPlayer(playerID);

        currPlayer->getOrtsUnits()[currObj] = newData;

        task.objs.insert(currObj);
        state->simAI->doPath(task);
      }
      ++lineNum;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(MoveToLocation) {
    assert(extraData != NULL);
    assert(extraData->obj == NULL); // should not have an object as a target

    switch(lineNum) {
    case 0:
      // everyone should arrive at the same time
      FORALL(objects,it) {
        TerrainBase::Task task(TerrainBase::Goal::LOCATION, TerrainBase::Goal::VICINITY,
                               TerrainBase::Task::INDIVIDUAL);

        GameObj* currObj = *it;

        ExtraObjectData newData;
        newData.location = extraData->pos;
        newData.state = ExtraObjectData::MOVE;

        Player* currPlayer = state->getPlayer(playerID);

        currPlayer->getOrtsUnits()[currObj] = newData;

        task.goal.loc = extraData->pos;

        // we can base the distance based on how many units we have...
        task.goal.distance = 80;
        //        task.goal.distance = (sint4)(objects.size() * 1.5);
        task.objs.insert(currObj);
        state->simAI->doPath(task);

      }
      ++lineNum;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(MoveToObject) {
    assert(extraData != NULL);
    assert(extraData->obj != NULL);

    switch(lineNum) {
    case 0:
	// everyone should arrive at the same time
      FORALL(objects,it) {
        if(*it == extraData->obj){ // the center object does not need to move
          ExtraObjectData newData;
          newData.isDone = true;
          Player* currPlayer = state->getPlayer(playerID);
          currPlayer->getOrtsUnits()[*it] = newData;
          continue;
        }
        TerrainBase::Task task(TerrainBase::Goal::OBJ, TerrainBase::Goal::TOUCH,
                               TerrainBase::Task::INDIVIDUAL);

        ExtraObjectData newData;
        newData.target = extraData->obj;
        newData.state = ExtraObjectData::MOVE;

        Player* currPlayer = state->getPlayer(playerID);

        currPlayer->getOrtsUnits()[*it] = newData;

        task.goal.obj = extraData->obj;
        task.objs.insert(*it);
        state->simAI->doPath(task);
      }
      ++lineNum;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(DefendObject) {
    assert(extraData != NULL);

    switch(lineNum) {
    case 0:
	  // everyone should arrive at the same time
      FORALL(objects,it){
        // this should NEVER happen, since you should not be defending an object that is getting orders
        // to defend itself!
        if(*it == extraData->obj){
          ExtraObjectData newData;
          newData.isDone = true;
          Player* currPlayer = state->getPlayer(playerID);
          currPlayer->getOrtsUnits()[*it] = newData;
          continue;
        }
	TerrainBase::Task task(TerrainBase::Goal::OBJ, TerrainBase::Goal::TOUCH,
			      TerrainBase::Task::INDIVIDUAL);

	ExtraObjectData newData;
	newData.target = extraData->obj;
	newData.state = ExtraObjectData::DEFEND;

	Player* currPlayer = state->getPlayer(playerID);

	currPlayer->getOrtsUnits()[*it] = newData;

	task.goal.obj = extraData->obj;
	task.objs.insert(*it);
	state->simAI->doPath(task);
      }
      ++lineNum;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(HalfUnitsBaseDefence_MassAttack){

    switch(lineNum) {
    case 0:{
      Player* currPlayer = state->getPlayer(playerID);
      // make sure that we have enough units to do anything.. if not, then just don't defend bases
      uint4 numUnits = currPlayer->getOrtsUnits().size();
      uint4 numBases = currPlayer->getOrtsBases().size();
      if(numUnits/2 >= numBases){
        uint4 numUnitsPerBase = (numUnits/2) / numBases;
        FORALL(currPlayer->getOrtsBases(),it){
          ObjectVector unitsToDefendBase;
          state->getNearestObjectsToObject(unitsToDefendBase,objects,it->first,numUnitsPerBase);
          new DefendObject(this,unitsToDefendBase,function,name,0.0,new MoveStruct(it->first));

	}
      }
      // the rest of the units that are not defending go join up and attack!
      new MassAttack(this,objects,function,name,0.0);
      ++lineNum;
      break;
    }
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(TouchObj) {
    assert(extraData != NULL);
    //assert(extraData->obj != NULL);

    switch(lineNum){
    case 0:
      if (extraData->obj != NULL){
	if (extraData->obj->get_max_speed() > 0 && *extraData->obj->sod.owner == (sint4)playerID){
	  TerrainBase::Loc center_loc;
	  extraData->obj->get_center(center_loc.x, center_loc.y);
	  new MoveToLocation(this,objects,function,name,0.0,new MoveStruct(center_loc));
	}
	else{
	  new MoveToObject(this, objects, function,name,0.0, new MoveStruct(extraData->obj));
	}
      }
      ++lineNum;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(JoinAndMoveToObj) {
    assert(extraData != NULL); // extradata should not be NULL for this task

    assert(extraData->pos != TerrainBase::Loc(-1, -1));

    switch(lineNum) {
    case 0:{
      GameObj* centerObject = state->getNearestObject(objects, state->getCenterOfObjects(objects));
      assert(centerObject != NULL);// takje this assert OUT later
      TerrainBase::Loc center_loc;
      centerObject->get_center(center_loc.x, center_loc.y);
    
      assert(centerObject->get_max_speed() > 0); //Objects that have a max speed of 0 will be in the representation, so this would fail
      new MoveToLocation(this,objects,function,name,0.0,new MoveStruct(center_loc));

      ++lineNum;
      break;
    }
    case 1:
      new MoveToLocation(this,objects,function,name,0.0,new MoveStruct(extraData->pos));

      ++lineNum;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(SmartJoin) {

    switch(lineNum) {
    case 0:{
      // first, find the final location we will join at.. in this case, it MUST be the location, since units will move.
      // once influence maps are in, it should find a safe spot to join, instead of just the average position of all
      // the units
      GameObj* centerObject = state->getNearestObject(objects, state->getCenterOfObjects(objects));
      assert(centerObject != NULL);// take this assert OUT later
      TerrainBase::Loc  center_loc;
      centerObject->get_center(center_loc.x, center_loc.y);

      if(state->findspot)
	center_loc = find_a_spot(center_loc, state->simAI->get_gsm());
      
      // now make groups of X units(in this case, 5).  kind of a dumb way to do it... basically, just take one unit, find X-1
      // closest units, finding their average position, and then joining at the unit closest to that position
      while(objects.size() > 0){
	ObjectVector unitsToJoin;

        state->getNearestObjectsToObject(unitsToJoin,objects,objects[0],state->smartJoinSize);
        new JoinAndMoveToObj(this,unitsToJoin,function,name,0.0, new MoveStruct(center_loc));
      }
      ++lineNum;
      break;
    }
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(Hunter) {
    switch (lineNum){
    case 0:

      // make groups of X units(in this case, 5).  kind of a dumb way to do it... basically, just take one unit, find X-1
      // closest units, finding their average position, and then joining at the unit closest to that position
      while(objects.size() > 0){
	ObjectVector unitsToJoin;
        state->getNearestObjectsToObject(unitsToJoin,objects,objects[0],state->groupSize);

        new JoinAndHunt(this,unitsToJoin,function,name,0.0);
      }

      lineNum++;
      break;
    case 1:
      //new MassAttack(this, objects, function, name, 0.0);
      new SpreadAttack(this, objects, function, name, 0.0);

      lineNum++;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(Hunter2) {
    switch (lineNum){
    case 0:

	// make groups of X units(in this case, 5).  kind of a dumb way to do it... basically, just take one unit, find X-1
	// closest units, finding their average position, and then joining at the unit closest to that position
      while(objects.size() > 0){
	ObjectVector unitsToJoin;
	state->getNearestObjectsToObject(unitsToJoin,objects,objects[0],state->groupSize);

	new JoinAndHunt2(this,unitsToJoin,function,name,0.0);
      }

      lineNum++;
      break;
    case 1:
//      new MassAttack(this, objects, function, name, 0.0);

      new SpreadAttack(this, objects, function, name, 0.0);
      lineNum++;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(JoinAndHunt){
    assert(extraData == NULL);

    switch (lineNum){
    case 0:{
      if(state->smartJoin){
        new SmartJoin(this, objects, function, name, 0.0);
      }else{
        new Join(this, objects, function, name, 0.0);
      }
      lineNum++;
      break;
    }
    case 1:{
      GameObj* target = state->getNearestEnemyUnitOfGroup(playerID, objects);
      if (target != NULL){
        if(state->useMoveForAttack){
          new MoveToObject(this, objects, function,name,0.0, new MoveStruct(target));
        }else{
          new Attack(this, objects, function,name,0.0, new MoveStruct(target));
        }
      }
      break;
    }
    default:
      return false;
    }
    return false;
  }
  DEFINE_TASK(JoinAndHunt2){
    assert(extraData == NULL);

    switch (lineNum){
    case 0:{
      if(state->smartJoin){
        new SmartJoin(this, objects, function, name, 0.0);
      }else{
        new Join(this, objects, function, name, 0.0);
      }
      lineNum++;
      break;
    }
    case 1:{
      GameObj* target = state->getAverageNearestEnemyUnitOfGroup(playerID, objects);
      if (target != NULL){
        if(state->useMoveForAttack){
	  new MoveToObject(this, objects, function,name,0.0, new MoveStruct(target));
        }else{
	 new Attack(this, objects, function,name,0.0, new MoveStruct(target));
        }
      }
      break;
    }
    default:
      return false;
    }
    return false;
  }


  DEFINE_TASK(JoinSpreadAttack) {
    switch (lineNum){
    case 0:

      while(objects.size() > 0){
	ObjectVector unitsToJoin;
	state->getNearestObjectsToObject(unitsToJoin,objects,objects[0],5);

	new JoinAndAttack(this,unitsToJoin,function,name,0.0);
      }

      lineNum++;
      break;
    case 1:
      new SpreadAttack(this, objects, function, name, 0.0);
      lineNum++;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(JoinAndAttack){
    assert(extraData == NULL);

    switch (lineNum){
    case 0:{
      if(state->smartJoin){
        new SmartJoin(this, objects, function, name, 0.0);
      }else{
        new Join(this, objects, function, name, 0.0);
      }
      lineNum++;
      break;
    }
    case 1:{
      GameObj* nearestBase = state->getNearestEnemyBaseOfGroup(playerID,objects);
      if (nearestBase != NULL){
	new MoveToObject(this, objects, function,name,0.0, new MoveStruct(nearestBase));
      }
      break;
    }
    default:
      return false;
    }
    return false;
  }

  Task* createSmartJoin(uint4 id,SimulationState* s,ObjectVector& o,real8 timeout) {
    return new SmartJoin(id,s,o,createSmartJoin,"SmartJoin",timeout);
  }
  Task* createSpreadAttack(uint4 id,SimulationState* s,ObjectVector& o,real8 timeout) {
    return new SpreadAttack(id,s,o,createSpreadAttack,"SpreadAttack",timeout);
  }
  Task* createMassAttack(uint4 id,SimulationState* s,ObjectVector& o,real8 timeout) {
    return new MassAttack(id,s,o,createMassAttack,"MassAttack",timeout);
  }
  Task* createHalfUnitsBaseDefence_MassAttack(uint4 id,SimulationState* s,ObjectVector& o,real8 timeout) {
    return new HalfUnitsBaseDefence_MassAttack(id,s,o,createHalfUnitsBaseDefence_MassAttack,
                                               "HalfUnitsBaseDefence_MassAttack",timeout);
  }
  Task* createHunter(uint4 id,SimulationState* s,ObjectVector& o,real8 timeout) {
    return new Hunter(id,s,o,createHunter,"Hunter",timeout);
  }
  Task* createHunter2(uint4 id,SimulationState* s,ObjectVector& o,real8 timeout) {
    return new Hunter2(id,s,o,createHunter2,"Hunter2",timeout);
  }
  Task* createJoinSpreadAttack(uint4 id,SimulationState* s,ObjectVector& o,real8 timeout) {
    return new JoinSpreadAttack(id,s,o,createJoinSpreadAttack,"JoinSpreadAttack",timeout);
  }

 /*********************JAMES' STUFF **************************/  
  TerrainBase::Loc find_a_spot(TerrainBase::Loc center_loc, GameStateModule& gsm){
               
    mapTerrain::TerrainMap terrMap(&gsm.get_game());
    TerrainBase::Loc loc(0,0);
    sint4 standard = 9; sint4 range = 128;
    bool found = terrMap.find_free_loc(center_loc, loc, range, standard);
    range+= 32;
    //std::cout << "found = " << found << ", and meeting spot is (" <<  center_loc.x / 16 << "," << center_loc.y / 16 << ")" << std::endl;
    while(!found  && standard > 4){
      found = terrMap.find_free_loc(center_loc, loc, range, standard--);
      //std::cout << "found = " << found << ", and meeting spot is (" <<  center_loc.x / 16 << "," << center_loc.y / 16 << ")" << std::endl;
    }
    terrMap.print_map();
    return TerrainBase::Loc(loc.x, loc.y);
  }
  /*********************JAMES' STUFF **************************/ 

}



