
void GfxClient::handle_events()
{
  SDL_Event event;

  while (SDL_PollEvent(&event)) {

    switch (event.type) {
    case SDL_QUIT:
      quit = true;
      break;
    case SDL_KEYUP:
      event_key_up(event);
      break;
    case SDL_KEYDOWN:
      event_key_down(event);
      break;
    case SDL_MOUSEBUTTONDOWN:
#if TRANSFORM_MOUSE_Y
      event.button.y = win_h-1-event.button.y;
#endif
      event_mouse_down(event);
      break;
    case SDL_MOUSEBUTTONUP:
#if TRANSFORM_MOUSE_Y
      event.button.y = win_h-1-event.button.y;
#endif
      event_mouse_up(event);
      break;
    case SDL_MOUSEMOTION:
#if TRANSFORM_MOUSE_Y
      event.motion.y = win_h-1-event.motion.y;
#endif
      event_mouse_motion(event);
      break;
    case SDL_ACTIVEEVENT:
      event_active(event);
      break;
    case SDL_VIDEORESIZE:
      event_resize(event);
      break;
    case SDL_VIDEOEXPOSE:
      event_expose(event);
      break;
    default:
      break;
    }
  }
}
//===================================================================

void GfxClient::event_key_down(SDL_Event &event)
{
  assert(event.key.type == SDL_KEYDOWN);

  ObjAction *c = gui_obj->action("event_key_down");
  if (c) {
    Vector<GameObj*> objs;
    objs.push_back(gui_obj);
    
    Vector<sint4> key_data;
    key_data.push_back(event.key.keysym.sym);
    key_data.push_back(event.key.keysym.mod);
    key_data.push_back(last_mouse_x);
    key_data.push_back(last_mouse_y);

    sint4 result = c->Act(objs, key_data);
    
    if (result) return;
  }

  switch (event.key.keysym.sym) {
  case SDLK_LCTRL:
    ctrl_down = true;
    break;
  case SDLK_LSHIFT:
    shift_down = true;
    break;
  case SDLK_ESCAPE:
    quit = true;
    break;
  case SDLK_LEFT:
    views[active_view].move_x(-KEYB_OFF * max_dim * 0.75); views[active_view].write();
    break;
  case SDLK_RIGHT:
    views[active_view].move_x(KEYB_OFF * max_dim * 0.75); views[active_view].write();
    break;
  case SDLK_UP:
    views[active_view].move_y(KEYB_OFF * max_dim * 0.75); views[active_view].write();
    break;
  case SDLK_DOWN:
    views[active_view].move_y(-KEYB_OFF * max_dim * 0.75); views[active_view].write();
    break;
  case ']':
    views[active_view].zoom(-ZOOM_OFF); views[active_view].write();
    break;
  case '[':
    views[active_view].zoom(ZOOM_OFF); views[active_view].write();
    break;
  case SDLK_COMMA:
    views[active_view].rotate_y(-ROTATE_OFF); views[active_view].write();
    break;
  case SDLK_PERIOD:
    views[active_view].rotate_y(ROTATE_OFF); views[active_view].write();
    break;
  case 'z':
    views[active_view].rotate_x(-ROTATE_OFF); views[active_view].write();
    break;
  case 'x':
    views[active_view].rotate_x(ROTATE_OFF); views[active_view].write();
    break;
  case 'q':
    GfxView::change_angle(lx_angle, -LROTATE_OFF);
    cout << "lxa=" << lx_angle << " lya=" << ly_angle << endl;
    break;
  case 'w':
    GfxView::change_angle(lx_angle, LROTATE_OFF);
    cout << "lxa=" << lx_angle << " lza=" << ly_angle << endl;    
    break;
  case 'e':
    GfxView::change_angle(ly_angle, -LROTATE_OFF);
    cout << "lxa=" << lx_angle << " lza=" << ly_angle << endl;    
    break;
  case 'r':
    GfxView::change_angle(ly_angle, LROTATE_OFF);
    cout << "lxa=" << lx_angle << " lza=" << ly_angle << endl;    
    break;


  case SDLK_KP1:
  case SDLK_KP2:
  case SDLK_KP3:
  case SDLK_KP4:
  case SDLK_KP6:
  case SDLK_KP7:
  case SDLK_KP8:
  case SDLK_KP9:
    {
      sint4 dx = 0;
      sint4 dy = 0;
      if (event.key.keysym.sym % 3 == SDLK_KP1 % 3) dx = -160;
      if (event.key.keysym.sym % 3 == SDLK_KP3 % 3) dx = +160;
      if (event.key.keysym.sym <= SDLK_KP3) dy = +160;
      if (event.key.keysym.sym >= SDLK_KP7) dy = -160;
      
      FORALL (sel_objs, i) {
        Vector<sint4> params;
        GameObj *gob = (*i)->gob;
        
        sint4 x = gob->sod.x + dx;
        sint4 y = gob->sod.y + dy;
        params.push_back(x);
        params.push_back(y);
        gob->set_action("move", params);
      }
    }
    break;

  case SDLK_KP5:
    FORALL (sel_objs, i) {
      Vector<sint4> params;
      GameObj *gob = (*i)->gob;
      gob->set_action("stop", params);
    }
    break;

  case 'p': perspective_mode = !perspective_mode; break;
  case 'f': wireframe = !wireframe; break;
//  case ' ': saveterrain();break;
    
  default:
    break;
  }
}


//===================================================================

void GfxClient::event_key_up(SDL_Event &event)
{
  assert(event.key.type == SDL_KEYUP);
  
  ObjAction *c = gui_obj->action("event_key_up");

  if (c) {
    Vector<GameObj*> objs;
    objs.push_back(gui_obj);
    
    Vector<sint4> key_data;
    key_data.push_back(event.key.keysym.sym);
    key_data.push_back(event.key.keysym.mod);
    key_data.push_back(last_mouse_x);
    key_data.push_back(last_mouse_y);
    
    sint4 result = c->Act(objs, key_data);
    
    if (result) return;
  }
  
  switch (event.key.keysym.sym) {
  case SDLK_LCTRL:
    ctrl_down = false;
    break;
  case SDLK_LSHIFT:
    shift_down = false;
    break;
  default:
    break;
  }
}

//===================================================================

void GfxClient::event_mouse_up(SDL_Event &event)
{
  last_mouse_x = event.button.x;
  last_mouse_y = event.button.y;
  
  uint1 button = event.button.button;

  // call the GUI mouse_up event handler
  // Sending the event to the GUI before the rest of this function shouldn't cause any side-effects,
  // the only actions that should care enough to block the event should have already trapped mouse_down.
  ObjAction *c = gui_obj->action("event_mouse_up");
  if (c) {
    Vector<GameObj*> objs;
    objs.push_back(gui_obj);
    
    Vector<sint4> click_data;
    click_data.push_back(last_mouse_x);
    click_data.push_back(last_mouse_y);
    click_data.push_back(button);
    
    sint4 result = c->Act(objs, click_data);
    
    if (result) return;
  }

  // ignore over minimap
  if (inside(last_mouse_x, last_mouse_y, mm_rect) && !net_active) return;
  
  if (button == 1) {

    if (cur_click_state == JUST_CLICKED) {
      cur_click_state = NORMAL;
      return;
    }
    if (cur_click_state != NORMAL) return;

    vec3 inter;
    Vector<IdZ> ids;
    SDL_Rect r;
    sint4 dx = abs(sint4(last_mouse_x) - sint4(button1_down_x));
    sint4 dy = abs(sint4(last_mouse_y) - sint4(button1_down_y));
    bool net_select = net_active && (dx > 0 && dy > 0 && (dx >= 5 || dy >= 5));

    if (net_select) {

      compute_rect(button1_down_x, button1_down_y, last_mouse_x, last_mouse_y, r);
      
    } else {

      if (!inside(last_mouse_x, last_mouse_y, views[active_view].viewport)) {
        net_active = false;
        return;
      }

      sint4 w = 2; // > 0

      SDL_Rect r0 = { event.button.x - w/2, event.button.y - w/2, w, w };
      r = r0;

    }

    net_active = false;

    rect_intersection(views[active_view].viewport, r, r);
    
    pick(views[active_view], r, inter, ids);

    if (!ids.empty()) {

      //      cout << ids.size() << " hit(s)" << endl;
      
      if (!shift_down) {

        // new selection if shift key not pressed
        FORALL (sel_objs, i) (*i)->selected = false;
        sel_objs.clear();
      }
      
      FORALL (ids, i) {

        GfxObject *gob = GfxObject::id2obj[(*i).id >> GfxObject::SUB_ID_LOG];
        if (gob == 0) { REM("foo"); continue; }
        if (dynamic_cast<GfxTerrainTile*>(gob)) continue;
        
        set<GfxObject*>::iterator j = sel_objs.find(gob);

        if (net_select || j == sel_objs.end()) {
          
          // add object to selection
          
          if (gob->gob) { // check if this is the main graphics object, not a component
            gob->selected = true;
            sel_objs.insert(gob);
          }
          
        } else {
          // remove object from selection
          gob->selected = false;
          sel_objs.erase(j);
        }
      }
      
      update_selected();


      if (sel_objs.size() > 1) {

        // remove other players' objects
        sint4 me = game->get_client_player();
        Vector<GfxObject*> del;
        FORALL (sel_objs, i) if ((*i)->gob->sod.owner != me) del.push_back(*i);
        FORALL (del, i) { (*i)->selected = false; sel_objs.erase(*i); }
        update_selected();
      }
    } else if (!net_select) {
      FORALL (sel_objs, i) (*i)->selected = false;
      sel_objs.clear();
      update_selected();
    }

    //    update_selected();
  }

  update_active_view();
}

//===================================================================

void GfxClient::event_mouse_down(SDL_Event &event)
{
  last_mouse_x = event.button.x;
  last_mouse_y = event.button.y;

  uint1 button = event.button.button;

  // call the GUI mouse_down event handler
  ObjAction *c = gui_obj->action("event_mouse_down");
  if (c) {
    Vector<GameObj*> objs;
    objs.push_back(gui_obj);
    
    Vector<sint4> click_data;
    click_data.push_back(last_mouse_x);
    click_data.push_back(last_mouse_y);
    click_data.push_back(button);
    
    sint4 result = c->Act(objs, click_data);
    
    if (result) {
      cur_click_state = JUST_CLICKED;
      return;
    }
  }


  if (button == 1) {

    button1_down_x = last_mouse_x;
    button1_down_y = last_mouse_y;
    net_active = false;

  }

  // clicks on minimap

  if (inside(last_mouse_x, last_mouse_y, mm_rect)) {

    if (button == 1) {

      // position active view

      real4 x, y;

      mm_to_playfield(last_mouse_x, last_mouse_y, x, y);
      views[active_view].x_off = -x;
      views[active_view].z_off = -y;

      cur_click_state = JUST_CLICKED;
      return;

    } else if (button == 3) {

      // send objects there

      if (sel_objs.size() > 0) {
        real4 pf_x, pf_y;
        mm_to_playfield(last_mouse_x, last_mouse_y, pf_x, pf_y);
        real4 off_x, off_y;
        if (game->is_rotated()) {

          off_x = off_y = (game->get_map().get_width()+game->get_map().get_height())*0.25;

        } else if (!rotate_minimap) {

          off_x = game->get_map().get_width()*0.5;
          off_y = game->get_map().get_height()*0.5;

        } else {
          off_x = game->get_map().get_width()*0.5;
          off_y = game->get_map().get_height()*0.5;
        }
        move_objects(pf_x+off_x, pf_y+off_y);
        return;
      }
    }
  }
  
  cur_click_state = NORMAL;

  if (!inside(event.button.x, event.button.y, views_rect)) return;

  if (button == 3) {

    // generate move actions for selected objects

    // compute target location from mouse coordinates

    sint4 w = 2;
    SDL_Rect r = { event.button.x - w/2, event.button.y-w/2, w, w };
    vec3 inter;
    Vector<IdZ> ids;

    cout << "SEND?" << endl;
    
    if (sel_objs.size() > 0 && pick(views[active_view], r, inter, ids) > 0) {
      move_objects(inter.x, inter.z);
    }
  }

  if (button == 4) { // wheel up
    views[active_view].zoom(-ZOOM_OFF);
  }
  if (button == 5) { // wheel down
    views[active_view].zoom(ZOOM_OFF);
  }

  update_active_view();
}

//===================================================================

void GfxClient::event_mouse_motion(SDL_Event &event)
{
  last_mouse_x = event.motion.x; last_mouse_y = event.motion.y;

  ///////////////////////////////////////////
  // call the GUI mouse_motion event handler
  ObjAction *c = gui_obj->action("event_mouse_motion");
  if (c) {
    Vector<GameObj*> objs;
    objs.push_back(gui_obj);
    
    Vector<sint4> motion_data;
    motion_data.push_back(last_mouse_x);
    motion_data.push_back(last_mouse_y);
    motion_data.push_back(event.motion.state);
    
    sint4 result = c->Act(objs, motion_data);
  
    if (result) return;
  }
  ///////////////////////////////////////////

  // motion on minimap
  if (inside(last_mouse_x, last_mouse_y, mm_rect) && !net_active) {

    if (event.motion.state & SDL_BUTTON(1)) {
      real4 x, y;
      mm_to_playfield(last_mouse_x, last_mouse_y, x, y);
      views[active_view].x_off = -x;
      views[active_view].z_off = -y;
      return;
    }
  }
  
  uint1 btn_state = event.motion.state;

  real4 dx = event.motion.xrel;
  real4 dy = event.motion.yrel;

#if (! TRANSFORM_MOUSE_Y)
  dy = -dy;
#endif

  if (btn_state & SDL_BUTTON(2)) { // middle mouse button

    if (ctrl_down) {
      views[active_view].rotate_y(dx * 0.3);
      views[active_view].rotate_x(dy * 0.3);
    } else if (shift_down) {
      views[active_view].zoom(dy * 0.01);
    } else {
      views[active_view].move_x(-dx * max_dim * 0.75);
      views[active_view].move_y(dy * max_dim * 0.75);
    }

  } else if (btn_state & SDL_BUTTON(1)) { // left mouse button

    if (cur_click_state == NORMAL)
      net_active = true;

  } else {

    update_active_view();

  }

}

//===================================================================

void GfxClient::event_resize(SDL_Event &event)
{
  win_w = event.resize.w;
  win_h = event.resize.h;

  SDL_Surface *surface = SDL_SetVideoMode(win_w, win_h, win_bpp, win_flags);
  if (!surface) ERR("Could not get a surface after resize");


  views_rect.x = views_rect.y = 0;
  views_rect.w = win_w;
  views_rect.h = win_h;

  ObjAction *c = gui_obj->action("event_resize");
  if (c) {
    Vector<GameObj*> objs;
    objs.push_back(gui_obj);
    
    Vector<sint4> resize_data;
    resize_data.push_back(win_w);
    resize_data.push_back(win_h);
    
    sint4 result = c->Act(objs, resize_data);
    if (result) return;

  } else {
    update_layout(prev_layout_name, true);
  }

  draw();
}

//===================================================================

void GfxClient::event_expose(SDL_Event &)
{
  //  update_layout(prev_layout_name, true);
  //  draw();
}

//===================================================================

void GfxClient::event_active(SDL_Event &) {}

//===================================================================

