/*
    (c) Igor Durdanovic, igord@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//: System.H (bof) (c) Igor Durdanovic

#ifndef System_H
#define System_H

#include <sys/time.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/vfs.h>
#include <sys/sysinfo.h>
#include <sys/utsname.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <sstream>

//#include "string.H"
#include "Global.H"

#define RLIMIT_RESOURCE __rlimit_resource
#define SOCKLEN_T socklen_t

class System
{
public:

  enum { IP_PROTO = 0, TCP_PROTO = 6 };

  static sint8 real_time(); // in micro sec
  static sint8 user_time(); // in micro sec

  static sint4 clock();          // curr time in sec since 19..
  static ccptr dtime(sint4 c); // converts clock() into string
  static ccptr stime(sint4 c); // converts clock() into SHORT string
  static ccptr ftime(sint4 c); // converts clock() into FILE-NAME string
  static ccptr ggftime(sint4 c); // converts clock() into GGF DT[] string

  static void free   (vptr Ptr);
  static vptr malloc (sint4 Len, bool report=true);
  static vptr realloc(vptr Ptr, sint4 Len, bool report=true);

  static sint4 chroot(ccptr Path, bool report=true);
  static sint4 chdir (ccptr Path, bool report=true);
  static void exit  (sint4 RetCode);

  static bool  file_legal(ccptr Path); // no .. or //
  static bool  file_exist(ccptr Path);
  static sint4 file_size (ccptr Path, bool report=true);
  static sint4 file_open (ccptr Path, sint4 Flags, mode_t Mode, bool report=true);
  static sint4 file_sync (sint4 Desc, bool report=true);
  static sint4 file_close(sint4 Desc, bool report=true);
  static sint4 file_write(sint4 Desc, cvptr Buff, sint4 Count, bool report=true);
  static sint4 file_read (sint4 Desc,  vptr Buff, sint4 Count, bool report=true);

  static struct statfs*  fs_stat(ccptr Path, bool report=true);
  static struct sysinfo* sysinfo(bool report=true);
  static struct utsname* uname(bool report=true);
  
  static sint4 sock_open    (sint4 Dom, sint4 Type, sint4 Prot, bool report=true);
  static sint4 sock_connect (ccptr Host, uint4 Port, bool report=true);
  static sint4 sock_accept  (sint4 Sock, bool report=true);
  static sint4 sock_bind    (sint4 Sock, sint4 Dom, uint4 Port, bool report=true);
  static sint4 sock_listen  (sint4 Sock, sint4 OnHold, bool report=true);
  static sint4 sock_shutdown(sint4 Sock, sint4 How, bool report=true);
  static sint4 sock_close   (sint4 Sock, bool report=true);
  static sint4 sock_read (sint4 Desc, vptr Buff, sint4 Count, bool report=true); 
  static sint4 sock_write(sint4 Desc, vptr Buff, sint4 Count, bool report=true); 
  
  static ccptr sock_stat(sint4 Sock);

  static sint4 rlimit_get(RLIMIT_RESOURCE What, struct rlimit& r, bool report=true);
  static sint4 rlimit_set(RLIMIT_RESOURCE What, struct rlimit& r, bool report=true);
  static sint4 rusage_get(struct rusage& ruse, bool report=true);

  static sint4 select_desc(sint4 N, fd_set& Read, fd_set& Write, fd_set& Excp,
			   struct timeval* T, bool report=true);

  static sint4 fcntl(sint4 Desc, sint4 Cmd, bool report=true);
  static sint4 fcntl_arg(sint4 Desc, sint4 Cmd, long Arg, bool report=true);

  static sint4 getsockopt(sint4 Sock, sint4 Level, sint4 Opt, vptr Val,
			  SOCKLEN_T* Len, bool report=true);
  static sint4 setsockopt(sint4 Sock, sint4 Level, sint4 Opt, vptr Val,
			  SOCKLEN_T  Len, bool report=true);
  static void  txtsockopt(sint4 Sock, ostream& os);

  static sint4 getpeername(sint4 Sock, struct sockaddr* sa,
			   SOCKLEN_T* Len, bool report=true);
  static sint4 getsockname(sint4 Sock, struct sockaddr* sa,
			   SOCKLEN_T* Len, bool report=true);

  static struct hostent* gethostbyaddr(const struct sockaddr* Addr,
				       sint4 Len, sint4 Type, bool report=true);
  static sint4           gethostname  (string& host_name, bool report=true);
  static sint4           getsockhost  (string& host_name, string& host_ip,
				       sint4& Port, sint4 Sock);

private:
  static bool shutdown;
};

#endif

//: System.H (eof) (c) Igor
