#include "Commander.H"
#include "Global.H"
#include "GameStateModule.H"
#include "GameChanges.H"
#include "Game.H"
#include "ClientModule.H"
#include "History.H"
#include "ModuleSet.H"
#include "TerrainBase.H"
#include "GatherCommander.H"

#define TEST(x)

using std::cerr;
using std::cout;
using std::endl;

class General : public Commander, public ClientModule
{
private:
    Vector<GameObj*> unitVector;
    bool receivedFirstView;
    GameStateModule &gsm;
    sint4 minerals;
    TerrainBase::Loc loc;
    GatherCommander *gatherCommander;
public:
    General(ModuleSet& ms, sint4 updateTime_=25) : 
        Commander(ms, updateTime_),
        receivedFirstView(false),
        gsm(ms.get_game_state()),
        gatherCommander(0)
    {
        gsm.add_handler(this);
        ms.get_history().trackBlueprint("mineral");
    }

    ~General()
    {
        gsm.remove_handler(this);
    }

    void exec() 
    {
        if (!receivedFirstView)
        {
            return;
        }
        if (gatherCommander == 0)
        {
            gatherCommander = new GatherCommander(this,
                                                  unitVector,
                                                  loc,
                                                  800,
                                                  1);
            unitVector.clear();
        }
    }

    void onReceiveUnit(GameObj * unit, Watcher::WatchID)
    {
        TEST(std::cerr << "General got " << unit->bp_name() << std::endl);

        unitVector.push_back(unit);
    }

    void onLoseUnit(GameObj *)
    {
       
    }

    bool onEvent(const Event &event) 
    {
        if (event.get_what() == GameStateModule::VIEW_MSG) 
        {
            if (!receivedFirstView) 
            {
                TEST(std::cerr << "Inserting units" << std::endl);
                insertUnits(gsm.get_changes().new_objs);
            }
            this->update();
            return true;
        }
        return false;
    }

    void insertUnits(const Vector<ScriptObj::WPtr>& objects)
    {
        TEST(cerr <<"Add units to Commander control" << endl);
        const sint4 me = gsm.get_game().get_client_player();
        
        FORALL_DYN(objects, it) 
        {
            GameObj* gob = (*it)->get_GameObj();
            if( !gob )
                continue;
            
            if( gob->bp_name() == "worker" && *(gob->sod.owner) == me ) 
            {
                TEST(std::cerr << "Inserting " << gob->bp_name() << std::endl);
                addUnit(gob);
            }
            if( gob->bp_name() == "controlCenter" && *(gob->sod.owner) == me ) 
            {
                sint4 x, y;
                gob->get_center(x,y);
                TerrainBase::Loc location(x,y);
                loc=location;
                TEST(std::cerr << "Inserting " << gob->bp_name() << std::endl);
                addUnit(gob);
                cout << x <<" " << y <<endl;
            }            
        }
        receivedFirstView = true;
    }
};
