/* 
 * Commander Class
 * Abstract Task Hierchary
 * Aaron Young
 * Ryan Neighbour
*/
#ifndef COMMANDER_H
#define COMMANDER_H
#include<vector>
#include "EventHandler.H"
#include "GameObj.H"
#include <ctime>
#include "TerrainModule.H"

using namespace std;

class Commander: public EventHandler
{
 protected:
  Commander* const myCommander;
  TerrainBase::Loc centrePoint;
  vector<Commander*> commVector;
  vector<GameObj*> unitVector;
  int radius, updateCounter, commanderID;

 

  /* Constructor */ 
  Commander(Commander* const mySuperior, TerrainBase::Loc centre, vector<GameObj*> units, int r):
    myCommander(mySuperior), centrePoint(centre), unitVector(units), radius(r)
  { 
    srand((unsigned)time(0));
    updateCounter=rand()%25+1;
    commanderID=EventFactory::new_who();
  }




  /* Main method to be extended by Concrete Class */
  virtual void exec() = 0;
  
  



  /* Pass this Commander a message */
  virtual bool handle_event(const Event &event) = 0;

  

  /* Recursively calls the commander tree to perform exec() */
  void update() {
    if (updateCounter==0) {
      exec();
      updateCounter=rand()%25+1;
    }
    vector<Commander*>::iterator it=commVector.begin();
    while (it!=commVector.end()) {
      Commander* c=*it;
      c->update();
      it++;
    }
    updateCounter--;
  }
  
  

  /* destroy commander, send units to superior */
  virtual ~Commander() {
    vector<Commander*>::iterator commIt=commVector.begin();
    while (commIt!=commVector.end()) {
      Commander* c=*commIt;
      delete c;
      commIt++;
   }
    vector<GameObj*>::iterator unitIt=unitVector.begin();
    while (unitIt!=unitVector.end()) myCommander->unitVector.push_back(*unitIt);
  }

};

#endif //COMMANDER_H
