// $Id: SI_Player.cpp,v 1.1 2007/04/05 21:13:47 kalf Exp $



#include "SI_Group.h"
#include "SI_Order.h"
#include "SI_MapObject.h"
#include "SI_Base.h"

#include "SI_SimulationState.h"

#include "SI_Player.h"
namespace Simul{

  Player::Player(uint4 playerID){
    this->playerID = playerID;
    currentStrategy = NULL;
  }

  Player::~Player(){
    // get rid of all the groups
    FORALL (groups, it) delete *it;
    groups.clear();

    // then get rid of all the bases
    FORALL (bases,it) delete *it;
    bases.clear();

    // finally, get rid of this player's strategy
    delete currentStrategy;
    currentStrategy = NULL;

  }


  Player::Player(const Player& p){
    playerID = p.playerID;

    FORALL(p.groups, it) {
      Group* newGroup = (*it)->clone();
      groups.push_back(newGroup);
    }

    FORALL(p.bases, it){
      Base* newBase = (*it)->clone();
      bases.push_back(newBase);
    }

    currentStrategy = NULL;
  }

  Player* Player::clone()
  {
    return new Player(*this);

  }

  void Player::addGroup(Group* newGroup){
    if(newGroup == NULL){
      ERR("trying to add NULL group");
    }else{
      groups.push_back(newGroup);
    }
  }

  void Player::addBase(Base* newBase){
    if(newBase == NULL){
      ERR("trying to add NULL base");
    }else{
      bases.push_back(newBase);
    }
  }
/*
  void Player::drawObjects(){
    // draw bases first, so groups will always be drawn over them
    FORALL(bases,it)  (it->second)->draw();
    FORALL(groups,it) (it->second)->draw();
  }

  void Player::drawObjectsInSelectionMode(){
    // FORALL(bases,it)  (it->second)->draw();
    FORALL(groups,it) (it->second)->drawInSelectionMode();
  }
*/

  void Player::executeOrders(real8 timeElapsed){
    FORALL(groups,it){
      Group* group = *it;
      if (group->getNumUnits() > 0){
	group->executeOrders(timeElapsed);
      }
    }
  }

  void Player::updateOrders(SimulationState& state){
    FORALL(groups,it){
      Group* group = *it;

      if(group->getOrder() == NULL) continue;

      if (group->getOrder()->isCompleted()) {
	group->executeOrderAction(state);
      }
    }
  }


  void Player::setStrategy(Task* strategy){
    delete currentStrategy;

    currentStrategy = strategy;
  }
}
