// $Id: SI_Strategies.cpp,v 1.4 2007/04/09 19:53:32 kalf Exp $


#include "SI_Strategies.h"

#include "SI_SimulationAI.H"
#include "SI_SimulationState.h"

#include "GameStateModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameObj.H"
//#include "GfxModule.H"
#include "TerrainModule.H"
#include "TerrainClassifier.H"

namespace Simul {
 //new FormUpMove(this,objects,function,name,0.0,new MoveStruct(center_loc));
#define DEFINE_TASK(name)						\
  name::name(uint4 playerID,SimulationState* state,ObjectVector& origObjects,	\
	     TaskCreateFunctionPtr f,std::string n,real8 timeout,MoveStruct* extraData) : \
    Task(playerID,state,origObjects,f,n,timeout,extraData){	\
    exec();								\
  }									\
    name::name(Task* parent,ObjectVector& origObjects,TaskCreateFunctionPtr function,std::string n, \
	       real8 timeout,MoveStruct* extraData) :	\
      Task(parent,origObjects,function,n,timeout,extraData){ \
      exec();								\
    }									\
    name::name(const name& t) : Task(t){}				\
    Task* name::clone(SimulationState* state,Task* newParent){		\
      name* newTask = new name(*this);					\
      newTask->parent = newParent;					\
      newTask->state = state;						\
      FORALL(children,it){						\
	newTask->children.push_back((*it)->clone(state,newTask));	\
      }									\
      return newTask;							\
    }									\
    bool name::exec()


  DEFINE_TASK(SpreadAttack) {
    ObjectVector temp;

    switch(lineNum) {
    case 0:
      // need a copy of this vector, since the constructor for tasks modifies the original by removing objects from it
      temp = objects;

      FORALL(temp, it) {
        ObjectVector object;
	object.push_back(*it);

        GameObj* nearestBase = state->getNearestEnemyBaseOfGroup(playerID,object);
	if(nearestBase != NULL) {
	  new Attack(this,object,function,name,0.0,new MoveStruct(nearestBase));
	}
      }
      default:
        return false;
    }
    return false;
  }

  DEFINE_TASK(MassAttack) {
    // ObjectVector temp;
    switch(lineNum){

    case 0:
      if(state->smartJoin){
        new SmartJoin(this, objects, function, name, 0.0);
      }else{
        new Join(this, objects, function, name, 0.0);
      }

      ++lineNum;
      break;
    case 1:{
    // need a copy of this vector, since the constructor for tasks modifies the original by removing objects from it
	GameObj* nearestBase = state->getNearestEnemyBaseOfGroup(playerID,objects);
        //temp = objects;

	if(nearestBase != NULL){
	    new Attack(this,objects,function,name,0.0,new MoveStruct(nearestBase));
	}
        break;
      }
      default:
        return false;
    }
    return false;
  }

  DEFINE_TASK(Join) {
    assert(extraData == NULL); // extradata should be NULL for this task

    switch(lineNum) {
    case 0:{
      GameObj* centerObject = state->getNearestObject(objects, state->getCenterOfObjects(objects));
      assert(centerObject != NULL && centerObject->get_shape() == Object::CIRCLE);// takje this assert OUT later
      TerrainBase::Loc center_loc;
      centerObject->get_center(center_loc.x, center_loc.y);
      
      if(state->findspot){
	GameStateModule & gsm = state->simAI->get_gsm();
	center_loc = find_a_spot(center_loc, gsm);
      }

       new MoveToLocation(this,objects,function,name,0.0,new MoveStruct(center_loc));
      //new FormUpMove(this,objects,function,name,0.0,new MoveStruct(center_loc));

      ++lineNum;
      break;
    }
    default:
      return false;
    }
    return false;
  }


  DEFINE_TASK(JoinRelative) {
      
    switch(lineNum) {
    case 0:{
      // first, find the final location we will join at.. in this case, it MUST be the location, since units will move.
      // once influence maps are in, it should find a safe spot to join, instead of just the average position of all
      // the units
      GameObj* centerObject = state->getNearestObject(objects, state->getCenterOfObjects(objects));
      assert(centerObject != NULL);// take this assert OUT later
      TerrainBase::Loc  center_loc;
      centerObject->get_center(center_loc.x, center_loc.y);

      if(state->findspot)
	center_loc = find_a_spot(center_loc, state->simAI->get_gsm());
      
      // now make groups of X units(in this case, 5. kind of a dumb way to do it... basically, just take one unit, find X-1
      // closest units, finding their average position, and then joining at the unit closest to that position
      fprintf(stderr,"objects size = %f\n",objects.size());
      while(objects.size() > 0){
	ObjectVector unitsToJoin;

	fprintf(stderr,"objects size = %f\n",objects.size());

        state->getNearestObjectsToObjectArb(unitsToJoin,objects,objects[0],playerID);

	////////////
	sint4 aveX = 0;
	sint4 aveY = 0;
	sint4 x = 0;
	sint4 y = 0;

	FORALL(unitsToJoin,it){
	  
	  GameObj* currObj = *it;
	  currObj->get_center(x,y);
	  aveX += x;
	  aveY += y;

	}
	center_loc.x = aveX / unitsToJoin.size();
	center_loc.y = aveY / unitsToJoin.size();


	center_loc = find_a_spot(center_loc, state->simAI->get_gsm());
	fprintf(stderr,"am i crashing after here \n");

	//////////////

        //  new JoinAndMoveToObj(this,unitsToJoin,function,name,0.0, new MoveStruct(center_loc));
	//new FormUpMove(this,unitsToJoin,function,name,0.0, new MoveStruct(center_loc), new MoveStruct(target),1);
	new FormUpMove(this,unitsToJoin,function,name,0.0, new MoveStruct(center_loc));
      }


      ++lineNum;
      break;
    }
    default:
        fprintf(stderr,"relative false default");
      return false;
    }
    fprintf(stderr,"relative false default");
    return false;
  }







  DEFINE_TASK(JoinRelative2) {
    assert(extraData == NULL); // extradata should be NULL for this task

    switch(lineNum) {
    case 0:{
      GameObj* centerObject = state->getNearestObject(objects, state->getCenterOfObjects(objects));
      assert(centerObject != NULL && centerObject->get_shape() == Object::CIRCLE);// takje this assert OUT later
      TerrainBase::Loc center_loc;
      centerObject->get_center(center_loc.x, center_loc.y);
      
      if(state->findspot){
	GameStateModule & gsm = state->simAI->get_gsm();
	center_loc = find_a_spot(center_loc, gsm);
      }

       new MoveToLocation(this,objects,function,name,0.0,new MoveStruct(center_loc));
      //new FormUpMove(this,objects,function,name,0.0,new MoveStruct(center_loc));



      ++lineNum;
      break;
    }
    default:
      return false;
    }
    return false;
  }






  DEFINE_TASK(Attack) {
    assert(extraData != NULL);

    switch(lineNum) {
    case 0:
      // everyone should arrive at the same time
      FORALL(objects,it) {
        TerrainBase::Task task(TerrainBase::Goal::OBJ, TerrainBase::Goal::TOUCH,
                               TerrainBase::Task::INDIVIDUAL);


        task.goal.obj = extraData->obj;

        GameObj* currObj = *it;

        ExtraObjectData newData;
        newData.target = extraData->obj;
        newData.state = ExtraObjectData::ATTACK;

        Player* currPlayer = state->getPlayer(playerID);

        currPlayer->getOrtsUnits()[currObj] = newData;

        task.objs.insert(currObj);
        state->simAI->doPath(task);
      }
      ++lineNum;
      break;
    default:
      return false;
    }
    return false;
  }

  ////////////////

  /*
    it is curretnly sloppy and some these varibles should be set by what calling it will make proper changes later and optimize the code as well if time permits.
   */
  DEFINE_TASK(FormUpMove){
    assert(extraData != NULL);
    assert(extraData->obj == NULL); // should not have an object as a target
    sint4 count = 0;
    sint4 row = 1;
    sint4 colSize = 1;
    sint4 col = 0;
    bool flag = false;
    sint4 formFlag = 0;
    real4 hyp = 0;
    real4 x = 0;
    real4 y = 0;
    sint4 posX = extraData -> pos.x;
    sint4 posY = extraData -> pos.y;
    real4 oldAngle = 0;
    real4 offset = 0;


   
    // sint4 targetX = extraData2->pos.x;
    // sint4 targetY = extraData2->pos.y;

    sint4 scale;
    sint4 posXori = posX;
    sint4 posYori = posY;


    sint4 groups = 0;
    sint4 groupCount = 0;

    ///////////////////formation info /////
 
    real4 angConst = M_PI;
    real4 facingDir = M_PI*2;//atan((targetY - posY)/(targetX - posX));
    real4 angle = (angConst) -1*facingDir;
    sint4 spacing = 25;
    sint4 grOffset = 0;

    groups = 1;

    formFlag = 1;//form;
    row = 1;
    colSize = row; 
    if(formFlag == 1)
    	colSize = 3;
    /////////////////////////////////

    switch(lineNum) {
    case 0:
      // everyone should arrive at the same time
      FORALL(objects,it) {
        TerrainBase::Task task(TerrainBase::Goal::LOCATION, TerrainBase::Goal::VICINITY,
                               TerrainBase::Task::INDIVIDUAL);

        GameObj* currObj = *it;


	//////////////////////////////////

    	if (groupCount % 2 == 0) {
	  scale = groupCount/2; 
    	}else{
	  scale = (groupCount+1)/2;
    	}


    	posX = posXori + (sint4)(cos(angle+M_PI/2)*scale*pow((real8)-1, (real8)groupCount)*grOffset);
    	posY = posYori + (sint4)(sin(angle+M_PI/2)*scale*pow((real8)-1, (real8)groupCount)*grOffset);
    	


	/////////////////////////////////
    	if (count % 3 == 2){ 
	  if(flag == false){
	    flag = true;
	    
	    hyp = pow((real8)(spacing*row),(real8)2) + pow((real8)0, (real8)2);
	    hyp = sqrt(hyp);
	    
	    extraData->pos.x = posX + (sint4)(hyp*cos(angle));
	    extraData->pos.y = posY + (sint4)(hyp*sin(angle));
	    ////////////////////////


	    if (col >= colSize){
	      groupCount++;
	      if(groupCount == groups){
		row++;
		groupCount = 0; 
		posX = posXori;
		posY = posYori;
    		
		if (formFlag == 0){
		  colSize = colSize + 1;
		}
    		
	      }   
	      
	      col = 0;
	      flag = 0;
	    }
	  }
	  else{
	    count++;
	    
	    if (col >= colSize){
	      groupCount++;
	      if(groupCount == groups){
		row++;
		groupCount = 0; 
		posX = posXori;
		posY = posYori;
    		
		if (formFlag == 0){
		  colSize = colSize + 1;
		}
  	      }  
	    }

	    col = 0;
	    flag = 0;


	  }
	}
	//////////////////////////////////

    	if (groupCount % 2 == 0) {
	  scale = groupCount/2; 
    	}else{
	  scale = (groupCount+1)/2;
    	}
    	posX = posXori + (sint4)(cos(angle+M_PI/2)*scale*pow((real8)-1, (real8)groupCount)*90);
    	posY = posYori + (sint4)(sin(angle+M_PI/2)*scale*pow((real8)-1, (real8)groupCount)*90);
    	


	/////////////////////////////////



	if (count % 3 == 0){ 		//1
	  
	  offset = col*spacing;
	  x=(spacing*row);
	  if(formFlag == 0)
	    y=(-1*(spacing*row  - offset));
	  else
	    y=(spacing+offset);
	  
	  hyp = pow(x,2) + pow(y, 2);
	  hyp = sqrt(hyp);
	  oldAngle = atan(y/x);
	  
	  extraData->pos.x = posX + (sint4)hyp*cos(oldAngle+angle);
	  extraData->pos.y = posY + (sint4)hyp*sin(oldAngle+angle);	

	}
	

	//////////////////////////////////

    	if (groupCount % 2 == 0) {
	  scale = groupCount/2; 
    	}else{
	  scale = (groupCount+1)/2;
    	}
    	posX = posXori + (sint4)(cos(angle+M_PI/2)*scale*pow((real8)-1, (real8)groupCount)*90);
    	posY = posYori + (sint4)(sin(angle+M_PI/2)*scale*pow((real8)-1, (real8)groupCount)*90);
    	


	/////////////////////////////////



	if (count % 3 == 1){ 
	  x=(spacing*row);
	  if(formFlag == 0)
	    y=(1*(spacing*row  - offset));
	  else
	    y=(-offset-spacing);
	  
	  hyp = pow(x,2) + pow(y, 2);
	  hyp = sqrt(hyp);
	  oldAngle = atan(y/x);
	  extraData->pos.x = posX + (sint4)hyp*cos(oldAngle+angle);
	  extraData->pos.y = posY + (sint4)hyp*sin(oldAngle+angle);	

	  
	  col++;
	}	
	count++;
	
	/////////////////////////////
        ExtraObjectData newData;
        newData.location = extraData->pos;
        newData.state = ExtraObjectData::MOVE;

        Player* currPlayer = state->getPlayer(playerID);

        currPlayer->getOrtsUnits()[currObj] = newData;

        task.goal.loc = extraData->pos;

        // we can base the distance based on how many units we have...
	task.goal.distance = 60;
	// task.goal.distance = 80;
        //        task.goal.distance = (sint4)(objects.size() * 1.5);
        task.objs.insert(currObj);
        state->simAI->doPath(task);

      }
      ++lineNum;
      break;
    default:
      return false;
    }
    return false;
  }

  ////////////////

  DEFINE_TASK(MoveToLocation) {
    assert(extraData != NULL);
    assert(extraData->obj == NULL); // should not have an object as a target

    switch(lineNum) {
    case 0:
      // everyone should arrive at the same time
      FORALL(objects,it) {
        TerrainBase::Task task(TerrainBase::Goal::LOCATION, TerrainBase::Goal::VICINITY,
                               TerrainBase::Task::INDIVIDUAL);

        GameObj* currObj = *it;

        ExtraObjectData newData;
        newData.location = extraData->pos;
        newData.state = ExtraObjectData::MOVE;

        Player* currPlayer = state->getPlayer(playerID);

        currPlayer->getOrtsUnits()[currObj] = newData;

        task.goal.loc = extraData->pos;

        // we can base the distance based on how many units we have...
        task.goal.distance = 80;
        //        task.goal.distance = (sint4)(objects.size() * 1.5);
        task.objs.insert(currObj);
        state->simAI->doPath(task);

      }
      ++lineNum;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(MoveToObject) {
    assert(extraData != NULL);
    assert(extraData->obj != NULL);

    switch(lineNum) {
    case 0:
	// everyone should arrive at the same time
      FORALL(objects,it) {
        if(*it == extraData->obj){ // the center object does not need to move
          ExtraObjectData newData;
          newData.isDone = true;
          Player* currPlayer = state->getPlayer(playerID);
          currPlayer->getOrtsUnits()[*it] = newData;
          continue;
        }
        TerrainBase::Task task(TerrainBase::Goal::OBJ, TerrainBase::Goal::TOUCH,
                               TerrainBase::Task::INDIVIDUAL);

        ExtraObjectData newData;
        newData.target = extraData->obj;
        newData.state = ExtraObjectData::MOVE;

        Player* currPlayer = state->getPlayer(playerID);

        currPlayer->getOrtsUnits()[*it] = newData;

        task.goal.obj = extraData->obj;
        task.objs.insert(*it);
        state->simAI->doPath(task);
      }
      ++lineNum;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(DefendObject) {
    assert(extraData != NULL);

    switch(lineNum) {
    case 0:
	  // everyone should arrive at the same time
      FORALL(objects,it){
        // this should NEVER happen, since you should not be defending an object that is getting orders
        // to defend itself!
        if(*it == extraData->obj){
          ExtraObjectData newData;
          newData.isDone = true;
          Player* currPlayer = state->getPlayer(playerID);
          currPlayer->getOrtsUnits()[*it] = newData;
          continue;
        }
	TerrainBase::Task task(TerrainBase::Goal::OBJ, TerrainBase::Goal::TOUCH,
			      TerrainBase::Task::INDIVIDUAL);

	ExtraObjectData newData;
	newData.target = extraData->obj;
	newData.state = ExtraObjectData::DEFEND;

	Player* currPlayer = state->getPlayer(playerID);

	currPlayer->getOrtsUnits()[*it] = newData;

	task.goal.obj = extraData->obj;
	task.objs.insert(*it);
	state->simAI->doPath(task);
      }
      ++lineNum;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(HalfUnitsBaseDefence_MassAttack){

    switch(lineNum) {
    case 0:{
      Player* currPlayer = state->getPlayer(playerID);
      // make sure that we have enough units to do anything.. if not, then just don't defend bases
      uint4 numUnits = currPlayer->getOrtsUnits().size();
      uint4 numBases = currPlayer->getOrtsBases().size();
      if(numUnits/2 >= numBases){
        uint4 numUnitsPerBase = (numUnits/2) / numBases;
        FORALL(currPlayer->getOrtsBases(),it){
          ObjectVector unitsToDefendBase;
          state->getNearestObjectsToObject(unitsToDefendBase,objects,it->first,numUnitsPerBase);
          new DefendObject(this,unitsToDefendBase,function,name,0.0,new MoveStruct(it->first));

	}
      }
      // the rest of the units that are not defending go join up and attack!
      new MassAttack(this,objects,function,name,0.0);
      ++lineNum;
      break;
    }
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(TouchObj) {
    assert(extraData != NULL);
    //assert(extraData->obj != NULL);

    switch(lineNum){
    case 0:
      if (extraData->obj != NULL){
	if (extraData->obj->get_max_speed() > 0 && *extraData->obj->sod.owner == (sint4)playerID){
	  TerrainBase::Loc center_loc;
	  extraData->obj->get_center(center_loc.x, center_loc.y);
	  new MoveToLocation(this,objects,function,name,0.0,new MoveStruct(center_loc));
	}
	else{
	  new MoveToObject(this, objects, function,name,0.0, new MoveStruct(extraData->obj));
	}
      }
      ++lineNum;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(JoinAndMoveToObj) {
    assert(extraData != NULL); // extradata should not be NULL for this task

    assert(extraData->pos != TerrainBase::Loc(-1, -1));

    switch(lineNum) {
    case 0:{
      GameObj* centerObject = state->getNearestObject(objects, state->getCenterOfObjects(objects));
      assert(centerObject != NULL);// takje this assert OUT later
      TerrainBase::Loc center_loc;
      centerObject->get_center(center_loc.x, center_loc.y);
    
      assert(centerObject->get_max_speed() > 0); //Objects that have a max speed of 0 will be in the representation, so this would fail
      new MoveToLocation(this,objects,function,name,0.0,new MoveStruct(center_loc));

      ++lineNum;
      break;
    }
    case 1:
      new MoveToLocation(this,objects,function,name,0.0,new MoveStruct(extraData->pos));

      ++lineNum;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(SmartJoin) {

    switch(lineNum) {
    case 0:{
      // first, find the final location we will join at.. in this case, it MUST be the location, since units will move.
      // once influence maps are in, it should find a safe spot to join, instead of just the average position of all
      // the units
      GameObj* centerObject = state->getNearestObject(objects, state->getCenterOfObjects(objects));
      assert(centerObject != NULL);// take this assert OUT later
      TerrainBase::Loc  center_loc;
      centerObject->get_center(center_loc.x, center_loc.y);

      if(state->findspot)
	center_loc = find_a_spot(center_loc, state->simAI->get_gsm());
      
      // now make groups of X units(in this case, 5).  kind of a dumb way to do it... basically, just take one unit, find X-1
      // closest units, finding their average position, and then joining at the unit closest to that position
      while(objects.size() > 0){
	ObjectVector unitsToJoin;

        state->getNearestObjectsToObject(unitsToJoin,objects,objects[0],state->smartJoinSize);
        new JoinAndMoveToObj(this,unitsToJoin,function,name,0.0, new MoveStruct(center_loc));

      }
      ++lineNum;
      break;
    }
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(Hunter) {
    switch (lineNum){
    case 0:

      // make groups of X units(in this case, 5).  kind of a dumb way to do it... basically, just take one unit, find X-1
      // closest units, finding their average position, and then joining at the unit closest to that position
      while(objects.size() > 0){
	ObjectVector unitsToJoin;
        state->getNearestObjectsToObject(unitsToJoin,objects,objects[0],state->groupSize);
        fprintf(stderr,"hunter is called \n");
        new JoinAndHunt(this,unitsToJoin,function,name,0.0);
      }

      lineNum++;
      break;
    case 1:
      //new MassAttack(this, objects, function, name, 0.0);
      new SpreadAttack(this, objects, function, name, 0.0);

      lineNum++;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(Hunter2) {
    switch (lineNum){
    case 0:

	// make groups of X units(in this case, 5).  kind of a dumb way to do it... basically, just take one unit, find X-1
	// closest units, finding their average position, and then joining at the unit closest to that position
      while(objects.size() > 0){
	ObjectVector unitsToJoin;
	state->getNearestObjectsToObject(unitsToJoin,objects,objects[0],state->groupSize);

	new JoinAndHunt2(this,unitsToJoin,function,name,0.0);
      }

      lineNum++;
      break;
    case 1:
//      new MassAttack(this, objects, function, name, 0.0);

      new SpreadAttack(this, objects, function, name, 0.0);
      lineNum++;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(JoinAndHunt){
    assert(extraData == NULL);
 
    switch (lineNum){
    case 0:{
      if(state->smartJoin){
        new SmartJoin(this, objects, function, name, 0.0);
        fprintf(stderr,"martjoin called \n");
      }else{
	// new Join(this, objects, function, name, 0.0);
	new JoinRelative(this, objects, function, name, 0.0);
         fprintf(stderr,"joinrelative called \n");
      }
      lineNum++;
      break;
    }
        
    case 1:{
      if(state->smartJoin){
        new SmartJoin(this, objects, function, name, 0.0);
        fprintf(stderr,"martjoin called \n");
      }else{
	new Join(this, objects, function, name, 0.0);
	fprintf(stderr,"join was called \n");
      }
      lineNum++;
      break;
    }
        
    case 2:{
      GameObj* target = state->getNearestEnemyUnitOfGroup(playerID, objects);
      if (target != NULL){
        if(state->useMoveForAttack){
          new MoveToObject(this, objects, function,name,0.0, new MoveStruct(target));
           fprintf(stderr,"MoveToObject called \n");
        }else{
            fprintf(stderr,"attack called \n");
	   new Attack(this, objects, function,name,0.0, new MoveStruct(target));
	  // TerrainBase::Loc center_loc;
	  // center_loc.x = 200;
	  // center_loc.y = 200;
	   // new FormUpMove(this,objects,function,name,0.0,new MoveStruct(center_loc));


        }
      }
      break;
    }
    default:
        fprintf(stderr,"relative reutnred false default");
        return false;
    }
    fprintf(stderr,"relative reutnred false");
    return false;
  }
  DEFINE_TASK(JoinAndHunt2){
    assert(extraData == NULL);

    switch (lineNum){
    case 0:{
      if(state->smartJoin){
        new SmartJoin(this, objects, function, name, 0.0);
      }else{
        new Join(this, objects, function, name, 0.0);
      }
      lineNum++;
      break;
    }
    case 1:{
      GameObj* target = state->getAverageNearestEnemyUnitOfGroup(playerID, objects);
      if (target != NULL){
        if(state->useMoveForAttack){
	  new MoveToObject(this, objects, function,name,0.0, new MoveStruct(target));
        }else{
	 new Attack(this, objects, function,name,0.0, new MoveStruct(target));
        }
      }
      break;
    }
    default:
      return false;
    }
    return false;
  }


  DEFINE_TASK(JoinSpreadAttack) {
    switch (lineNum){
    case 0:

      while(objects.size() > 0){
	ObjectVector unitsToJoin;
	state->getNearestObjectsToObject(unitsToJoin,objects,objects[0],5);

	new JoinAndAttack(this,unitsToJoin,function,name,0.0);
      }

      lineNum++;
      break;
    case 1:
      new SpreadAttack(this, objects, function, name, 0.0);
      lineNum++;
      break;
    default:
      return false;
    }
    return false;
  }

  DEFINE_TASK(JoinAndAttack){
    assert(extraData == NULL);

    switch (lineNum){
    case 0:{
      if(state->smartJoin){
        new SmartJoin(this, objects, function, name, 0.0);
      }else{
        new Join(this, objects, function, name, 0.0);
      }
      lineNum++;
      break;
    }
    case 1:{
      GameObj* nearestBase = state->getNearestEnemyBaseOfGroup(playerID,objects);
      if (nearestBase != NULL){
	new MoveToObject(this, objects, function,name,0.0, new MoveStruct(nearestBase));
      }
      break;
    }
    default:
      return false;
    }
    return false;
  }

  Task* createSmartJoin(uint4 id,SimulationState* s,ObjectVector& o,real8 timeout) {
    return new SmartJoin(id,s,o,createSmartJoin,"SmartJoin",timeout);
  }
  Task* createSpreadAttack(uint4 id,SimulationState* s,ObjectVector& o,real8 timeout) {
    return new SpreadAttack(id,s,o,createSpreadAttack,"SpreadAttack",timeout);
  }
  Task* createMassAttack(uint4 id,SimulationState* s,ObjectVector& o,real8 timeout) {
    return new MassAttack(id,s,o,createMassAttack,"MassAttack",timeout);
  }
  Task* createHalfUnitsBaseDefence_MassAttack(uint4 id,SimulationState* s,ObjectVector& o,real8 timeout) {
    return new HalfUnitsBaseDefence_MassAttack(id,s,o,createHalfUnitsBaseDefence_MassAttack,
                                               "HalfUnitsBaseDefence_MassAttack",timeout);
  }
  Task* createHunter(uint4 id,SimulationState* s,ObjectVector& o,real8 timeout) {
    return new Hunter(id,s,o,createHunter,"Hunter",timeout);
  }
  Task* createHunter2(uint4 id,SimulationState* s,ObjectVector& o,real8 timeout) {
    return new Hunter2(id,s,o,createHunter2,"Hunter2",timeout);
  }
  Task* createJoinSpreadAttack(uint4 id,SimulationState* s,ObjectVector& o,real8 timeout) {
    return new JoinSpreadAttack(id,s,o,createJoinSpreadAttack,"JoinSpreadAttack",timeout);
  }

 /*********************JAMES' STUFF **************************/  
  TerrainBase::Loc find_a_spot(TerrainBase::Loc center_loc, GameStateModule& gsm){
               
    mapTerrain::TerrainMap terrMap(&gsm.get_game());
    TerrainBase::Loc loc(0,0);
    sint4 standard = 9; sint4 range = 64;
    bool found = terrMap.find_free_loc(center_loc, loc, range, standard);
    range+= 32;
    //std::cout << "found = " << found << ", and meeting spot is (" <<  center_loc.x / 16 << "," << center_loc.y / 16 << ")" << std::endl;
    while(!found  && standard > 4){
      found = terrMap.find_free_loc(center_loc, loc, range, standard--);
      //std::cout << "found = " << found << ", and meeting spot is (" <<  center_loc.x / 16 << "," << center_loc.y / 16 << ")" << std::endl;
    }
    terrMap.print_map();
    return TerrainBase::Loc(loc.x, loc.y);
  }
  /*********************JAMES' STUFF **************************/ 

}



