#include "AttackCommander.H"
#include "TerrainModule.H"

static sint4 CalcDistSq(const TerrainBase::Loc & loc1,
                        const TerrainBase::Loc & loc2)
{
    sint4 dx = loc1.x - loc2.x;
    sint4 dy = loc1.y - loc2.y;
    return dx*dx + dy*dy;
}

AttackCommander::AttackCommander(Commander* const mySuperior, 
                                 Vector<GameObj*> units,
                                 sint4 owner) :
    Commander(mySuperior, 16),
    target(NULL),
    me(owner),
    mode(1)
{
    add_handler(&moduleSet.get_terrain());
    moduleSet.get_terrain().add_handler(this);
    
    add_handler(&moduleSet.get_combat());
    moduleSet.get_combat().add_handler(this);
    
    this->addUnits(units);

    Vector<GameObj::WPtr> wvec;
    FORALL(units, u) wvec.push_back(*u);

    call_handlers( SquadCombatEvent( me, 
                                     SquadCombatAI::ATTACK_FREELANCE, 
                                     wvec ) );
}

void AttackCommander::findCentroid()
{
    center.x = center.y = 0;
    FORALL(unit2state, u2s)
    {
        TerrainBase::Loc pos; 
        u2s->first->get_center(pos.x, pos.y);
        center.x += pos.x; 
        center.y += pos.y;
    }
    center.x /= unit2state.size();
    center.y /= unit2state.size();
}

bool AttackCommander::onEvent (const Event& event) 
{  
    if(event.get_what() == TerrainModule::PATH_SUCCESS) 
    {
        const TerrainModule::PathStatusEvent & e =
            reinterpret_cast<const TerrainModule::PathStatusEvent &>(event);
        GameObj * obj = e.get_obj();
        std::map<GameObj*, UnitState>::iterator it;
        it = unit2state.find(obj);
        if(it != unit2state.end())
        {
            it->second.mode = UnitState::FIGHTING;
        }
        
        return true;
    }
  
    if(event.get_what() == TerrainModule::PATH_FAILURE) 
    {
        const TerrainModule::PathStatusEvent & e =
            reinterpret_cast<const TerrainModule::PathStatusEvent &>(event);
        GameObj * obj = e.get_obj();
        std::map<GameObj*, UnitState>::iterator it;
        it = unit2state.find(obj);
        if(it != unit2state.end())
        {
            it->second.mode = UnitState::IDLE;
        }

        return true;
    }

    return false;
}

/*********************JAMES' STUFF **************************/  
TerrainBase::Loc AttackCommander::find_a_spot(TerrainBase::Loc center_loc)
{ 
  mapTerrain::TerrainMap terrMap(&moduleSet.get_game_state().get_game());
  TerrainBase::Loc loc(0,0);
  sint4 standard = 9; sint4 range = 64;
  bool found = terrMap.find_free_loc(center_loc, loc, range, standard);
  range+= 32;
  while(!found  && standard > 4){
    found = terrMap.find_free_loc(center_loc, loc, range, standard--);
  }
  terrMap.print_map();
  return TerrainBase::Loc(loc.x, loc.y);
}
/*********************JAMES' STUFF **************************/ 

void AttackCommander::updateEnemies()
{
    sint4 enemy = (me == 0) ? 1 : 0;

    enemies.clear();
    myWatcher.queryUnits(OwnerIs(enemy), enemies);
}

GameObj * AttackCommander::getNearestUnit(const std::string & bp_name,
                                          real8 * distance) const
{
    // Select the closest target
    GameObj * target = NULL;
    sint4 targetDistSq = 0;
    FORALL(enemies,e)
    {
	GameObj * enemy = (*e);
        if( enemy->bp_name() != bp_name )
        {
            continue;
        }
        TerrainBase::Loc pos;
        enemy->get_center(pos.x, pos.y);
        sint4 distSq = CalcDistSq(center, pos);
        if( target == NULL || distSq < targetDistSq )
        {
            target = enemy;
            targetDistSq = distSq;
        }
    }
    if(distance != NULL)
    {
        *distance = sqrt( (real8)targetDistSq );
    }
    return target;
}

void AttackCommander::exec()
{
    const sint4 enterCombatDist = 8 * 16; // Just outside firing range
    const sint4 exitCombatDist = 10 * 16; // Well outside firing range

    const sint4 enterCombatDistSq = enterCombatDist * enterCombatDist;
    const sint4 exitCombatDistSq = exitCombatDist * exitCombatDist;

    if(unit2state.empty())
    {
        return;
    }

    findCentroid();

    updateEnemies();

    target = getNearestUnit("tank");
    if( target != NULL )
    {
        std::cerr << "Hunting tanks" << std::endl;
    }
    else
    {
        target = getNearestUnit("controlCenter");
        if( target != NULL )
        {
            std::cerr << "Looting bases" << std::endl;
        }
        else
        {
            std::cerr << "Nothing left to attack" << std::endl;
            return;
        }
    }

    TerrainBase::Loc targetLoc = formLoc;
    target->get_center(targetLoc.x, targetLoc.y);

    SquadCombatAI & scai = moduleSet.get_combat();

    // Choose an action for every unit
    attackers.clear();
    FORALL(unit2state, u2s)
    {
        GameObj *   unit  = u2s->first;
        UnitState & state = u2s->second;

        TerrainBase::Loc unitLoc;
        unit->get_center(unitLoc.x, unitLoc.y);
        sint4 distSq = CalcDistSq(targetLoc, unitLoc);

        if(state.mode == UnitState::IDLE)
        {
            if(distSq < enterCombatDistSq)
            {
                state.mode = UnitState::FIGHTING;
            }
            else
            {
                call_handlers( TerrainModule::PathEvent( me,
                               TerrainModule::FIND_PATH_MSG,
			       TerrainBase::Obj2ObjTask(unit, target) ) );
                state.mode = UnitState::MOVING;
            }
        }
        
        if(state.mode == UnitState::MOVING)
        {
            if(distSq < enterCombatDistSq)
            {
                state.mode = UnitState::FIGHTING;
            }
        }
      
        if(state.mode == UnitState::FIGHTING)
        {
            if(distSq > exitCombatDistSq)
            {
                state.mode = UnitState::IDLE;
            }
        }

        if(state.mode == UnitState::FIGHTING)
        {
            scai.enable_unit_motion(unit);
        }
        else
        {
            scai.disable_unit_motion(unit);
        }
    }
}
