#include "GatherCommander.H"

#include "Gather_accessories.H"
#include <numeric>
#include "Route.H"
#include "RouteAllocator.H"
#include <math.h>

#define MINERAL_ACCESS_POINT_THRESHOLD 6

#define TEST(x)

const static real8 PI = 3.141592653589793;
const static real8 DIST = 10.0;
const static real8 LS = cos( PI * 0.125 ) * DIST;
const static real8 SS = sin( PI * 0.125 ) * DIST;

//****************************Function Objects**************************************

void GatherCommander::computeMineralClusters()
{
    std::cout << "compute MineralClusters " << std::endl;

    mineralClusters.clear();
    if (mineralRecs.empty())
    {
        // Do nothing
    }
    else if (mineralRecs.size() <= 10) 
    {
        boost::shared_ptr<MineralCluster> cluster(new MineralCluster(mineralRecs));
        mineralClusters.push_back(cluster);
    } 
    else 
    {
        // List all mineral locations
        Vector<sint4> clusterOf;
        Vector<ScalarPoint> mineralLocations;
        mineralLocations.reserve(mineralRecs.size());
        FORALL(mineralRecs, mit) 
        {
            const TerrainBase::Loc pos = (*mit)->getPosition();
            ScalarPoint point(pos.x, pos.y);
            mineralLocations.push_back(point);
        }
        
        // Cluster
        simple_clustering(mineralLocations, 2, 8, clusterOf);
        // Store the results
        sint4 clusterCount = *max_element(clusterOf.begin(), clusterOf.end()) + 1;
        mineralClusters.reserve(clusterCount);
        
        FORU(i, clusterCount) 
        {
            boost::shared_ptr<MineralCluster> cluster(new MineralCluster());
            mineralClusters.push_back(cluster);
        }

        Vector<sint4>::iterator cit = clusterOf.begin();
        FORALL(mineralRecs, mrit) 
        {
            mineralClusters[*cit]->addMineral(*mrit);
            cit++;
        }
    }
    
    // Compute cluster centroids
    FORALL(mineralClusters, mcit) 
    {
        (*mcit)->calculateCentroid();
    }
}

void GatherCommander::computeRoutes() 
{
    TEST(std::cout << "computing Routes............" << std::endl);

    if(controlCenters.empty())
    {
        return;
    }

    TerrainBase::Loc ccLoc;
    controlCenters[0]->get_center(ccLoc.x, ccLoc.y);

    sint4 NSEW = 10;
    //sint4 diagx = 9;
    //sint4 diagy = 5;
    sint4 diag = 7;

    TEST(std::cout << "mineral Clusters: " << mineralClusters.size() << std::endl);
    Vector<MineralClusterPtr>::iterator cluster = mineralClusters.begin();
    for (cluster = mineralClusters.begin(); 
         cluster != mineralClusters.end(); 
         cluster++)
    {
        std::set<const History::Record*> mineralsInCluster;
        (*cluster)->getMinerals(mineralsInCluster);
        TEST(std::cout << "\tminerals In Cluster: " << mineralsInCluster.size() 
             << std::endl);

        Vector<MineralPtr> minerals;

        FORALL(mineralsInCluster, record)
        {
            TEST(std::cout << "creating initial points " << minerals.size() 
                 << std::endl);
            TerrainBase::Loc c = (*record)->getPosition();
            Vector<TerrainBase::Loc> tempPoints;

            tempPoints.push_back(TerrainBase::Loc(c.x + LS, c.y + SS));
            tempPoints.push_back(TerrainBase::Loc(c.x + SS, c.y + LS));
            tempPoints.push_back(TerrainBase::Loc(c.x - SS, c.y + LS));
            tempPoints.push_back(TerrainBase::Loc(c.x - LS, c.y + SS));
            tempPoints.push_back(TerrainBase::Loc(c.x - LS, c.y - SS));
            tempPoints.push_back(TerrainBase::Loc(c.x - SS, c.y - LS));
            tempPoints.push_back(TerrainBase::Loc(c.x + SS, c.y - LS));
            tempPoints.push_back(TerrainBase::Loc(c.x + LS, c.y - SS));
            
            boost::shared_ptr<Mineral> mineral(new Mineral(tempPoints, c));
            minerals.push_back(mineral);
        }
        FORALL(minerals, mineralToTest)
        {
            TEST(std::cout << "mineral: " << (*mineralToTest)->centre << std::endl);
            FORALL_DYN((*mineralToTest)->points, pointToTest)
            {
                TEST(std::cout << "\tpoint: " << *pointToTest << std::endl);
                bool clashes = false;
                if (pointToTest->distance((*cluster)->getCentroid()) <
                    (*mineralToTest)->centre.distance((*cluster)->getCentroid()))
                {
                    TEST(std::cout << "\t\tp: " << *pointToTest << " wrong side." 
                         << std::endl);
                    pointToTest = (*mineralToTest)->points.erase(pointToTest);
                    pointToTest--;
                    continue;
                }
                FORALL(minerals, otherMineral)
                {
                    TEST(std::cout << "\t\tagainst mineral: " 
                         << (*otherMineral)->centre << std::endl);
                    if (mineralToTest == otherMineral)
                    {
                        continue;
                    }
                    if (pointToTest->distance((*otherMineral)->centre) < 9) 
                    {
                        TEST(std::cout << "\t\t\tp: " << *pointToTest 
                             << " too close to m: " 
                             << (*mineralToTest)->centre << std::endl);
                        clashes = true;
                        pointToTest = (*mineralToTest)->points.erase(pointToTest);
                        pointToTest--;
                        break;
                    }
                }
                if (clashes)
                    continue;
                FORALL(minerals, otherMineral)
                {
                    TEST(std::cout << "\t\tagainst mineral: " 
                         << (*otherMineral)->centre << std::endl);
                    if (otherMineral == mineralToTest ||
                        pointToTest->distance((*otherMineral)->centre) > 15)
                    {
                        continue;
                    }
                    FORALL((*otherMineral)->points, otherPoint)
                    {
                        TEST(std::cout << "\t\t\tagainst point: " << *otherPoint 
                             << std::endl);
                        if (pointToTest->distance(*otherPoint) < 8)
                        {
                            TEST(std::cout << "\t\t\t\tp: " << *pointToTest 
                                 << " too close to p: " << *otherPoint << std::endl);
                            clashes = true;
                            pointToTest = (*mineralToTest)->points.erase(pointToTest);
                            pointToTest--;
                            break;
                        }
                    }
                    if (clashes)
                        break;
                }
            }
        }
        TEST(std::cout <<  "end of cluster" << std::endl);
        TEST(std::cout << minerals.size() << " " << mineralsInCluster.size() 
             << " " << mineralClusters.size() << std::endl);
        Vector<boost::shared_ptr<Mineral> >::iterator mit = minerals.begin();
        FORALL(mineralsInCluster, mrit)
        {
            FORALL((*mit)->points, p)
            {
                TEST(std::cout << "adding Route at: " << *p << " for mineral " 
                     << (*mit)->centre << std::endl);
                
                RoutePtr route(new Route(this,
                                         *mrit,
                                         *p,
                                         controlCenters[0]));
                routeAllocator->addRoute(route);
            }
            mit++;
        }
    }
    TEST(std::cout << "......finished computing routes" << std::endl);
}
/*
sint4 GatherCommander::computeNumberOfWorkersOnRoute(AccessPoint ap) {
    ap.setReturnPoint(Loc(4,4));
    return 1;//((ap.loc.distance(ap.returnPoint) / 4) * 2 + 50) / 50;
}

*/
