#include "MineralCluster.H"

#include <numeric>

/**
 * sumXYPositions are used by computeMineralClusters() as a function object for
 * the accumulate STL algorithm.
 */

inline sint4 sumXPosition(sint4 v, const History::Record *mineral) 
{
    return v + mineral->getPosition().x;
}

inline sint4 sumYPosition(sint4 v, const History::Record *mineral) 
{
    return v + mineral->getPosition().y;
}

MineralCluster::MineralCluster() { }

MineralCluster::MineralCluster(std::set<const History::Record*> &minerals_)
    : myMinerals(minerals_) { }

void MineralCluster::addMineral(const History::Record *mineral) 
{
    myMinerals.insert(mineral);
}

void MineralCluster::calculateCentroid() 
{
    real8 count = myMinerals.size();

    centroid.x = (sint4) (accumulate(myMinerals.begin(), 
                                     myMinerals.end(), 
                                     0, 
                                     sumXPosition) 
                          / count);

    centroid.y = (sint4) (accumulate(myMinerals.begin(), 
                                           myMinerals.end(), 
                                           0, 
                                           sumYPosition) 
                                / count);
}

TerrainBase::Loc MineralCluster::getCentroid() const 
{
    return centroid;
}

void MineralCluster::getMinerals(std::set<const History::Record *> &minerals) const 
{
    minerals = myMinerals;
}

sint4 MineralCluster::getNumMinerals() const 
{
    return myMinerals.size();
}

void MineralCluster::setMinerals(std::set<const History::Record *> &minerals) 
{
    myMinerals = minerals;
}
