#include "Route.H"

#include "GatherCommander.H"

#define TEST(x) x

Route::Route(GatherCommander *commander_,
             const History::Record *mineralRec_,
             TerrainBase::Loc accessPoint_,
             GameObj *cc_,
             sint4 max) 
    : commander(commander_),
      mineralRec(mineralRec_),
      accessPoint(accessPoint_),
      controlCenter(cc_),
      currentWorkers(0),
      maxWorkers(max),
      routeLength(0)
{ 
    setRouteLength(getSimpleRouteLength());
}

Route::~Route() { }

void Route::addWorker()
{
    currentWorkers++;
}

TerrainBase::Loc Route::getAccessPoint() const
{
    return accessPoint;
}

GameObj* Route::getCC() const
{
    return controlCenter;
}

sint4 Route::getCurrentWorkers() const
{
    return currentWorkers;
}

sint4 Route::getID() const
{
    return id;
}

sint4 Route::getMaxWorkers() const
{
    return maxWorkers;
}

const History::Record* Route::getMineralRec() const 
{ 
    return mineralRec; 
}

real8 Route::getRealRouteLength() const
{
    /*
    TerrainBase::Loc ccLoc;
    controlCenter->get_center(ccLoc.x, ccLoc.y);
    real8 dist = commander->getDistance(controlCenter, accessPoint, 3);
    TEST(std::cout << "real dist: " << dist << " for " << accessPoint << std::endl);
    */
    return -1;//dist;
}

real8 Route::getRouteLength() const
{
    return routeLength;
}

real8 Route::getSimpleRouteLength() const
{
    TerrainBase::Loc ccLoc;
    controlCenter->get_center(ccLoc.x, ccLoc.y);
    return accessPoint.distance(ccLoc);
}

bool Route::mineralIsActive() const
{
    return mineralRec->isActive();
}

void Route::moveToMineral(GameObj *worker) const
{
    commander->moveTo(worker, accessPoint);
}

void Route::moveToCC(GameObj *worker) const
{
    commander->moveTo(worker, controlCenter);
}

void Route::removeWorker() 
{ 
    currentWorkers--; 
}

void Route::setID(sint4 id_) 
{ 
    id = id_; 
}

void Route:: setRouteLength(real8 length)
{
    routeLength = length;
}

void Route::planSuccessOccurred() 
{
    planFailures = 0;
}

void Route::planFailureOccurred()
{
    planFailures++;
}
