#include "RouteAllocator.H"

#include "Route.H"
#include <algorithm>

#define TEST(x)

using namespace std;

///////////////////////////////CONSTRUCTORS/DESTRUCTORS/////////////////////////////

RouteAllocator::RouteAllocator()
    : workers(0),
      pos(0),
      idPos(0)
{ }

RouteAllocator::RouteAllocator(sint4 workers_)
    : workers(workers_),
      pos(0),
      idPos(0)
{ }

RouteAllocator::~RouteAllocator()
{ }

////////////////////////////////////////PRIVATE/////////////////////////////////////

sint4 RouteAllocator::newID() {
    idPos++;
    return idPos;
}

/////////////////////////////////////////PUBLIC/////////////////////////////////////

void RouteAllocator::addRoute(RoutePtr route_) 
{
        route_->setID(newID());
        routes.push_back(route_);
}

void RouteAllocator::addWorker()
{
    workers++;
}

RoutePtr RouteAllocator::getBestRoute(TerrainBase::Loc locationOfWorker)
{
    RoutePtr smallestRoute;
    real8 smallestDist = 10000;
    FORALL(routes, rit)
    {
        if ((*rit)->getCurrentWorkers() >= (*rit)->getMaxWorkers())
        {
            continue;
        }

        real8 distance = (*rit)->getAccessPoint().distance(locationOfWorker);
        if (distance < smallestDist) 
        {
            smallestDist = distance;
            smallestRoute = *rit;
        }
    }

    return smallestRoute;
} 

sint4 RouteAllocator::getNumRoutes() 
{
    return routes.size();
}

uint4 RouteAllocator::getNumAvailableRoutes()
{
    sint4 count = 0;
    FORALL(routes, rit)
    {
        if ((*rit)->getCurrentWorkers() < (*rit)->getMaxWorkers())
        {
            count++;
        }
    }
    return count;
}

void RouteAllocator::markRouteAsBad(RoutePtr route)
{
    route->planFailureOccurred();
    if(route->getPlanFailures() > 10)
    {
        FORALL_DYN(routes, rit)
        {
            if (route == *rit)
            {
                routes.erase(rit);
                break;
            }
        }
    }
}

RoutePtr RouteAllocator::getWorstUsedRoute() 
{
    RoutePtr worstRoute;
    Vector<RoutePtr>::reverse_iterator rit;
    for(rit = routes.rbegin(); rit != routes.rend(); rit++)
    {
        if ((*rit)->getCurrentWorkers() > 0) 
        {
            worstRoute = *rit;
        }
    }
    return worstRoute;
}

void RouteAllocator::removeRoute(RoutePtr route_) 
{
    FORALL_DYN(routes, rit)
    {
        if (*rit == route_)
        {
            routes.erase(rit);
            break;
        }
    }
}

void RouteAllocator::removeWorker()
{
    workers++;
}

void RouteAllocator::removeInaccessibleRoutes()
{
    Vector<RoutePtr>::iterator rit = routes.begin();
    while (rit != routes.end())
    {
        if ((*rit)->getRouteLength() < 1)
        {
            TEST(std::cout << "removing inaccessible route for " 
                 << (*rit)->getAccessPoint() << std::endl);
            rit = routes.erase(rit);
        }
        else
        {
            rit++;
        }
    }
}

void RouteAllocator::setWorkers(sint4 numWorkers)
{
    workers = numWorkers;
}

void RouteAllocator::sortRoutesAccordingToLocation(TerrainBase::Loc location) 
{
    removeInaccessibleRoutes();
    sort(routes.begin(), routes.end(), CompareRoutesByDistanceToLocation(location));
}

size_t RouteAllocator::countWorkersNeeded() const
{
    size_t count = 0;
    FORALL(routes, rit)
    {
        count += (*rit)->getMaxWorkers();
    }
    return count;
}
