#include "Worker.H"

#include "Route.H"
#include "PlayerInfo.H"
#include "GameStateModule.H"
#include "Game.H"
#include "GameObj.H"
#include "GatherCommander.H"

const std::string miningComponent = "pickaxe";
const std::string miningAction = "mine";
const std::string returnResourcesAction= "return_resources";

/////////////////////////////// Constructors ///////////////////////////

Worker::Worker(GameObj* gob_, GatherCommander *commander_, GameStateModule *gsm_)
    : gsm(gsm_),
      commander(commander_),
      gob(gob_),
      state(IDLE),
      knowsRoute(false)
{ }

Worker::~Worker() 
{ 
    if(route != 0)
    {
        route->removeWorker();
    }
}

//////////////////////////////////////PUBLIC///////////////////////////////////////

void Worker::clearRoute() 
{
    if(route != 0)
    {
        route->removeWorker();
        route.reset();
    }
}

bool Worker::closeEnoughToDropOff() const 
{
    assert(route != 0);

    sint4 dist = (sint4)(Game::distance(*gob, *route->getCC()));
    return dist <= WORKER_CC_DISTANCE_THRESHOLD;
}

bool Worker::closeEnoughToMine() const 
{
    assert(route != 0);

    if (route->mineralIsActive())
    {
        sint4 dist = (sint4)(Game::distance(*gob, *(route->getMineralRec()->getObject())));
        if (dist > WORKER_MINERAL_DISTANCE_THRESHOLD) 
        {
            commander->moveTo(gob, route->getMineralRec()->getObject());
            return false;
        }
        else
            return true;
    }
    return false;
}

void Worker::dropOffMinerals() 
{
    assert(route != 0);

    PlayerInfo& pi = gsm->get_game().get_cplayer_info();
    Vector<sint4> args(1, pi.get_id(route->getCC()));
    gob->set_action(returnResourcesAction, args);
    setState(IDLE);
}

GameObj* Worker::getControlCenter() const
{
    assert(route != 0);

    return route->getCC();
}

sint4 Worker::getCurrentMinerals() const 
{ 
    return gob->get_int("minerals"); 
}

GameObj* Worker::get_gob() const 
{
    return gob;
}

TerrainBase::Loc Worker::getLocation() const 
{ 
    return TerrainBase::Loc(*(gob->sod.x),*(gob->sod.y)); 
}

const History::Record* Worker::getMineralRec() const 
{ 
    if(route != 0)
    {
        return route->getMineralRec(); 
    }
    else
    {
        return NULL;
    }
}

RoutePtr Worker::getRoute() const 
{
    // PRECONDITION //
    assert( route != 0 );

    return route;
}

sint4 Worker::getState() const
{
    return state;
}

bool Worker::hasRoute() const 
{
    return route != 0;
}

bool Worker::isMining() const 
{
    return state == MINING;
}

void Worker::mine() 
{
    assert(route != 0);

    setState(MINING);
    PlayerInfo& pi = gsm->get_game().get_cplayer_info();
    Vector<sint4> args(1, pi.get_id(route->getMineralRec()->getObject()));
    gob->component(miningComponent)->set_action(miningAction, args);
}

void Worker::moveToCC() 
{
    assert(route != 0);

    route->moveToCC(gob);
    setState(RETURNING);
}

void Worker::moveToMineral() 
{
    assert(route != 0);

    setState(MOVING);
    route->moveToMineral(gob);
}

void Worker::setRoute(RoutePtr route) 
{
    assert(route != NULL);

    if(this->route != 0)
    {
        this->route->removeWorker();
    }

    this->route = route;
    this->route->addWorker();
}

void Worker::setState(sint4 state_)
{
    state = state_;
}
