#ifndef __FINANCE_H__
#define __FINANCE_H__

#include "EventHandler.H"

class GameStateModule;
namespace History { class HistoryModule; }
class ScriptObj;

namespace Finance
{
    typedef uint4 HoldID;

    class Module;
    class Hold;

    class Ticket
    {
        friend class Module;
        friend class Hold;

        Module & fm;
        uint4    value;

        Ticket(Module & fm, uint4 value);
        ~Ticket();
    public:
        uint4    available() const;
        Ticket * reserve(uint4 amount);
        void     trim(uint4 amount);
        void     spend();
        void     release();
    };

    class Hold
    {
        friend class Module;

        Module & fm;
        uint4    value;

        Hold(Module & fm, uint4 value);
        ~Hold();
    public:
        bool     ready() const;
        void     cancel();
        Ticket * ticket();
    };

    class Module : public EventHandler
    {
        friend class Ticket;
        friend class Hold;

        GameStateModule & gsm;
        History::HistoryModule &   hm;

        // Financial information
        std::list<Hold*> holds;
        uint4 totalResources;
        uint4 totalHeld;
        uint4 totalReserved;
        uint4 totalSpent;

        // Net worth information
        uint4 updateMapCounter;
        uint4 maxMapPatches;
        uint4 maxMapMinerals;
        uint4 curMapMinerals;

        std::set<ScriptObj*> alliedWorkers;
        uint4 alliedCurrentMining;
        uint4 alliedLiveUnitWorth;
        uint4 alliedDeadUnitWorth;
        uint4 enemyDeadUnitWorth;
    public:
        Module(GameStateModule & gsm, History::HistoryModule & hm);
        ~Module();

        bool handle_event(const Event & event);

        /// Determine the amount of resources
        /// which are immediately available
        uint4 available() const;

        /// Reserve immediately available resources
        /// to be spent at a later time
        Ticket * reserve(uint4 amount);

        /// Reserve all incoming resources until a
        /// certain amount have been accumulated
        Hold * hold(uint4 amount);

        /// Get the cost of a given unit via its
        /// blueprint name
        uint4 get_cost(const std::string & bp_name) const;

        /// Get the total number of minerals that have been mined
        uint4 getTotalMined() const;

        /// Get the total worth of all living units and players
        uint4 getTotalWorth() const;

        /// Get the total worth of all allied units, and allied resources
        uint4 getAlliedWorth() const;

        /// Get the total worth of all enemy units, and enemy resources
        uint4 getEnemyWorth() const;
    };
}

#endif
