// $Id: SquadCombatAIWidget.C,v 1.2 2007/05/28 02:12:47 orts_furtak Exp $

// SquadCombatAI Widget Implementation

#include "SquadCombatAIWidget.H"
#include "GfxModule.H"
#include "GfxGlobal.H"
#include "Profiler.H"
#include "InfluenceMap.H"

#include "Game.H"
#include "GameStateModule.H"

using namespace std;

//===================================================================

SquadCombatAIOverlay::SquadCombatAIOverlay(const SquadCombatAI &scai, GfxModule &gfxm_)
  : dot(gfxm_), gfxm(&gfxm_)
{
  enabled = true;
  show_avg_dist = true;
  show_dir = true;
  show_friend_influence = false;
  show_target_influence = false;
  show_motion_field = false;
  show_occupied_field = false;
  color_units = true;
  sc = &scai;
}

//-------------------------------------------------------------------

bool SquadCombatAIOverlay::handle_event(const Event &e)
{
  if (e.get_who() == GfxModule::FROM) {
    if (e.get_what() == GfxModule::DRAW_MSG) {
      draw();
      return true;
    }
  }
  return false;
}

//-------------------------------------------------------------------

void SquadCombatAIOverlay::draw()
{
  PROFILE("SC Overlay");

  if (!enabled) return;

  // --- friendly attack influence -------------------------

  if (sc->friendlies && show_friend_influence) {
    InfluenceMap &im = sc->friendlies->attack_influence;
    im.draw(dot, gfxm->get_game(), Vec3<real4>(1.0, 0.0, 0.0));
  }

  // --- enemy target influence ----------------------------
  
  if (sc->enemies && show_target_influence) {
    InfluenceMap &im = sc->enemies->target_influence;
    im.draw(dot, gfxm->get_game(), Vec3<real4>(0.0, 0.5, 1.0));
  }

  // -------------------------------------------------------

  FORALL (sc->groups, g) {
    //    InfluenceMap &im = (*g)->occupied_influence;
    //    im.draw(dot, gfxm->get_game(), Vec3<real4>(0.0, 1.0, 0.0));
  }

#if 0
  sint4 mw = sc->terrain_field.mw;
  sint4 mh = sc->terrain_field.mh;

  dot.begin_lines();

  if (sc->terrain_field.marked) {
    for (sint4 tx = 0; tx <= mw; ++tx) {
      for (sint4 ty = 0; ty <= mh; ++ty) {
        sint4 t = ty * mw + tx;
        if (sc->terrain_field.marked[t]) {
          real4 px = tx / 2.0;
          real4 py = ty / 2.0;

          dot.draw_circle(px, py, 0.5, Vec3<real4>(1,0,1));
        }
      }
    }
  }

  dot.end_lines();
#endif

#if 0

  dot.begin_lines();
  FORALL (sc->squads, sq) {
    SquadCombatAI::Squad *s = *sq;
    //    dot.draw_circle(s->x / 16.0, s->y / 16.0, 5, Vec3<real4>(1,0.4,0.4));
    if (s->objective == SquadCombatAI::Squad::ATTACK_TARGET) {
      dot.draw_circle(s->x / 16.0, s->y / 16.0, 5, Vec3<real4>(0,0,1));
    }
    FORALL (s->units, u) {
      GameObj *g = *u;
      sint4 x, y;
      g->get_center(x, y);
      if (s->objective == SquadCombatAI::Squad::ATTACK_TARGET) {
        dot.draw_line(s->x / 16.0, s->y / 16.0, x / 16.0, y / 16.0, Vec3<real4>(1,0,0));
      }
      dot.draw_circle(x / 16.0, y / 16.0, 1, Vec3<real4>(1,0,0));
    }
  }
  dot.end_lines();
#endif


  // -------------------------------------------------------

  if (show_avg_dist && sc->friendlies) {
    dot.begin_lines();
    FORALL (sc->friendlies->units, uu) {
      GameObj *u = *uu;
      SquadCombatAI::ShooterMap::const_iterator it = sc->shooter_data.find(u);
      if (it == sc->shooter_data.end()) continue;
      const SquadCombatAI::ShooterData &sd = it->second;

      // --- diff to enemy ---------------------------------

      if (show_avg_dist) {
        real4 i = sd.diff_to_enemy / 32.0;
        if (i > 0.8) i = 0.8;
        dot.draw_circle(sd.x / 16.0,
                        sd.y / 16.0,
                        0.2,
                        Vec3<real4>(0.0, 0.2 + i, 0.0));
      }

      // --- direction -------------------------------------

      if (show_dir) {
        dot.draw_line(sd.x / 16.0,
                      sd.y / 16.0,
                      sd.x / 16.0 + sd.force1.x,
                      sd.y / 16.0 + sd.force1.y,
                      Vec3<real4>(1.0, 0.0, 0.0));

        dot.draw_line(sd.x / 16.0,
                      sd.y / 16.0,
                      sd.x / 16.0 + sd.force2.x,
                      sd.y / 16.0 + sd.force2.y,
                      Vec3<real4>(1.0, 1.0, 0.0));
      }

    }

    dot.end_lines();
  }

  // --- isolated target -----------------------------------

  dot.begin_lines();
  FORALL (sc->squads, sq) {
    SquadCombatAI::Squad *s = *sq;
    dot.draw_circle(s->x / 16.0, s->y / 16.0, 0.4, Vec3<real4>(1,1,0));
    dot.draw_circle(s->x / 16.0, s->y / 16.0, 0.7, Vec3<real4>(1,1,0));
    dot.draw_circle(s->x / 16.0, s->y / 16.0, 1.0, Vec3<real4>(1,1,0));
  }
  FORALL (sc->squads, sq) {
    SquadCombatAI::Squad *s = *sq;
    dot.draw_circle(s->pos_x / 16.0, s->pos_y / 16.0, 0.4, Vec3<real4>(0,1,1));
    dot.draw_circle(s->pos_x / 16.0, s->pos_y / 16.0, 0.7, Vec3<real4>(0,1,1));
    dot.draw_circle(s->pos_x / 16.0, s->pos_y / 16.0, 1.0, Vec3<real4>(0,1,1));
  }
  dot.end_lines();

  // --- motion field --------------------------------------

  if (show_motion_field) {
    FORALL (sc->groups, g) (*g)->motion_field.draw(dot, gfxm->get_game(), Vec3<real4>(0.0, 1.0, 0.0));
  }

  // --- occupied field ------------------------------------

  if (show_occupied_field) {
    sc->occupied_field.draw(dot, gfxm->get_game(), Vec3<real4>(1.0, 0.0, 1.0));
  }

  // --- color units ---------------------------------------

  if (color_units) {
    dot.begin_squares();
    if (sc->enemies) {
      FORALL (sc->enemies->units, u) {
        GameObj *g = *u;
        sint4 x, y;
        g->get_center(x, y);
        dot.draw_square(x/16.0 - 0.2, y/16.0 - 0.2,
                        x/16.0 + 0.2, y/16.0 + 0.2,
                        Vec3<real4>(1,0,0));
      }
    }
    if (sc->friendlies) {
      FORALL (sc->friendlies->units, u) {
        GameObj *g = *u;
        sint4 x, y;
        g->get_center(x, y);
        dot.draw_square(x/16.0 - 0.2, y/16.0 - 0.2,
                        x/16.0 + 0.2, y/16.0 + 0.2,
                        Vec3<real4>(0,0,1));
        /*
        SquadCombatAI::ShooterMap::const_iterator it = sc->shooter_data.find(g);
        if (it == sc->shooter_data.end()) continue;
        const SquadCombatAI::ShooterData &sd = it->second;
        if (!sd.motion_enabled) continue;

        dot.draw_square(x/16.0 - 0.3, y/16.0 - 0.3,
                        x/16.0 + 0.3, y/16.0 + 0.3,
                        Vec3<real4>(0,1,0));
        */
      }
    }
    dot.end_squares();
  }

  // -------------------------------------------------------

#if 0
  if (sc->friendlies) {
    VectorField &vf = sc->friendlies->attack_field;
    vf.draw(dot, gfxm->get_game(), Vec3<real4>(1.0, 0.0, 0.0));

    dot.begin_lines();

    FORALL (sc->friendlies->units, uu) {
      
      GameObj *u = *uu;
      SquadCombatAI::ShooterMap::const_iterator it = sc->shooter_data.find(u);
      if (it == sc->shooter_data.end()) continue;
      const SquadCombatAI::ShooterData &sd = it->second;

      sint4 r = sd.r < 0 ? 0 : sd.r;
      if (sd.range >= 0) r += sd.range;

      real4 i = sd.support;
      if (i > 0.8) i = 0.8;

      dot.draw_circle(sd.x / 16.0,
                      sd.y / 16.0,
                      r / 16.0,
                      Vec3<real4>(0.2 + i, 0.2 + i, 0.0));
    }

    dot.end_lines();
  }
#endif

#if 0
  if (sc->enemies) {
    VectorField &vf = sc->enemies->target_field;
    vf.draw(dot, gfxm->get_game(), Vec3<real4>(0.0, 0.0, 1.0));
  }
#endif
  
  //-------------------------------------

#if 0
  {
    const TerrainVectorField &vf = sc->terrain_field;
    vf.draw(dot, gfxm->get_game(), Vec3<real4>(0.0, 1.0, 0.0));
  }
#endif

  //-------------------------------------

  glEnable(GL_TEXTURE_2D);
}

//===================================================================
//===================================================================

SquadCombatAIWidget::SquadCombatAIWidget(SquadCombatAI &scai, GfxModule *gfxm)
{
  if (gfxm) {
    overlay = new SquadCombatAIOverlay(scai, *gfxm);
    gfxm->add_handler(overlay);
    overlay->enable();
  } else {
    overlay = 0;
  }

  set_name("SquadCombat");

  sc = &scai;
}

//-------------------------------------------------------------------

SquadCombatAIWidget::~SquadCombatAIWidget()
{
  if (overlay) delete overlay;
}

//===================================================================

stype SquadCombatAIWidget::event_key_down(stype key, stype /*mod*/, stype x, stype y)
{
  if (is_outside(x,y)) return stype();
  cout << "event_key_down" << endl;
  
  char index = (char)key.to_int();
  
  switch(index){
    
  case '\t':
    if (overlay) {
      if (overlay->is_enabled()) {
        cout << "overlay off" << endl;
        overlay->disable();
      } else {
        cout << "overlay on" << endl;
        overlay->enable();
      }
    }
    return stype(1);

  case '1': if (overlay) overlay->toggle_friend_influence(); return stype(1);
  case '2': if (overlay) overlay->toggle_target_influence(); return stype(1);
  case '3': if (overlay) overlay->toggle_dir(); return stype(1);
  case '4': if (overlay) overlay->toggle_motion_field(); return stype(1);
  case '5': if (overlay) overlay->toggle_occupied_field(); return stype(1);
  case '6': if (overlay) overlay->toggle_color_units(); return stype(1);

  case 'd':
    if (overlay) overlay->toggle_avg_dist();
    return stype(1);

  case 'a':
    if (sc) {
      if (sc->is_enabled_attack()) {
        cout << "disable attack" << endl;
        sc->disable_attack();
      } else {
        cout << "enable attack" << endl;
        sc->enable_attack();
      }
    }
    return stype(1);

  case 'm':
    if (sc) {
      if (sc->is_enabled_motion()) {
        cout << "disable motion" << endl;
        sc->disable_motion();
      } else {
        cout << "enable motion" << endl;
        sc->enable_motion();
      }
    }
    return stype(1);


  default:
    break;
  }

  return stype();
}

//===================================================================

stype SquadCombatAIWidget::draw()
{
  if (!!off) return stype();
  if (sc->friendlies == 0) return stype();

  InfluenceMap &im = sc->friendlies->attack_influence;

  sint4 map_width = im.get_width();
  sint4 map_height = im.get_height();

  sint4 x = this->x.to_int();
  sint4 y = this->y.to_int();
  sint4 w = this->w.to_int();
  sint4 h = this->h.to_int();

  real8 xd = (real8)w/(real8)map_width;
  real8 yd = (real8)h/(real8)map_height;

  draw_grid(map_width, map_height);
  
  FORS (xi, map_width) {
    FORS (yi, map_height) {

      real4 v = im(xi, yi);
      
      if (v > 0.0) {
        
        if (v > 1.0) v = 1.0;
        Vec3<real4> c(v, 0, 0);

        glColor4f(c.r, c.g, c.b, 1.0);

        glBegin(GL_QUADS);
        
        glVertex2f( x+xi*xd    , y+h-yi*yd   -yd );
        glVertex2f( x+xi*xd+xd , y+h-yi*yd   -yd );
        glVertex2f( x+xi*xd+xd , y+h-yi*yd+yd-yd );
        glVertex2f( x+xi*xd    , y+h-yi*yd+yd-yd );
        
        glEnd();
      }
    }
  }
  
  glEnable(GL_TEXTURE_2D);
  
  return stype();
}

//===================================================================

