#ifndef __BASE_CLASSIFIER_H__
#define __BASE_CLASSIFIER_H__

#include "TerrainAnalysis.H"
#include "History.H"
#include "ModuleSet.H"

namespace TerrainAnalysis
{
    class BaseInfo
    {
    public:
        enum OWNER { UNCLAIMED, ALLIED, ENEMY, CONTESTED };
    private:
        ResourceClusterInfo resources;
        std::set<const History::Record *> alliedBuildings;
        std::set<const History::Record *> enemyBuildings;
    public:
        BaseInfo(const ResourceClusterInfo & resources);
        ~BaseInfo() {}
        void addAlliedBuilding(const History::Record * building);
        void addEnemyBuilding(const History::Record * building);
        size_t countAlliedBuildings() const;
        size_t countAlliedBuildings(const std::string & bp_name) const;
        size_t countEnemyBuildings() const;
        size_t countEnemyBuildings(const std::string & bp_name) const;
        TerrainBase::Loc getResourceLocation() const;
        TerrainBase::Loc getCentralLocation() const;
        sint4 resourceWorth() const;
        sint4 alliedWorth() const;
        sint4 enemyWorth() const;
        OWNER getOwner() const;
    };

    class BaseClassifier
    {
        ModuleSet &                  ms;

        Vector<BaseInfo*>       bases;
        Vector<const BaseInfo*> const_bases;
    public:
        BaseClassifier(ModuleSet & ms);
        ~BaseClassifier();

        void clear();
        void update();
        const Vector<const BaseInfo*> & get_bases() const;
    };
}

#endif
