// $Id: AI_PathFind.H,v 1.2 2004/10/14 23:43:36 orts_mburo Exp $

// This is an ORTS file (c) Michael Chung, licensed under the GPL

#ifndef AI_PathFind_H
#define AI_PathFind_H

#include "pathfind.h"

#include "Coord.H"

/** Description

    Code for pathfinding in the AI
*/

#define node_id_type int

class AI_Board;
class AI_Path;

/** pathfinding function 
    @param path  object to receive the new path
*/
bool find_path(
  AI_Path &path,
  const AI_Board &board,
  Point pt_start,
  Point pt_goal);

class AI_PathFind_Environment : public PathFind::Environment {

public:
  AI_PathFind_Environment();

  void attach_board(const AI_Board *board);
  void detach_board();

  node_id_type point_to_id(const Point &pt) const;
  Point id_to_point(node_id_type id) const;

  void set_start(node_id_type id);
  void set_goal(node_id_type id);

public:
  // PathFind::Environment methods
  virtual int getHeuristic(node_id_type start, node_id_type target) const;
  virtual int getMaxCost() const;
  virtual int getMinCost() const;
  virtual int getNumberNodes() const;
        
  /** Generate successor nodes for the search.
      @param nodeId Current node
      @param lastNodeId
      Can be used to prune nodes,
      (is set to NO_NODE in Search::checkPathExists).
      @param result Resulting successors.
      Passed in as a reference to avoid dynamic memory allocation.
      The vector will be cleared and filled with the successors.
  */
  virtual void getSuccessors(node_id_type nodeId, node_id_type lastNodeId,
    std::vector<Successor>& result,
    bool usePruning = false) const;
        
  virtual bool isValidNodeId(node_id_type nodeId) const;
  
private:
  // pointer to attached board object
  const AI_Board *m_board;

  node_id_type m_start;
  node_id_type m_goal;
};

#endif
